// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;

namespace Smdn.Net.Imap4.Client.Session {
  internal class ImapReferralResponseProcessor : ImapResponseProcessorBase {
    public ImapReferralResponseProcessor(ImapSession session)
      : base(session)
    {
    }

    public override bool Process(ImapResponse response, IImapTransaction t)
    {
      var status = response as ImapStatusResponse;

      if (status == null)
        return false;

      if (status.ResponseText.Code != ImapResponseCode.Referral)
        return false;

      if ((t is GreetingTransaction) || (t is AuthenticateTransaction) || (t is LoginTransaction))
        return ProcessLoginReferrals(status, t);
      else
        return ProcessMailboxReferrals(status, t);
    }

    // RFC 2221 IMAP4 Login Referrals
    // http://tools.ietf.org/html/rfc2221
    private bool ProcessLoginReferrals(ImapStatusResponse status, IImapTransaction t)
    {
      if (t is GreetingTransaction) {
        if (status.Condition != ImapResponseCondition.Bye)
          return false;

        // 4.2. BYE at connection startup referral
        //    An IMAP4 server MAY respond with an untagged BYE and a REFERRAL
        //    response code that contains an IMAP URL to a home server if it is not
        //    willing to accept connections and wishes to direct the client to
        //    another IMAP4 server.
        var referToUri = ImapResponseTextConverter.FromReferral(status.ResponseText)[0];

        if (Session.HandlesReferralAsException) {
          throw new ImapLoginReferralException(string.Format("try another server: '{0}'", status.ResponseText.Text), referToUri);
        }
        else {
          Trace.Info("login referral: '{0}'", status.ResponseText.Text);
          Trace.Info("  try to connect to {0}", referToUri);

          return true;
        }
      }
      else {
        var tagged = status as ImapTaggedStatusResponse;

        if (tagged == null)
          return false;

        // 4.1. LOGIN and AUTHENTICATE Referrals
        //    An IMAP4 server MAY respond to a LOGIN or AUTHENTICATE command with a
        //    home server referral if it wishes to direct the user to another IMAP4
        //    server.
        var referToUri = ImapResponseTextConverter.FromReferral(status.ResponseText)[0];

        if (Session.HandlesReferralAsException) {
          throw new ImapLoginReferralException(string.Format("try another server: '{0}'", tagged.ResponseText.Text), referToUri);
        }
        else {
          Trace.Info("login referral: '{0}'", status.ResponseText.Text);

          if (tagged.Condition == ImapResponseCondition.Ok)
            Trace.Info("  another server available at {0}", referToUri);
          else
            Trace.Info("  try to connect to {0}", referToUri);

          return true;
        }
      }
    }

    // RFC 2193 IMAP4 Mailbox Referrals
    // http://tools.ietf.org/html/rfc2193
    private bool ProcessMailboxReferrals(ImapStatusResponse status, IImapTransaction t)
    {
      var tagged = status as ImapTaggedStatusResponse;

      if (tagged == null)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.

      // 4.2. CREATE Referrals
      //    An IMAP4 server MAY respond to the CREATE command with one or more
      //    IMAP mailbox referrals, if it wishes to direct the client to issue
      //    the CREATE against another server.  The server can employ any means,
      //    such as examining the hierarchy of the specified mailbox name, in
      //    determining which server the mailbox should be created on.

      // 4.3. RENAME Referrals
      //    An IMAP4 server MAY respond to the RENAME command with one or more
      //    pairs of IMAP mailbox referrals.  In each pair of IMAP mailbox
      //    referrals, the first one is an URL to the existing mailbox name and
      //    the second is an URL to the requested new mailbox name.

      // 4.4. COPY Referrals
      //    An IMAP4 server MAY respond to the COPY command with one or more IMAP
      //    mailbox referrals.  This indicates that the destination mailbox is on
      //    a remote server.  To achieve the same behavior of a server COPY, the
      //    client MAY issue the constituent FETCH and APPEND commands against
      //    both servers.
      var referrals = ImapResponseTextConverter.FromReferral(status.ResponseText);

      if (Session.HandlesReferralAsException) {
        throw new ImapMailboxReferralException(string.Format("try another server: '{0}'", tagged.ResponseText.Text), referrals);
      }
      else {
        Trace.Info("mailbox referral: '{0}'", tagged.ResponseText.Text);
        Trace.Info("  try mailboxes below:");
        foreach (var referral in referrals) {
          Trace.Info("    {0}", referral);
        }

        return true;
      }
    }
  }
}