// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using System.Threading;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client.Session {
  // Layers:
  // * Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public class ImapSession : IDisposable {
    /*
     * properties
     */
    internal int Id {
      get { return connection.Id; }
    }

    public bool IsDisposed {
      get { return disposed; }
    }

    public bool HandlesReferralAsException {
      get { return handlesReferralAsException; }
      set { handlesReferralAsException = value; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get { CheckDisposed(); return connection.SendTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { CheckDisposed(); return connection.ReceiveTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.ReceiveTimeout = value; }
    }

    public bool IsSecureConnection {
      get { CheckDisposed(); return connection.IsSecureConnection; }
    }

    public int TransactionTimeout {
      get { CheckDisposed(); return transactionTimeout; }
      set
      {
        CheckDisposed();
        RejectTransactionProceeding();

        if (value < -1)
          throw new ArgumentOutOfRangeException("TransactionTimeout", "must be greater than or equals to -1");
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (!transactionProceedingSemaphore.WaitOne(0)) {
          return true;
        }
        else {
          transactionProceedingSemaphore.Release();
          return false;
        }
      }
    }

    /*
     * construction / destruction
     */
    public ImapSession(string host)
      : this(host, ImapDefaultPort.Imap, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port)
      : this(host, port, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout)
      : this(host, port, false, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, transactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException, int transactionTimeout)
      : this(host, ImapDefaultPort.Imap, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout)
      : this(host, port, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (transactionTimeout < -1)
        throw new ArgumentOutOfRangeException("transactionTimeout", "must be greater than or equals to -1");

      this.handlesReferralAsException = handlesReferralAsException;
      this.transactionTimeout = transactionTimeout;

      this.readonlyServerID = serverID.AsReadOnly();
      this.readonlyHierarchyDelimiters = hierarchyDelimiters.AsReadOnly();

      InitializeResponseProcessors();

      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={0}}}",
                           connection,
                           state);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;

    private bool disposed = false;
    private bool handlesReferralAsException;
    private bool handlesIncapableAsException = false;
    private int transactionTimeout = DefaultTransactionTimeout;
    private Semaphore transactionProceedingSemaphore = new Semaphore(1, 1);

    /*
     * session state properties
     */
    internal ImapConnection Connection {
      get { return connection; }
    }

    public bool IsIdling {
      get { CheckDisposed(); return connection.IsIdling; }
    }

    public ImapSessionState State {
      get { CheckDisposed(); return state; }
    }

    public ImapUpdatedStatus LastUpdatedStatus {
      get { CheckDisposed(); return lastUpdatedStatus; }
      internal set { lastUpdatedStatus = value; }
    }

    public ImapMailbox SelectedMailbox {
      get { CheckDisposed(); return selectedMailbox; }
    }

    public Uri Authority {
      get { CheckDisposed(); return authority; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { CheckDisposed(); return serverCapabilities; }
    }

    public IDictionary<string, string> HierarchyDelimiters {
      get { CheckDisposed(); return readonlyHierarchyDelimiters; }
    }

    public ImapNamespace Namespaces {
      get { CheckDisposed(); return namespaces; }
    }

    public IDictionary<string, string> ServerID {
      get { CheckDisposed(); return readonlyServerID; }
    }

    public string SelectedLanguage {
      get { CheckDisposed(); return selectedLanguage; }
    }

    public ImapCollationAlgorithm ActiveComparator {
      get { CheckDisposed(); return activeComparator; }
    }

    private ImapConnection connection = null;
    private ImapMailboxManager mailboxManager = null;
    private ImapSessionState state = ImapSessionState.NotConnected;
    private ImapUpdatedStatus lastUpdatedStatus = new ImapUpdatedStatus();
    private ImapTransactionResult lastTransactionResult = null;
    private ImapMailbox selectedMailbox = null;
    private Uri authority = null;
    private ImapCapabilityList serverCapabilities;
    private Dictionary<string, string> hierarchyDelimiters = new Dictionary<string, string>();
    private ImapNamespace namespaces = new ImapNamespace();
    private Dictionary<string, string> serverID = new Dictionary<string, string>();
    private string selectedLanguage = null;
    private ImapCollationAlgorithm activeComparator = ImapCollationAlgorithm.Default;

    private IDictionary<string, string> readonlyServerID;
    private IDictionary<string, string> readonlyHierarchyDelimiters;

    /*
     * transaction methods : connect/disconnect
     */
    private void Connect(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      TraceInfo("connecting");

      this.connection = new ImapConnection(host, port, createAuthenticatedStreamCallback);

      TraceInfo("connected");

      SetServerCapabilities(new ImapCapabilityList(new[] {
        ImapCapability.Imap4Rev1 /* required by GreetingTransaction */
      }));

      using (var t = new GreetingTransaction(connection)) {
        try {
          if (ProcessTransaction(t, handlesIncapableAsException).Failed)
            throw new ImapConnectionException("initial greeting failed");
        }
        catch (TimeoutException) {
          throw new ImapConnectionException("initial greeting failed");
        }

        if (t.Result.Code == ImapTransactionResultCode.PreAuth) {
          UpdateAuthority(null, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }
        else {
          TransitStateTo(ImapSessionState.NotAuthenticated);
        }

        if (!t.Result.ReceivedResponses.Exists(delegate(ImapResponse response) {
          return (response is ImapStatusResponse) &&
                 (response as ImapStatusResponse).ResponseText.Code == ImapResponseCode.Capability;
        })) {
          // clear server capabilities
          SetServerCapabilities(new ImapCapabilityList());
        }
      }
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect()
    {
      Disconnect(true);
    }

    /// <summary>disconnects session</summary>
    public void Disconnect(bool logout)
    {
      CheckDisposed();

      if (logout) {
        if (state == ImapSessionState.Selected)
          Close();
        if (state == ImapSessionState.Authenticated)
          Logout();
      }

      (this as IDisposable).Dispose();
    }

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        TraceInfo("disconnecting");

        connection.Close();

        TraceInfo("disconnected");
      }
      finally {
        TransitStateTo(ImapSessionState.NotConnected);

        connection = null;
      }
    }

    void IDisposable.Dispose()
    {
      if (disposed)
        return;

      CloseConnection();

      transactionProceedingSemaphore.Close();
      transactionProceedingSemaphore = null;

      disposed = true;
    }

    /*
     * transaction methods : any state
     */

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Capability()
    {
      ImapCapabilityList discard;

      return Capability(out discard);
    }

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Capability(out ImapCapabilityList capabilities)
    {
      RejectNonConnectedState();

      capabilities = null;

      using (var t = new CapabilityTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          SetServerCapabilities(t.Result.Value);
          capabilities = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult NoOp()
    {
      RejectNonConnectedState();

      using (var t = new NoOpTransaction(connection)) {
        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends LOGOUT command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Logout()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already logged out or disconnected");

      using (var t = new LogoutTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Code == ImapTransactionResultCode.Ok) {
          TransitStateTo(ImapSessionState.NotAuthenticated);
          CloseConnection();
        }

        return t.Result;
      }
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult ID(Dictionary<string, string> clientParameterList)
    {
      Dictionary<string, string> discard;

      return ID(clientParameterList, out discard);
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult ID(Dictionary<string, string> clientParameterList, out Dictionary<string, string> serverParameterList)
    {
      RejectNonConnectedState();

      serverParameterList = null;

      using (var t = new IdTransaction(connection)) {
        // client parameter list
        if (clientParameterList != null && 0 < clientParameterList.Count) {
          // RFC 2971 IMAP4 ID extension
          // http://tools.ietf.org/html/rfc2971
          // 3.3. Defined Field Values
          //    Field strings MUST NOT be longer than 30 octets.  Value strings MUST
          //    NOT be longer than 1024 octets.  Implementations MUST NOT send more
          //    than 30 field-value pairs.
          if (30 < clientParameterList.Count)
            throw new ArgumentException("number of client parameter list pairs must be less than or equals to 30");

          var idParamsList = new List<ImapString>();

          try {
            foreach (var param in clientParameterList) {
              if (30 < NetworkTransferEncoding.Transfer7Bit.GetByteCount(param.Key))
                throw new ArgumentException("field string of client parameter list must be shorter than or equals to 30 octets");

              idParamsList.Add(new ImapQuotedString(param.Key));

              if (param.Value == null) {
                idParamsList.Add(new ImapNilString());
              }
              else {
                if (1024 < NetworkTransferEncoding.Transfer7Bit.GetByteCount(param.Value))
                  throw new ArgumentException("value string of client parameter list must be shorter than or equals to 1024 octets");
                idParamsList.Add(new ImapQuotedString(param.Value));
              }
            }
          }
          catch (EncoderFallbackException) {
            throw new ArgumentException("both field and value string must not contain non-ASCII chars");
          }

          t.Request.Arguments["client parameter list"] = new ImapParenthesizedString(idParamsList.ToArray());
        }

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.serverID = t.Result.Value;
          this.readonlyServerID = this.serverID.AsReadOnly();

          serverParameterList = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPRESS command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Compress(ImapCompressionMechanism compressionMechanism)
    {
      RejectNonConnectedState();

      if (compressionMechanism == null)
        throw new ArgumentNullException("compressionMechanism");

      using (var t = new CompressTransaction(connection)) {
        t.Request.Arguments["compression mechanism"] = compressionMechanism;

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Language(out string[] supportedLanguageTags)
    {
      RejectNonConnectedState();

      supportedLanguageTags = null;

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.2. LANGUAGE Command
       *    If there aren't any arguments, the server SHOULD send an untagged
       *    LANGUAGE response listing the languages it supports.  If the server
       *    is unable to enumerate the list of languages it supports it MAY
       *    return a tagged NO response to the enumeration request.
       */
      using (var t = new LanguageTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          supportedLanguageTags = t.Result.Value.Item1;

        return t.Result;
      }
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Language(CultureInfo cultureInfo, params CultureInfo[] cultureInfos)
    {
      string discard;

      return Language(out discard, cultureInfo, cultureInfos);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Language(out string selectedLanguageTag, CultureInfo cultureInfo, params CultureInfo[] cultureInfos)
    {
      if (cultureInfo == null)
        throw new ArgumentNullException("cultureInfo");

      return Language(out selectedLanguageTag, cultureInfo.IetfLanguageTag, Array.ConvertAll(cultureInfos, delegate(CultureInfo info) {
        return info.IetfLanguageTag;
      }));
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>
    /// valid in any state.
    /// this method sends LANGUAGE command with language of "default".
    /// </remarks>
    public ImapTransactionResult Language()
    {
      string discard;

      return Language(out discard);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>
    /// valid in any state.
    /// this method sends LANGUAGE command with language of "default".
    /// </remarks>
    public ImapTransactionResult Language(out string selectedLanguageTag)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.2. LANGUAGE Command
       *    The special "default" language range argument indicates a request to
       *    use a language designated as preferred by the server administrator.
       *    The preferred language MAY vary based on the currently active user.
       */
      return Language(out selectedLanguageTag, "default");
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Language(string languageRange, params string[] languageRanges)
    {
      string discard;

      return Language(out discard, languageRange, languageRanges);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Language(out string selectedLanguageTag, string languageRange, params string[] languageRanges)
    {
      RejectNonConnectedState();

      if (languageRange == null)
        throw new ArgumentNullException("languageRange");
      else if (languageRange.Length == 0)
        throw new ArgumentException("language range must be non-empty string", "languageRange");

      var languageRangeArguments = new List<string>();

      languageRangeArguments.Add(languageRange);

      if (languageRanges != null) {
        for (var i = 0; i < languageRanges.Length; i++) {
          if (languageRanges[i] == null)
            throw new ArgumentNullException(string.Format("languageRanges[{0}]", i));
          else if (languageRanges[i].Length == 0)
            throw new ArgumentException("language range must be non-empty string", string.Format("languageRanges[{0}]", i));

          languageRangeArguments.Add(languageRanges[i]);
        }
      }

      selectedLanguageTag = null;

      using (var t = new LanguageTransaction(connection)) {
        t.Request.Arguments["language range arguments"] = new ImapStringList(languageRangeArguments.ToArray());

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.selectedLanguage = t.Result.Value.Item1[0];
          this.namespaces = t.Result.Value.Item2 ?? namespaces;

          selectedLanguageTag = t.Result.Value.Item1[0];
        }

        return t.Result;
      }
    }

    /*
     * transaction methods : non-authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      return StartTls(createAuthenticatedStreamCallback, false);
    }

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback, bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("already authenticated");

      if (createAuthenticatedStreamCallback == null)
        throw new ArgumentNullException("createAuthenticatedStreamCallback");

      ImapTransactionResult result;

      using (var t = new StartTlsTransaction(connection, createAuthenticatedStreamCallback)) {
        if ((result = ProcessTransaction(t, handlesIncapableAsException)).Succeeded)
          // 6.2.1. STARTTLS Command
          //       Once [TLS] has been started, the client MUST discard cached
          //       information about server capabilities and SHOULD re-issue the
          //       CAPABILITY command.
          SetServerCapabilities(new ImapCapabilityList());
        else
          return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials, ImapAuthenticationMechanism authenticationMechanism)
    {
      return Authenticate(credentials, null, authenticationMechanism);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authenticationMechanism)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already authenticated");

      if (credentials == null)
        throw new ArgumentNullException("credentials");
      if (authenticationMechanism == null)
        throw new ArgumentNullException("authenticationMechanism");

      var credential = credentials.LookupCredential(connection, username, authenticationMechanism);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError, string.Format("credential not found for {0};AUTH={1}@{2}:{3}", username, authenticationMechanism, connection.Host, connection.Port));

      using (var t = new AuthenticateTransaction(connection, credential, serverCapabilities.Has(ImapCapability.SaslIR))) {
        t.Request.Arguments["authentication mechanism name"] = authenticationMechanism;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateAuthority(credential.UserName, authenticationMechanism);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials)
    {
      return Login(credentials, null);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials, string username)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already logged in");

      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (serverCapabilities.Has(Imap4.ImapCapability.LoginDisabled))
        throw new ImapIncapableException("LOGIN is disabled");

      if (credentials == null)
        throw new ArgumentNullException("credentials");

      var credential = credentials.LookupCredential(connection, username, null);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError, string.Format("credential not found for {0}@{1}:{2}", username, connection.Host, connection.Port));

      using (var t = new LoginTransaction(connection, credential)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateAuthority(credential.UserName, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    internal void UpdateAuthority(string username, ImapAuthenticationMechanism authType)
    {
      // TODO: use ImapUriBuilder
      var scheme = string.Format("{0}://", connection.IsSecurePortConnection ? ImapUri.UriSchemeImaps : ImapUri.UriSchemeImap);
      var userInfo = (username == null)
        ? string.Empty
        : (authType == null)
          ? string.Format("{0}@", username)
          : string.Format("{0};AUTH={1}@", username, authType);

      authority = new Uri(string.Format("{0}{1}{2}:{3}", scheme, userInfo, connection.Host, connection.Port));

      TraceInfo("authority: {0}", authority);
    }

    /*
     * transaction methods : authenticated state
     */

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Select(ImapMailbox mailbox)
    {
      return Select(mailbox.Name);
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Select(string mailboxName)
    {
      using (var t = new SelectTransaction(connection, null)) {
        return SelectExamineInternal(t, mailboxName, null);
      }
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult SelectCondStore(ImapMailbox mailbox)
    {
      return SelectCondStore(mailbox.Name);
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult SelectCondStore(string mailboxName)
    {
      using (var t = new SelectTransaction(connection, ImapCapability.CondStore)) {
        return SelectExamineInternal(t, mailboxName, new ImapParenthesizedString("CONDSTORE"));
      }
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Examine(ImapMailbox mailbox)
    {
      return Examine(mailbox.Name);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Examine(string mailboxName)
    {
      using (var t = new ExamineTransaction(connection, null)) {
        return SelectExamineInternal(t, mailboxName, null);
      }
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult ExamineCondStore(ImapMailbox mailbox)
    {
      return ExamineCondStore(mailbox.Name);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult ExamineCondStore(string mailboxName)
    {
      using (var t = new ExamineTransaction(connection, ImapCapability.CondStore)) {
        return SelectExamineInternal(t, mailboxName, new ImapParenthesizedString("CONDSTORE"));
      }
    }

    private ImapTransactionResult SelectExamineInternal(SelectTransactionBase t, string mailboxName, ImapParenthesizedString selectParameters)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      var selectingMailbox = mailboxManager.GetExist(mailboxName);
      var selectingMailboxExists = (selectingMailbox != null);

      if (selectingMailboxExists) {
        if (selectedMailbox == selectingMailbox) {
          return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "mailbox has already been selected");
        }
        else if (selectingMailbox.Flags != null &&
            selectingMailbox.Flags.Has(ImapMailboxFlag.NoSelect)) {
          throw new ImapProtocolViolationException("mailbox is not selectable");
        }
      }
      else {
        selectingMailbox = new ImapMailbox(mailboxName);
      }

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    examine         = "EXAMINE" SP mailbox [select-params]
       *                      ;; modifies the original IMAP EXAMINE command
       *                      ;; to accept optional parameters
       *    select          = "SELECT" SP mailbox [select-params]
       *                      ;; modifies the original IMAP SELECT command to
       *                      ;; accept optional parameters
       *    select-params   = SP "(" select-param *(SP select-param) ")"
       *    select-param    = select-param-name [SP select-param-value]
       *                      ;; a parameter to SELECT may contain one or
       *                      ;; more atoms and/or strings and/or lists.
       *    select-param-name= tagged-ext-label
       *    select-param-value= tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions
       */
      t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

      // select-params
      if (selectParameters != null)
        t.Request.Arguments["select parameters"] = selectParameters;

      selectedMailbox = selectingMailbox; // set temporarily (to update size and status)

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
        if (selectingMailboxExists)
          selectedMailbox = selectingMailbox;
        else
          selectedMailbox = mailboxManager.Add(selectingMailbox);

        TransitStateTo(ImapSessionState.Selected);
      }
      else {
        selectedMailbox = null;
        if (state != ImapSessionState.NotConnected)
          TransitStateTo(ImapSessionState.Authenticated);
      }

      return t.Result;
    }

    /// <summary>sends CREATE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Create(string mailboxName, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      /*
       * 6.3.3. CREATE Command
       *       It is an error to attempt to create INBOX or a mailbox
       *       with a name that refers to an extant mailbox. 
       */
      if (ImapMailbox.IsNameInbox(mailboxName))
        throw new ImapProtocolViolationException("It is an error to attempt to create INBOX.");

      if (mailboxManager.GetExist(mailboxName) != null)
        throw new ImapProtocolViolationException(string.Format("It is an error to attempt to create a mailbox with a name that refers to an extent mailbox. (mailboxName: '{0}')", mailboxName));

      createdMailbox = null;

      using (var t = new CreateTransaction(connection)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    create          = "CREATE" SP mailbox
         *                      [create-params]
         *                      ;; Use of INBOX gives a NO error.
         *    create-params   = SP "(" create-param *( SP create-param) ")"
         *    create-param-name = tagged-ext-label
         *    create-param      = create-param-name [SP create-param-value]
         *    create-param-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         */
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // create-params
        //t.Request.Arguments["create parameters"] = new ImapParenthesizedString();

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          createdMailbox = mailboxManager.Add(new ImapMailbox(mailboxName));

        return t.Result;
      }
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Delete(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Delete(mailbox.Name);
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Delete(string mailboxName)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      /*
       * 6.3.4. DELETE Command
       *       It is an error to attempt to delete INBOX or a
       *       mailbox name that does not exist.
       */
      if (ImapMailbox.IsNameInbox(mailboxName))
        throw new ImapProtocolViolationException("It is an error to attempt to delete INBOX.");

      /*
       * 6.3.4. DELETE Command
       *       It is an error to attempt to
       *       delete a name that has inferior hierarchical names and also has
       *       the \Noselect mailbox name attribute (see the description of the
       *       LIST response for more details).
       * 
       *       It is permitted to delete a name that has inferior hierarchical
       *       names and does not have the \Noselect mailbox name attribute.  In
       *       this case, all messages in that mailbox are removed, and the name
       *       will acquire the \Noselect mailbox name attribute.
       */
      var deletingMailbox = mailboxManager.GetExist(mailboxName);

      if (deletingMailbox != null) {
        if ((deletingMailbox.Flags != null && deletingMailbox.Flags.Has(ImapMailboxFlag.NoSelect)) &&
            mailboxManager.ExistChildrenOf(deletingMailbox))
          throw new ImapProtocolViolationException("It is an error to attempt to delete a name that has inferior hierarchical names and also has the \\Noselect mailbox name attribute.");
      }

      using (var t = new DeleteTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          if (selectedMailbox != null && deletingMailbox == selectedMailbox) {
            selectedMailbox = null;
            TransitStateTo(ImapSessionState.Authenticated);
          }

          mailboxManager.Delete(mailboxName);
        }

        return t.Result;
      }
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(ImapMailbox existingMailbox, string newMailboxName)
    {
      ValidateMailboxRelationship(existingMailbox);

      ImapMailbox discard;

      return Rename(existingMailbox.Name, newMailboxName, out discard);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(ImapMailbox existingMailbox, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      ValidateMailboxRelationship(existingMailbox);

      return Rename(existingMailbox.Name, newMailboxName, out renamedMailbox);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(string existingMailboxName, string newMailboxName)
    {
      ImapMailbox discard;

      return Rename(existingMailboxName, newMailboxName, out discard);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(string existingMailboxName, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(existingMailboxName);
      RejectInvalidMailboxNameArgument(newMailboxName);

      /* 
       * 6.3.5. RENAME Command
       *       It is
       *       an error to attempt to rename from a mailbox name that does not
       *       exist or to a mailbox name that already exists.
       */
      if (mailboxManager.GetExist(newMailboxName) != null)
        throw new ImapProtocolViolationException(string.Format("It is an error to attempt to rename to a mailbox name that already exists. (newMailboxName: '{0}')", newMailboxName));

      renamedMailbox = null;

      using (var t = new RenameTransaction(connection)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    rename          = "RENAME" SP mailbox SP mailbox
         *                      [rename-params]
         *                      ;; Use of INBOX as a destination gives
         *                      ;; a NO error, unless rename-params
         *                      ;; is not empty.
         *    rename-params     = SP "(" rename-param *( SP rename-param) ")"
         *    rename-param      = rename-param-name [SP rename-param-value]
         *    rename-param-name = tagged-ext-label
         *    rename-param-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         */
        t.Request.Arguments["existing mailbox name"] = new ImapMailboxNameString(existingMailboxName);
        t.Request.Arguments["new mailbox name"] = new ImapMailboxNameString(newMailboxName);

        // rename-params
        //t.Request.Arguments["rename parameters"] = new ImapParenthesizedString();

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          renamedMailbox = mailboxManager.Rename(existingMailboxName, newMailboxName);

        return t.Result;
      }
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Subscribe(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Subscribe(mailbox.Name);
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Subscribe(string mailboxName)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      using (var t = new SubscribeTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Unsubscribe(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Unsubscribe(mailbox.Name);
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Unsubscribe(string mailboxName)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      using (var t = new UnsubscribeTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty mailbox name and an empty reference name.
    /// </remarks>
    public ImapTransactionResult ListRoot(out ImapMailboxList root)
    {
      return ListRoot(string.Empty, out root);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty mailbox name.
    /// </remarks>
    public ImapTransactionResult ListRoot(string referenceName, out ImapMailboxList root)
    {
      RejectNonAuthenticatedState();

      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      root = null;

      using (var t = new ListTransaction(connection)) {
        t.Request.Arguments["reference name"] = new ImapQuotedString(referenceName);
        t.Request.Arguments["mailbox name"] = new ImapQuotedString(string.Empty);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          hierarchyDelimiters[referenceName] = t.Result.Value[0].HierarchyDelimiter;

          root = t.Result.Value[0];
        }

        return t.Result;
      }
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapTransactionResult List(out ImapMailbox[] mailboxes)
    {
      return ListInternal(string.Empty, "*", out mailboxes);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return ListInternal(string.Empty, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox[] discard;

      return ListInternal(string.Empty, new ImapMailboxNameString(mailbox.Name), out discard);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return ListInternal(referenceName, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    private ImapTransactionResult ListInternal(string referenceName,
                                               ImapString mailboxName,
                                               out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      using (var t = new ListTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LSUB command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapTransactionResult Lsub(out ImapMailbox[] mailboxes)
    {
      return LsubInternal(string.Empty, "*", out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return LsubInternal(string.Empty, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox[] discard;

      return LsubInternal(string.Empty, new ImapMailboxNameString(mailbox.Name), out discard);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return LsubInternal(referenceName, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    private ImapTransactionResult LsubInternal(string referenceName,
                                               ImapString mailboxName,
                                               out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      using (var t = new LsubTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends Gimap XLIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends Gimap XLIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapTransactionResult XList(out ImapMailbox[] mailboxes)
    {
      return XListInternal(string.Empty, "*", out mailboxes);
    }

    /// <summary>sends Gimap LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult XList(string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return XListInternal(string.Empty, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    /// <summary>sends Gimap LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult XList(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox[] discard;

      return XListInternal(string.Empty, new ImapMailboxNameString(mailbox.Name), out discard);
    }

    /// <summary>sends Gimap LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult XList(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return XListInternal(referenceName, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    private ImapTransactionResult XListInternal(string referenceName,
                                               ImapString mailboxName,
                                               out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      using (var t = new XListTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends RLIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapTransactionResult RList(out ImapMailbox[] mailboxes)
    {
      return RListInternal(string.Empty, "*", out mailboxes);
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RList(string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return RListInternal(string.Empty, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RList(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return RListInternal(referenceName, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    private ImapTransactionResult RListInternal(string referenceName,
                                                ImapString mailboxName,
                                                out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      using (var t = new RListTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapTransactionResult RLsub(out ImapMailbox[] mailboxes)
    {
      return RLsubInternal(string.Empty, "*", out mailboxes);
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RLsub(string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return RLsubInternal(string.Empty, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RLsub(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");

      return RLsubInternal(referenceName, new ImapMailboxNameString(mailboxName), out mailboxes);
    }

    private ImapTransactionResult RLsubInternal(string referenceName,
                                                ImapString mailboxName,
                                                out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");

      using (var t = new RLsubTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

#region "extended LIST with no reference name"
#region "extended LIST with multiple mailbox name pattern"
    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(out ImapMailbox[] mailboxes,
                                              string mailboxNamePattern,
                                              params string[] mailboxNamePatterns)
    {
      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, null, null, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(ImapListSelectionOptions selectionOptions,
                                              out ImapMailbox[] mailboxes,
                                              string mailboxNamePattern,
                                              params string[] mailboxNamePatterns)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException("selectionOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, selectionOptions, null, out mailboxes);
    }

      /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(ImapListReturnOptions returnOptions,
                                              out ImapMailbox[] mailboxes,
                                              string mailboxNamePattern,
                                              params string[] mailboxNamePatterns)
    {
      if (returnOptions == null)
        throw new ArgumentNullException("returnOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, null, returnOptions, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(ImapListSelectionOptions selectionOptions,
                                              ImapListReturnOptions returnOptions,
                                              out ImapMailbox[] mailboxes,
                                              string mailboxNamePattern,
                                              params string[] mailboxNamePatterns)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException("selectionOptions");
      if (returnOptions == null)
        throw new ArgumentNullException("returnOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, selectionOptions, returnOptions, out mailboxes);
    }
#endregion

#region "extended LIST with single mailbox name pattern"
    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(string mailboxNamePattern,
                                              out ImapMailbox[] mailboxes)
    {
      return ListExtendedInternalNoRefName(mailboxNamePattern, null, null, null, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(string mailboxNamePattern,
                                              ImapListSelectionOptions selectionOptions,
                                              out ImapMailbox[] mailboxes)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException("selectionOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, selectionOptions, null, out mailboxes);
    }

      /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(string mailboxNamePattern,
                                              ImapListReturnOptions returnOptions,
                                              out ImapMailbox[] mailboxes)
    {
      if (returnOptions == null)
        throw new ArgumentNullException("returnOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, null, returnOptions, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(string mailboxNamePattern,
                                              ImapListSelectionOptions selectionOptions,
                                              ImapListReturnOptions returnOptions,
                                              out ImapMailbox[] mailboxes)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException("selectionOptions");
      if (returnOptions == null)
        throw new ArgumentNullException("returnOptions");

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, selectionOptions, returnOptions, out mailboxes);
    }
#endregion

    private ImapTransactionResult ListExtendedInternalNoRefName(string mailboxNamePattern,
                                                                string[] mailboxNamePatterns,
                                                                ImapListSelectionOptions selectionOptions,
                                                                ImapListReturnOptions returnOptions,
                                                                out ImapMailbox[] mailboxes)
    {
      if (mailboxNamePattern == null)
        throw new ArgumentNullException("mailboxNamePattern");

      if (mailboxNamePatterns == null || mailboxNamePatterns.Length == 0)
        return ListExtendedInternal(string.Empty, new[] {mailboxNamePattern}, selectionOptions, returnOptions, out mailboxes);
      else
        return ListExtendedInternal(string.Empty, mailboxNamePatterns.Prepend(mailboxNamePattern), selectionOptions, returnOptions, out mailboxes);
    }
#endregion

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult ListExtended(string referenceName,
                                              string[] mailboxNamePatterns,
                                              ImapListSelectionOptions selectionOptions,
                                              ImapListReturnOptions returnOptions,
                                              out ImapMailbox[] mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException("referenceName");
      if (mailboxNamePatterns == null)
        throw new ArgumentNullException("mailboxNamePatterns");
      else if (mailboxNamePatterns.Length == 0)
        throw new ArgumentException("must be non-empty array", "mailboxNamePatterns");
      if (selectionOptions == null)
        throw new ArgumentNullException("selectionOptions");
      if (returnOptions == null)
        throw new ArgumentNullException("returnOptions");

      return ListExtendedInternal(referenceName, mailboxNamePatterns, selectionOptions, returnOptions, out mailboxes);
    }

    private ImapTransactionResult ListExtendedInternal(string referenceName,
                                                       string[] mailboxNamePatterns,
                                                       ImapListSelectionOptions selectionOptions,
                                                       ImapListReturnOptions returnOptions,
                                                       out ImapMailbox[] mailboxes)
    {
      using (var t = new ListExtendedTransaction(connection)) {
        if (selectionOptions != null)
        t.Request.Arguments["selection options"] = selectionOptions;

        if (returnOptions != null)
          t.Request.Arguments["return options"] = returnOptions;

        var quotedMailboxNamePatterns = Array.ConvertAll(mailboxNamePatterns, delegate(string pattern) {
          return new ImapQuotedString(pattern);
        });

        if (quotedMailboxNamePatterns.Length == 1)
          ListLsubInternal(t, referenceName, quotedMailboxNamePatterns[0], out mailboxes);
        else
          ListLsubInternal(t, referenceName, new ImapParenthesizedString(quotedMailboxNamePatterns), out mailboxes);

        /*
         * draft-ietf-morg-status-in-list-01 - IMAP4 Extension for returning STATUS information in extended LIST
         * http://tools.ietf.org/html/draft-ietf-morg-status-in-list-01
         */
        if (returnOptions != null && returnOptions.RequiredCapabilities.Contains(ImapCapability.ListStatus)) {
          // XXX: converting STATUS response
          foreach (var response in t.Result.ReceivedResponses) {
            var data = response as ImapDataResponse;

            if (data == null || data.Type != ImapDataResponseType.Status)
              continue;

            string mailboxName;
            var statusAttr = ImapDataResponseConverter.FromStatus(data, out mailboxName);

            var statusMailbox = Array.Find(mailboxes, delegate(ImapMailbox mailbox) {
              return mailbox.Name == mailboxName;
            });

            if (statusMailbox != null)
              statusMailbox.UpdateStatus(statusAttr);
          }
        }

        return t.Result;
      }
    }

    private ImapTransactionResult ListLsubInternal(ListTransactionBase t,
                                                   string referenceName,
                                                   ImapString mailboxName,
                                                   out ImapMailbox[] mailboxes)
    {
      RejectNonAuthenticatedState();

      mailboxes = null;

      t.Request.Arguments["reference name"] = new ImapQuotedString(referenceName);
      t.Request.Arguments["mailbox name"] = mailboxName;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        mailboxes = Array.ConvertAll<ImapMailboxList, ImapMailbox>(t.Result.Value, mailboxManager.GetExistOrCreate);

      return t.Result;
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(ImapMailbox mailbox, ImapStatusDataItem statusDataItem)
    {
      ImapStatusAttributeList discard1;
      ImapMailbox discard2;

      ValidateMailboxRelationship(mailbox);

      return StatusInternal(mailbox.Name, statusDataItem, out discard1, out discard2);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(string mailboxName, ImapStatusDataItem statusDataItem)
    {
      ImapStatusAttributeList discard1;
      ImapMailbox discard2;

      return StatusInternal(mailboxName, statusDataItem, out discard1, out discard2);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(string mailboxName, ImapStatusDataItem statusDataItem, out ImapStatusAttributeList statusAttributes)
    {
      ImapMailbox discard;

      return StatusInternal(mailboxName, statusDataItem, out statusAttributes, out discard);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(string mailboxName, ImapStatusDataItem statusDataItem, out ImapMailbox statusMailbox)
    {
      ImapStatusAttributeList discard;

      return StatusInternal(mailboxName, statusDataItem, out discard, out statusMailbox);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    private ImapTransactionResult StatusInternal(string mailboxName, ImapStatusDataItem statusDataItem, out ImapStatusAttributeList statusAttributes, out ImapMailbox statusMailbox)
    {
      RejectNonAuthenticatedState();

      RejectInvalidMailboxNameArgument(mailboxName);

      if (statusDataItem == null)
        throw new ArgumentNullException("statusDataItem");

      /*
       * 6.3.10. STATUS Command
       * 
       *            Note: The STATUS command is intended to access the
       *            status of mailboxes other than the currently selected
       *            mailbox.  Because the STATUS command can cause the
       *            mailbox to be opened internally, and because this
       *            information is available by other means on the selected
       *            mailbox, the STATUS command SHOULD NOT be used on the
       *            currently selected mailbox.
       */
      var existStatusMailbox = mailboxManager.GetExist(mailboxName);

      if (existStatusMailbox != null) {
        if (existStatusMailbox == selectedMailbox)
          throw new ImapProtocolViolationException("the STATUS command SHOULD NOT be used on the currently selected mailbox");

        if (existStatusMailbox.Flags != null && existStatusMailbox.Flags.Has(ImapMailboxFlag.NoSelect))
          throw new ImapProtocolViolationException("the STATUS command SHOULD NOT be used on the mailbox that has \\Noselect attribute");
      }

      statusAttributes = null;
      statusMailbox = null;

      using (var t = new StatusTransaction(connection)) {
        // mailbox name
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // status data item names
        t.Request.Arguments["status data item names"] = statusDataItem;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          statusMailbox = mailboxManager.GetExistOrCreate(mailboxName);

          statusAttributes = t.Result.Value;
          statusMailbox.UpdateStatus(statusAttributes);
        }

        return t.Result;
      }
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Append(ImapMessage message, ImapMailbox mailbox)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ValidateMailboxRelationship(mailbox);

      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(new[] {message}, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(messages, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, ImapMailbox mailbox, out ImapAppendedUidSet appendedUids)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return AppendInternal(new[] {message}, mailbox.Name, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, ImapMailbox mailbox, out ImapAppendedUidSet appendedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return AppendInternal(messages, mailbox.Name, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(new[] {message}, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName)
    {
      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(messages, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapAppendedUidSet appendedUids)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ImapMailbox discard;

      return AppendInternal(new[] {message}, mailboxName, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
    {
      ImapMailbox discard;

      return AppendInternal(messages, mailboxName, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapMailbox createdMailbox)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ImapAppendedUidSet discard;

      return AppendInternal(new[] {message}, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapAppendedUidSet discard;

      return AppendInternal(messages, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      return AppendInternal(new[] {message}, mailboxName, true, out appendedUids, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      return AppendInternal(messages, mailboxName, true, out appendedUids, out createdMailbox);
    }

    private ImapTransactionResult AppendInternal(IEnumerable<ImapMessage> messages, string mailboxName, bool tryCreate, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState();

      if (messages == null)
        throw new ArgumentNullException("messages");

      RejectInvalidMailboxNameArgument(mailboxName);

      appendedUids = null;
      createdMailbox = null;

      // append message
      var messagesToUpload = new List<ImapString>();
      var messageCount = 0;

      foreach (var message in messages) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         * 
         *    append          = "APPEND" SP mailbox 1*append-message
         *                      ;; only a single append-message may appear
         *                      ;; if MULTIAPPEND [MULTIAPPEND] capability
         *                      ;; is not present
         *    append-message  = append-opts SP append-data
         *    append-ext      = append-ext-name SP append-ext-value
         *                      ;; This non-terminal define extensions to
         *                      ;; to message metadata.
         *    append-ext-name = tagged-ext-label
         *    append-ext-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    append-data     = literal / literal8 / append-data-ext
         *    append-data-ext = tagged-ext
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions,
         *                      ;; i.e., a mandatory label followed
         *                      ;; by parameters.
         *    append-opts     = [SP flag-list] [SP date-time] *(SP append-ext)
         *                      ;; message metadata
         */

        // flag-list
        if (message.Flags != null)
          messagesToUpload.Add(message.Flags.GetNonApplicableFlagsRemoved().ToParenthesizedString());

        // date-time
        if (message.InternalDate.HasValue)
          messagesToUpload.Add(ImapDateTimeFormat.ToDateTimeString(message.InternalDate.Value));

        // append-data
        messagesToUpload.Add(new ImapLiteralStream(new MemoryStream(message.Body, false)));

        messageCount++;
      }

      ImapTransactionResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new AppendTransaction(connection, 1 < messageCount)) {
          // mailbox name
          t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

          // messages to upload
          t.Request.Arguments["messages to upload"] = new ImapStringList(messagesToUpload.ToArray());

          if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
            appendedUids = t.Result.Value;
            return t.Result;
          }
          else if (!tryCreate) {
            return t.Result;
          }

          failedResult = t.Result;

          // 6.3.11. APPEND Command
          //       If the destination mailbox does not exist, a server MUST return an
          //       error, and MUST NOT automatically create the mailbox.  Unless it
          //       is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the APPEND
          //       if the CREATE is successful.
          respTryCreate = (t.Result.GetResponseCode(ImapResponseCode.TryCreate) is ImapTaggedStatusResponse);
        }

        // try create
        if (i == 0 && respTryCreate)
          if (Create(mailboxName, out createdMailbox).Failed)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>begins to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public IAsyncResult BeginAppend(Stream messageBodyStream, DateTimeOffset? internalDate, ImapMessageFlagList flags, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return BeginAppend(messageBodyStream, internalDate, flags, mailbox.Name);
    }

    /// <summary>begins to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public IAsyncResult BeginAppend(Stream messageBodyStream, DateTimeOffset? internalDate, ImapMessageFlagList flags, string mailboxName)
    {
      RejectNonAuthenticatedState();
      RejectTransactionProceeding();

      if (messageBodyStream == null)
        throw new ArgumentNullException("messageBodyStream");

      RejectInvalidMailboxNameArgument(mailboxName);

      // append message
      var messagesToUpload = new List<ImapString>();

      // flag-list
      if (flags != null)
        messagesToUpload.Add(flags.GetNonApplicableFlagsRemoved().ToParenthesizedString());

      // date-time
      if (internalDate.HasValue)
        messagesToUpload.Add(ImapDateTimeFormat.ToDateTimeString(internalDate.Value));

      // append-data
      messagesToUpload.Add(new ImapLiteralStream(messageBodyStream, ImapLiteralSyncMode.Synchronous));

      AppendTransaction t = null;

      try {
        t = new AppendTransaction(connection, false);

        // mailbox name
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // messages to upload
        t.Request.Arguments["messages to upload"] = new ImapStringList(messagesToUpload.ToArray());

        var asyncResult = BeginProcessTransaction(t, handlesIncapableAsException);

        // wait for started (or completed)
        for (;;) {
          if (asyncResult.IsCompleted)
            break;
          else if (IsTransactionProceeding)
            break;
          else
            System.Threading.Thread.Sleep(10);
        }

        return asyncResult;
      }
      catch {
        if (t != null) {
          t.Dispose();
          t = null;
        }

        throw;
      }
    }

    /// <summary>ends to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult EndAppend(IAsyncResult asyncResult)
    {
      ImapAppendedUidSet discard;

      return EndAppend(asyncResult, out discard);
    }

    /// <summary>ends to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult EndAppend(IAsyncResult asyncResult, out ImapAppendedUidSet appendedUid)
    {
      appendedUid = null;

      var appendAsyncResult = asyncResult as TransactionAsyncResult;

      if (appendAsyncResult == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      using (var t = appendAsyncResult.Transaction as AppendTransaction) {
        if (EndProcessTransaction(appendAsyncResult).Succeeded)
          appendedUid = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends ENABLE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Enable(params string[] capabilityNames)
    {
      ImapCapabilityList discard;

      return Enable(out discard, capabilityNames);
    }

    /// <summary>sends ENABLE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Enable(out ImapCapabilityList enabledCapabilities, params string[] capabilityNames)
    {
      RejectNonAuthenticatedState();

      if (capabilityNames == null || capabilityNames.Length == 0)
        throw new ArgumentException("at least one capability name is required", "capabilityNames");

      enabledCapabilities = null;

      using (var t = new EnableTransaction(connection)) {
        // capability names
        t.Request.Arguments["capability names"] = new ImapStringList(capabilityNames);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          enabledCapabilities = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * transaction methods : selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Check()
    {
      RejectNonSelectedState();

      using (var t = new CheckTransaction(connection)) {
        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Close()
    {
      RejectNonAuthenticatedState();

      if (state != ImapSessionState.Selected)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "mailbox has already closed or has not selected yet");

      using (var t = new CloseTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge()
    {
      long[] discard;

      return Expunge(out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge(out long[] expungedMessages)
    {
      using (var t = new ExpungeTransaction(connection, false)) {
        return ExpungeInternal(t, null, out expungedMessages);
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet)
    {
      long[] discard;

      return UidExpunge(uidSet, out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      if (uidSet == null)
        throw new ArgumentNullException("uidSet");
      if (!uidSet.IsUidSet)
        throw new ArgumentException("sequence set is not UID", "uidSet");
      if (uidSet.IsEmpty)
        throw new ArgumentException("uid set is empty", "uidSet");

      using (var t = new ExpungeTransaction(connection, true)) {
        return ExpungeInternal(t, uidSet, out expungedMessages);
      }
    }

    private ImapTransactionResult ExpungeInternal(ExpungeTransaction t, ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      RejectNonSelectedState();

      expungedMessages = null;

      // sequence set for UID EXPUNGE
      if (uidSet != null)
        t.Request.Arguments["sequence set"] = uidSet.ToString();

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        expungedMessages = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Search(searchingCriteria, null, out matched);
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESearch(searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, false, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSearch(searchingCriteria, null, out matched);
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESearch(searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, true, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Sort(sortingCriteria, searchingCriteria, null, out matched);
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESort(sortingCriteria, searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, ImapCapability.ESort)) {
        return ESearchESortInternal(t, false, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSort(sortingCriteria, searchingCriteria, (Encoding)null, out matched);
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESort(sortingCriteria, searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, ImapCapability.ESort)) {
        return ESearchESortInternal(t, true, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    private ImapTransactionResult ESearchESortInternal(SearchTransactionBase t, bool uid, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      if (resultOptions == null)
        throw new ArgumentNullException("resultOptions");

      /*
       * RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
       * http://tools.ietf.org/html/rfc5182
       */
      var saveResult = resultOptions.ContainsOneOf(ImapSearchResultOptions.Save);

      // ESEARCH or ESORT capability
      if (t is SearchTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESearch);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESort);
      }
      else if (t is SortTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESort);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESearch);
      }

      SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);

      if (saveResult) {
        var saved = t.Result.Succeeded && (t.Result.GetResponseCode(ImapResponseCode.NotSaved) == null);

        if (saved)
          matched = ImapMatchedSequenceSet.CreateSavedResult(matched == null
                                                             ? matched = new ImapMatchedSequenceSet(ImapSequenceSet.CreateSet(uid, new long[] {}))
                                                             : matched);
      }

      return t.Result;
    }

    private ImapTransactionResult SearchSortInternal(SearchTransactionBase t, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      if (t is SortTransaction && sortingCriteria == null)
        throw new ArgumentNullException("sortingCriteria");
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");

      RejectNonSelectedState();

      matched = null;

      /*
       * 6.4.4. SEARCH Command
       *    US-ASCII MUST be supported; other [CHARSET]s MAY be supported.
       * 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       *    The US-ASCII and [UTF-8] charsets MUST be implemented.
       *    All other charsets are optional.
       */
      if (charset == null) {
        if (t is SortTransaction)
          charset = Encoding.UTF8;
        else
          charset = Encoding.ASCII;
      }

      var charsetSpecified = searchingCriteria.SetCharset(charset);

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    search          = "SEARCH" [search-return-opts]
       *                      SP search-program
       *    search-correlator  = SP "(" "TAG" SP tag-string ")"
       *    search-program     = ["CHARSET" SP charset SP]
       *                         search-key *(SP search-key)
       *                         ;; CHARSET argument to SEARCH MUST be
       *                         ;; registered with IANA.
       *    search-return-data = search-modifier-name SP search-return-value
       *                         ;; Note that not every SEARCH return option
       *                         ;; is required to have the corresponding
       *                         ;; ESEARCH return data.
       *    search-return-opts = SP "RETURN" SP "(" [search-return-opt
       *                         *(SP search-return-opt)] ")"
       *    search-return-opt = search-modifier-name [SP search-mod-params]
       *    search-return-value = tagged-ext-val
       *                         ;; Data for the returned search option.
       *                         ;; A single "nz-number"/"number" value
       *                         ;; can be returned as an atom (i.e., without
       *                         ;; quoting).  A sequence-set can be returned
       *                         ;; as an atom as well.
       *    search-modifier-name = tagged-ext-label
       *    search-mod-params = tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions.
       */

      /* 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       * The charset argument is mandatory (unlike SEARCH)
       */
      // charset specification
      if (t is SortTransaction || charsetSpecified)
        t.Request.Arguments["charset specification"] = charset.WebName;

      // search-return-opt
      if (resultOptions != null)
        t.Request.Arguments["result specifier"] = resultOptions;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

        // sort criteria
      if (sortingCriteria != null)
        t.Request.Arguments["sort criteria"] = sortingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        matched = t.Result.Value;

      // TODO: BADCHARSET
      // 6.4.4. SEARCH Command
      //       If the server does not support the specified [CHARSET], it MUST
      //       return a tagged NO response (not a BAD).  This response SHOULD
      //       contain the BADCHARSET response code, which MAY list the
      //       [CHARSET]s supported by the server.

      return t.Result;
    }

    internal ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, string charset, out ImapMatchedSequenceSet matched)
    {
      if (sortingCriteria == null)
        throw new ArgumentNullException("sortingCriteria");
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");

      RejectNonSelectedState();

      matched = null;

      if (charset == null)
        charset = Encoding.UTF8.WebName;

      using (var t = new SortTransaction(connection, true, null)) {
        // charset specification
        t.Request.Arguments["charset specification"] = charset;

        // searching criteria
        t.Request.Arguments["searching criteria"] = searchingCriteria;

        // sort criteria
        t.Request.Arguments["sort criteria"] = sortingCriteria;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          matched = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    private ImapTransactionResult ThreadInternal(ThreadTransaction t, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");
      if (charset == null)
        throw new ArgumentNullException("charset");

      /*
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.THREAD. THREAD Command
       *   The US-ASCII and [UTF-8] charsets MUST be implemented.
       */
      if (charset == null)
        charset = Encoding.UTF8;

      searchingCriteria.SetCharset(charset);

      return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset.WebName, out threadList);
    }

    internal ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, string charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset ?? Encoding.UTF8.WebName, out threadList);
      }
    }

    private ImapTransactionResult ThreadInternal(ThreadTransaction t, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, string charset, out ImapThreadList threadList)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException("threadingAlgorithm");

      RejectNonSelectedState();

      threadList = null;

      // threading algorithm
      t.Request.Arguments["threading algorithm"] = threadingAlgorithm;

      // charset specification
      t.Request.Arguments["charset specification"] = charset;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        threadList = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Fetch(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, out ImapMessage[] messages)
    {
      return FetchInternal(sequenceOrUidSet, messageDataItems, null, null, out messages);
    }

    /// <summary>sends FETCH command with CHANGEDSINCE FETCH modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult FetchChangedSince(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, ulong modSeq, out ImapMessage[] messages)
    {
      var fetchModifiers = new ImapParenthesizedString("CHANGEDSINCE", modSeq.ToString());

      return FetchInternal(sequenceOrUidSet, messageDataItems, fetchModifiers, ImapCapability.CondStore, out messages);
    }

    private ImapTransactionResult FetchInternal(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, ImapParenthesizedString fetchModifiers, ImapCapability fetchModifiersCapabilityRequirement, out ImapMessage[] messages)
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      if (messageDataItems == null)
        throw new ArgumentNullException("messageDataItems");

      messages = null;

      using (var t = new FetchTransaction(connection, sequenceOrUidSet.IsUidSet, fetchModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    fetch           = "FETCH" SP sequence-set SP ("ALL" / "FULL" /
         *                      "FAST" / fetch-att /
         *                      "(" fetch-att *(SP fetch-att) ")")
         *                      [fetch-modifiers]
         *                      ;; modifies the original IMAP4 FETCH command to
         *                      ;; accept optional modifiers
         *    fetch-modifiers = SP "(" fetch-modifier *(SP fetch-modifier) ")"
         *    fetch-modifier  = fetch-modifier-name [ SP fetch-modif-params ]
         *    fetch-modif-params  = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    fetch-modifier-name = tagged-ext-label
         */
        // sequence set
        t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

        // message data item names or macro
        t.Request.Arguments["message data item names or macro"] = messageDataItems;

        // fetch-modifiers
        if (fetchModifiers != null)
          t.Request.Arguments["fetch modifiers"] = fetchModifiers;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          if (t.Result.Value == null) {
            messages = null;
          }
          else {
            messages = Array.ConvertAll(t.Result.Value, delegate(ImapMailboxMessage message) {
              message.SetBaseUrl(SelectedMailbox.Url);
              return (ImapMessage)message;
            });
          }
        }

        return t.Result;
      }
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem)
    {
      ImapMessage[] discard;

      return Store(sequenceOrUidSet, storeDataItem, out discard);
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, out ImapMessage[] messages)
    {
      return StoreInternal(sequenceOrUidSet, storeDataItem, null, null, out messages);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq)
    {
      ImapMessage[] discard;
      ImapSequenceSet discard2;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out discard2);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessage[] messages)
    {
      ImapSequenceSet discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out messages, out discard);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapSequenceSet failedMessageSet)
    {
      ImapMessage[] discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out failedMessageSet);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessage[] messages, out ImapSequenceSet failedMessageSet)
    {
      failedMessageSet = null;
      messages = null;

      var storeModifiers = new ImapParenthesizedString("UNCHANGEDSINCE", modSeq.ToString());

      var ret = StoreInternal(sequenceOrUidSet, storeDataItem, storeModifiers, ImapCapability.CondStore, out messages);

      var modified = lastTransactionResult.GetResponseCode(ImapResponseCode.Modified);

      if (modified is ImapTaggedStatusResponse)
        failedMessageSet = ImapResponseTextConverter.FromModified(modified.ResponseText, sequenceOrUidSet.IsUidSet);

      return ret;
    }

    private ImapTransactionResult StoreInternal(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ImapParenthesizedString storeModifiers, ImapCapability storeModifiersCapabilityRequirement, out ImapMessage[] messages)
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      if (storeDataItem == null)
        throw new ArgumentNullException("storeDataItem");

      messages = null;

      using (var t = new StoreTransaction(connection, sequenceOrUidSet.IsUidSet, storeModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    store           = "STORE" SP sequence-set [store-modifiers]
         *                      SP store-att-flags
         *                      ;; extend [IMAP4] STORE command syntax
         *                      ;; to allow for optional store-modifiers
         *    store-modifiers =  SP "(" store-modifier *(SP store-modifier)
         *                        ")"
         *    store-modifier  = store-modifier-name [SP store-modif-params]
         *    store-modif-params = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    store-modifier-name = tagged-ext-label
         */
        // message set
        t.Request.Arguments["message set"] = sequenceOrUidSet.ToString();

        // store-modifiers
        if (storeModifiers != null)
          t.Request.Arguments["store modifiers"] = storeModifiers;

        // message data item name
        t.Request.Arguments["message data item name"] = storeDataItem.ItemName;

        // value for message data item
        t.Request.Arguments["value for message data item"] = storeDataItem;

        var ret = ProcessTransaction(t, handlesIncapableAsException);

        if (t.Result.Value == null) {
          messages = null;
        }
        else {
          messages = Array.ConvertAll(t.Result.Value, delegate(ImapMailboxMessage message) {
            message.SetBaseUrl(SelectedMailbox.Url);
            return (ImapMessage)message;
          });
        }

        return ret;
      }
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox, out ImapCopiedUidSet copiedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName)
    {
      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
    {
      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapCopiedUidSet discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      return CopyInternal(sequenceOrUidSet, mailboxName, true, out copiedUids, out createdMailbox);
    }

    private ImapTransactionResult CopyInternal(ImapSequenceSet sequenceOrUidSet, string mailboxName, bool tryCreate, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      RejectInvalidMailboxNameArgument(mailboxName);

      copiedUids = null;
      createdMailbox = null;

      ImapTransactionResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new CopyTransaction(connection, sequenceOrUidSet.IsUidSet)) {
          // sequence set
          t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

          // mailbox name
          t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

          if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
            copiedUids = t.Result.Value;
            return t.Result;
          }
          else if (!tryCreate) {
            return t.Result;
          }

          failedResult = t.Result;

          // 6.4.7. COPY Command
          //       If the destination mailbox does not exist, a server SHOULD return
          //       an error.  It SHOULD NOT automatically create the mailbox.  Unless
          //       it is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the COPY if
          //       the CREATE is successful.
          respTryCreate = (t.Result.GetResponseCode(ImapResponseCode.TryCreate) is ImapTaggedStatusResponse);
        }

        // try create
        if (i == 0 && respTryCreate)
          if (Create(mailboxName, out createdMailbox).Failed)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>sends UNSELECT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Unselect()
    {
      RejectNonSelectedState();

      using (var t = new UnselectTransaction(connection)) {
        var r = ProcessTransaction(t, handlesIncapableAsException);

        if (r.Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return r;
      }
    }

    /*
     * transaction methods : authenticated/selected state
     */

    public ImapTransactionResult Idle(TimeSpan timeout)
    {
      return IdleInternal((int)timeout.TotalMilliseconds, null, null);
    }

    public ImapTransactionResult Idle(int millisecondsTimeout)
    {
      return IdleInternal(millisecondsTimeout, null, null);
    }

    public ImapTransactionResult Idle(object state, ImapKeepIdleCallback keepIdleCallback)
    {
      return Idle(Timeout.Infinite, state, keepIdleCallback);
    }

    public ImapTransactionResult Idle(TimeSpan idleTimeout, object state, ImapKeepIdleCallback keepIdleCallback)
    {
      return Idle((int)idleTimeout.TotalMilliseconds, state, keepIdleCallback);
    }

    public ImapTransactionResult Idle(int idleMillisecondsTimeout, object state, ImapKeepIdleCallback keepIdleCallback)
    {
      if (keepIdleCallback == null)
        throw new ArgumentNullException("keepIdleCallback");

      return IdleInternal(idleMillisecondsTimeout, state, keepIdleCallback);
    }

    private ImapTransactionResult IdleInternal(int idleMillisecondsTimeout, object keepIdleState, ImapKeepIdleCallback keepIdleCallback)
    {
      if (idleMillisecondsTimeout < -1)
        throw new ArgumentOutOfRangeException("idleMillisecondsTimeout", "must be greater than or equals to -1");

      var idleAsyncResult = BeginIdle(keepIdleState, keepIdleCallback);

      idleAsyncResult.AsyncWaitHandle.WaitOne(idleMillisecondsTimeout);

      return EndIdle(idleAsyncResult);
    }

    /// <summary>sends IDLE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public IAsyncResult BeginIdle()
    {
      return BeginIdle(null, null);
    }

    private IAsyncResult BeginIdle(object keepIdleState, ImapKeepIdleCallback keepIdleCallback)
    {
      IdleTransaction t = null;

      try {
        RejectNonAuthenticatedState();
        RejectIdling();

        t = new IdleTransaction(connection, keepIdleState, keepIdleCallback);

        var asyncResult = BeginProcessTransaction(t, handlesIncapableAsException);

        // wait for IDLE started or respond tagged error response
        t.WaitForIdleStateChanged();

        return asyncResult;
      }
      catch {
        if (t != null) {
          t.Dispose();
          t = null;
        }

        throw;
      }
    }

    /// <summary>sends DONE for IDLE command continuation request responses</summary>
    /// <remarks>valid in idling</remarks>
    public ImapTransactionResult EndIdle(IAsyncResult asyncResult)
    {
      var idleAsyncResult = asyncResult as TransactionAsyncResult;

      if (idleAsyncResult == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      using (var t = idleAsyncResult.Transaction as IdleTransaction) {
        // send 'DONE' if idling
        t.Done();

        return EndProcessTransaction(idleAsyncResult);
      }
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Namespace()
    {
      ImapNamespace discard;

      return Namespace(out discard);
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Namespace(out ImapNamespace namespaces)
    {
      RejectNonAuthenticatedState();

      namespaces = null;

      using (var t = new NamespaceTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.namespaces = t.Result.Value;
          namespaces = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Comparator()
    {
      ImapCollationAlgorithm discard;

      return Comparator(out discard);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Comparator(out ImapCollationAlgorithm activeComparator)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with no arguments, it results in a
       *    COMPARATOR response indicating the currently active comparator.
       */
      activeComparator = null;

      using (var t = new ComparatorTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.activeComparator = t.Result.Value.Item1;

          activeComparator = t.Result.Value.Item1;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Comparator(ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      ImapCollationAlgorithm discard1;
      ImapCollationAlgorithm[] discard2;

      return Comparator(out discard1, out discard2, comparator, comparators);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Comparator(out ImapCollationAlgorithm[] matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      ImapCollationAlgorithm discard;

      return Comparator(out discard, out matchingComparators, comparator, comparators);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Comparator(out ImapCollationAlgorithm activeComparator, out ImapCollationAlgorithm[] matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      RejectNonAuthenticatedState();

      if (comparator == null)
        throw new ArgumentNullException("comparator");

      var comparatorArguments = new List<ImapCollationAlgorithm>();

      comparatorArguments.Add(comparator);

      if (comparators != null) {
        for (var i = 0; i < comparators.Length; i++) {
          if (comparators[i] == null)
            throw new ArgumentNullException(string.Format("comparators[{0}]", i));

          comparatorArguments.Add(comparators[i]);
        }
      }

      activeComparator = null;
      matchingComparators = null;

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with one or more comparator arguments, it changes the
       *    active comparator as directed.
       */
      using (var t = new ComparatorTransaction(connection)) {
        t.Request.Arguments["comparator order arguments"] = new ImapStringList(comparatorArguments.ConvertAll(delegate(ImapCollationAlgorithm comp) {
          return new ImapQuotedString(comp.ToString());
        }).ToArray());

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.activeComparator = t.Result.Value.Item1;

          activeComparator = t.Result.Value.Item1;
          matchingComparators = t.Result.Value.Item2 ?? new ImapCollationAlgorithm[0];
        }

        return t.Result;
      }
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, params ImapString[] arguments)
    {
      ImapDataResponse[] discard;

      return GenericCommand(command, out discard, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, out ImapDataResponse[] dataResponses, params ImapString[] arguments)
    {
      RejectNonConnectedState();

      if (command == null)
        throw new ArgumentNullException("command");
      else if (command.Length == 0)
        throw new ArgumentException("command must be a non-empty string", "command");

      dataResponses = null;

      using (var t = new GenericCommandTransaction(connection, command)) {
        if (arguments != null && 0 < arguments.Length)
          t.Request.Arguments["arguments"] = new ImapStringList(arguments);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          dataResponses = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * methods for processing transaction
     */

    private class TransactionAsyncResult : IAsyncResult {
      object IAsyncResult.AsyncState {
        get { return processTransactionAsyncResult.AsyncState; }
      }

      System.Threading.WaitHandle IAsyncResult.AsyncWaitHandle {
        get { return processTransactionAsyncResult.AsyncWaitHandle; }
      }

      bool IAsyncResult.CompletedSynchronously {
        get { return processTransactionAsyncResult.CompletedSynchronously; }
      }

      bool IAsyncResult.IsCompleted {
        get { return processTransactionAsyncResult.IsCompleted; }
      }

      public IImapTransaction Transaction { get { return transaction; } }
      public IAsyncResult ProcessTransactionAsyncResult { get { return processTransactionAsyncResult; } }

      public TransactionAsyncResult(IImapTransaction transaction, IAsyncResult processTransactionAsyncResult)
      {
        this.transaction = transaction;
        this.processTransactionAsyncResult = processTransactionAsyncResult;
      }

      private /*readonly*/ IImapTransaction transaction;
      private /*readonly*/ IAsyncResult processTransactionAsyncResult;
    }

    public ImapTransactionResult ProcessTransaction(IImapTransaction t)
    {
      return ProcessTransaction(t, false);
    }

    public ImapTransactionResult ProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      if (transactionTimeout == Timeout.Infinite) {
        // no timeout
        PreProcessTransaction(t, exceptionIfIncapable);

        ProcessTransactionInternal(t);

        return PostProcessTransaction(t);
      }
      else {
        var async = BeginProcessTransaction(t, exceptionIfIncapable);

        if (async.AsyncWaitHandle.WaitOne(transactionTimeout)) {
          return EndProcessTransaction(async);
        }
        else {
          CloseConnection();
          throw new TimeoutException(string.Format("transaction timeout ({0})", t.GetType().FullName));
        }
      }
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t)
    {
      return BeginProcessTransaction(t, false);
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      return BeginProcessTransaction(t, exceptionIfIncapable, null, null);
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t, bool exceptionIfIncapable, AsyncCallback callback, object state)
    {
      PreProcessTransaction(t, exceptionIfIncapable);

      var processTransaction = new ProcessTransactionDelegate(ProcessTransactionInternal);

      return new TransactionAsyncResult(t, processTransaction.BeginInvoke(t, callback, state));
    }

    public ImapTransactionResult EndProcessTransaction(IAsyncResult asyncResult)
    {
      var ar = asyncResult as TransactionAsyncResult;

      if (ar == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      var ptar = ar.ProcessTransactionAsyncResult as System.Runtime.Remoting.Messaging.AsyncResult;

      if (ptar.EndInvokeCalled)
        throw new InvalidOperationException("EndProcessTransaction already called");

      var processTransaction = ptar.AsyncDelegate as ProcessTransactionDelegate;

      return PostProcessTransaction(processTransaction.EndInvoke(ptar));
    }

    private delegate IImapTransaction ProcessTransactionDelegate(IImapTransaction t);

    private IImapTransaction ProcessTransactionInternal(IImapTransaction t)
    {
      try {
        transactionProceedingSemaphore.WaitOne();

        Trace.LogRequest(t);

        t.Start();

        if (t is IdleTransaction)
          TraceInfo("idling");

        for (;;) {
          if (t.State == ImapTransactionState.Finished) {
            // transaction is finished
            if (t is IdleTransaction)
              TraceInfo("done");

            Trace.LogResponse(t);

            return t;
          }

          t.Process();
        }
      }
      finally {
        transactionProceedingSemaphore.Release();
      }
    }

    private void PreProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      // TODO: '5.5. Multiple Commands in Progress'
      //   The following are examples of valid non-waiting command sequences:
      //      FETCH + STORE + SEARCH + CHECK
      //      STORE + COPY + EXPUNGE
      if (IsTransactionProceeding)
        throw new InvalidOperationException("another transaction proceesing");

      RejectIdling();

      // check capability
      if (!exceptionIfIncapable)
        return;

      if (!serverCapabilities.Has(Imap4.ImapCapability.Imap4Rev1) && !(t is CapabilityTransaction))
        throw new ImapIncapableException(Imap4.ImapCapability.Imap4Rev1);

      if (t is IImapExtension)
        CheckServerCapability(t as IImapExtension);
      else if (t is IImapMultipleExtension)
        CheckServerCapability(t as IImapMultipleExtension);

      foreach (var arg in t.Request.Arguments.Values) {
        if (arg is IImapExtension)
          CheckServerCapability(arg as IImapExtension);
        else if (arg is IImapMultipleExtension)
          CheckServerCapability(arg as IImapMultipleExtension);
      }
    }

    private ImapResponseProcessorBase[] responseProcessors;

    private void InitializeResponseProcessors()
    {
      responseProcessors = new ImapResponseProcessorBase[] {
        new ImapCapabilityResponseProcessor(this),
        new ImapUpdatedStatusResponseProcessor(this),
        new ImapReferralResponseProcessor(this),
      };
    }

    private ImapTransactionResult PostProcessTransaction(IImapTransaction t)
    {
      lastTransactionResult = t.Result;

      switch (t.Result.Code) {
        /*
         * disconnect without processing responses
         */
        case ImapTransactionResultCode.InternalError:
          CloseConnection();
          throw new ImapException(t.Result.Description, t.Result.Exception);

        case ImapTransactionResultCode.SocketTimeout:
          CloseConnection();
          throw new TimeoutException(string.Format("socket timeout in {0}", t.GetType().FullName), t.Result.Exception);
       }

      lastUpdatedStatus = new ImapUpdatedStatus();

      var unprocessedResponses = new List<ImapResponse>(t.Result.ReceivedResponses);

      foreach (var processor in responseProcessors) {
        foreach (var resp in unprocessedResponses.ToArray()) {
          if (processor.Process(resp, t))
            unprocessedResponses.Remove(resp);
        }
      }

      if (t.Result.Code == ImapTransactionResultCode.Bye) {
        TransitStateTo(ImapSessionState.NotAuthenticated);
        CloseConnection();
      }

      return t.Result;
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(ImapCapabilityList newCapabilities)
    {
      serverCapabilities = new ImapCapabilityList(true, newCapabilities);

      connection.SynchronizeLiterals = !serverCapabilities.Has(ImapCapability.LiteralNonSync);
    }

    private void TransitStateTo(ImapSessionState newState)
    {
      if (newState == state)
        return;

      switch (newState) {
        case ImapSessionState.NotConnected:
          authority = null;
          selectedMailbox = null;
          if (mailboxManager != null) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
          }
          TraceInfo("now in non-connected state");
          break;

        case ImapSessionState.NotAuthenticated:
          UpdateAuthority(null, null);
          if (state == ImapSessionState.Selected || state == ImapSessionState.Authenticated) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
            TraceInfo("logged out");
          }
          TraceInfo("now in non-authenticated state");
          break;

        case ImapSessionState.Authenticated:
          TraceInfo("now in authenticated state");
          if (state == ImapSessionState.NotAuthenticated || state == ImapSessionState.NotConnected) {
            mailboxManager = new ImapMailboxManager(this);
            TraceInfo("logged in");
          }
          break;

        case ImapSessionState.Selected:
          TraceInfo("now in selected state (selected '{0}')", selectedMailbox.Name);
          break;
      }

      state = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        throw new ImapProtocolViolationException("session has been disconnected");
    }

    private void RejectNonAuthenticatedState()
    {
      RejectNonConnectedState();

      if (state == ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("session is not in authenticated state");
    }

    private void RejectNonSelectedState()
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.Selected)
        throw new ImapProtocolViolationException("session is not in selected state");
    }

    private void RejectIdling()
    {
      if (IsIdling)
        throw new ImapProtocolViolationException("session is idling now");
    }

    private void ValidateMailboxRelationship(ImapMailbox mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      if (mailbox.Session != this)
        throw new ArgumentException("mailbox is not attached to current session", "mailbox");
    }

    private void CheckServerCapability(IImapExtension extension)
    {
      CheckServerCapability(extension.RequiredCapability);
    }

    private void CheckServerCapability(IImapMultipleExtension extensions)
    {
      foreach (var capability in extensions.RequiredCapabilities) {
        CheckServerCapability(capability);
      }
    }

    private void CheckServerCapability(ImapCapability capability)
    {
      if (capability != null && !serverCapabilities.Has(capability))
        throw new ImapIncapableException(capability);
    }

    private void RejectInvalidMailboxNameArgument(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("invalid mailbox name", "mailboxName");
    }

    private void RejectTransactionProceeding()
    {
      if (IsTransactionProceeding)
        throw new InvalidOperationException("transaction proceeding");
    }

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}
