// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;

namespace Smdn.Net.Imap4.Client.Session {
  internal class ImapUpdatedStatusResponseProcessor : ImapResponseProcessorBase {
    public ImapUpdatedStatusResponseProcessor(ImapSession session)
      : base(session)
    {
    }

    // 7.2. Server Responses - Server and Mailbox Status
    // 7.2.6. FLAGS Response
    // 7.3. Server Responses - Mailbox Size
    // 7.3.1. EXISTS Response
    // 7.3.2. RECENT Response
    // 7.4. Server Responses - Message Status
    // 7.4.1. EXPUNGE Response
    // 7.4.2. FETCH Response

    // SelectedMailbox must be updated
    public override bool Process(ImapResponse response, IImapTransaction t)
    {
      if (response is ImapDataResponse)
        return Process(response as ImapDataResponse);
      else if (response is ImapStatusResponse)
        return Process(response as ImapStatusResponse);
      else
        return false;
    }

    private bool Process(ImapStatusResponse status)
    {
      if (Session.SelectedMailbox == null)
        return false;

      if (status.Condition == ImapResponseCondition.Ok) {
        if (status.ResponseText.Code == ImapResponseCode.Unseen) {
          Session.SelectedMailbox.FirstUnseen = ImapResponseTextConverter.FromUnseen(status.ResponseText);
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.PermanentFlags) {
          Session.SelectedMailbox.PermanentFlags = ImapResponseTextConverter.FromPermanentFlags(status.ResponseText);
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.UidNext) {
          Session.SelectedMailbox.UidNext = ImapResponseTextConverter.FromUidNext(status.ResponseText);
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.UidValidity) {
          Session.SelectedMailbox.UidValidity = ImapResponseTextConverter.FromUidValidity(status.ResponseText);
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.ReadOnly) {
          Session.SelectedMailbox.ReadOnly = true;
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.ReadWrite) {
          Session.SelectedMailbox.ReadOnly = false;
          return true;
        }
        /*
         * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
         * http://tools.ietf.org/html/rfc4551
         * 3. IMAP Protocol Changes
         * 3.1. New OK Untagged Responses for SELECT and EXAMINE
         */
        else if (status.ResponseText.Code == ImapResponseCode.HighestModSeq) {
          Session.SelectedMailbox.ModificationSequences = true;
          Session.SelectedMailbox.HighestModSeq = ImapResponseTextConverter.FromHighestModSeq(status.ResponseText);
          return true;
        }
        else if (status.ResponseText.Code == ImapResponseCode.NoModSeq) {
          Session.SelectedMailbox.ModificationSequences = false;
          return true;
        }
      }
      else if (status.Condition == ImapResponseCondition.No) {
        if (status.ResponseText.Code == ImapResponseCode.UidNotSticky) {
          Session.SelectedMailbox.UidPersistent = false;
          return true;
        }
      }

      return false;
    }

    private bool Process(ImapDataResponse data)
    {
      ImapUpdatedStatus updatedStatus = null;

      if (data.Type == ImapDataResponseType.Flags) {
        var flags = ImapDataResponseConverter.FromFlags(data);

        //       The update from the FLAGS response MUST be recorded by the client.
        if (Session.SelectedMailbox != null)
          Session.SelectedMailbox.ApplicableFlags = flags;

        updatedStatus = ImapUpdatedStatus.CreateFlagsUpdated(flags);
      }
      else if (data.Type == ImapDataResponseType.Exists) {
        var exists = ImapDataResponseConverter.FromExists(data);

        //       The update from the EXISTS response MUST be recorded by the
        //       client.
        if (Session.SelectedMailbox != null)
          Session.SelectedMailbox.ExistsMessage = exists;

        updatedStatus = ImapUpdatedStatus.CreateExistsUpdated(exists);
      }
      else if (data.Type == ImapDataResponseType.Recent) {
        var recent = ImapDataResponseConverter.FromRecent(data);

        //       The update from the RECENT response MUST be recorded by the
        //       client.
        if (Session.SelectedMailbox != null)
          Session.SelectedMailbox.RecentMessage = recent;

        updatedStatus = ImapUpdatedStatus.CreateRecentUpdated(recent);
      }
      else if (data.Type == ImapDataResponseType.Expunge) {
        var expunge = ImapDataResponseConverter.FromExpunge(data);

        // TODO:
        //       The update from the EXPUNGE response MUST be recorded by the
        //       client.
        /*
                 The EXPUNGE response reports that the specified message sequence
                 number has been permanently removed from the mailbox.  The message
                 sequence number for each successive message in the mailbox is
                 immediately decremented by 1, and this decrement is reflected in
                 message sequence numbers in subsequent responses (including other
                 untagged EXPUNGE responses).
         */
        if (Session.SelectedMailbox != null)
          Session.SelectedMailbox.ExistsMessage -= 1;

        updatedStatus = ImapUpdatedStatus.CreateExpungedUpdated(expunge);
      }
      else if (data.Type == ImapDataResponseType.Fetch) {
        var messageAttribute = ImapDataResponseConverter.FromFetchMessageDynamicAttribute(data);

        updatedStatus = ImapUpdatedStatus.CreateMessageAttributeUpdated(messageAttribute);
      }

      if (updatedStatus == null) {
        return false;
      }
      else {
        Session.LastUpdatedStatus = Session.LastUpdatedStatus.MergeWith(updatedStatus);
        return true;
      }
    }
  }
}