// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Transaction.BuiltIn {
  internal abstract class SearchTransactionBase : ImapTransactionBase<ImapTransactionResult<ImapMatchedSequenceSet>>, IImapExtension {
    protected virtual ImapCapability RequiredCapability {
      get { return resultSpecifierCapabilityRequirement; }
    }

    ImapCapability IImapExtension.RequiredCapability {
      get { return RequiredCapability; }
    }

    protected SearchTransactionBase(string command, ImapConnection connection, bool uid, ImapCapability resultSpecifierCapabilityRequirement)
      : base(connection)
    {
      commandString = command;

      if (uid)
        commandString = "UID " + commandString;

      this.uid = uid;
      this.resultSpecifierCapabilityRequirement = resultSpecifierCapabilityRequirement;
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("searching criteria"))
        return ProcessArgumentNotSetted;

      if (this is SearchTransaction) {
        return ProcessSearch;
      }
      else {
        // http://tools.ietf.org/html/rfc5256
        // BASE.6.4.SORT. SORT Command
        //   The charset argument is mandatory (unlike SEARCH)
        if (!Request.Arguments.ContainsKey("charset specification"))
          return ProcessArgumentNotSetted;
        if (!Request.Arguments.ContainsKey("sort criteria"))
          return ProcessArgumentNotSetted;

        return ProcessSearch;
      }
    }

    private void ProcessArgumentNotSetted()
    {
      if (this is SearchTransaction)
        FinishError(ImapTransactionResultCode.RequestError, "arguments 'searching criteria' must be setted");
      else
        FinishError(ImapTransactionResultCode.RequestError, "arguments 'sort criteria', 'charset specification' and 'searching criteria' must be setted");
    }

    // RFC 4466 - Collected Extensions to IMAP4 ABNF
    // http://tools.ietf.org/html/rfc4466
    // 2.6. Extensions to SEARCH Command
    // 2.6.1. Extended SEARCH Command
    //    Arguments:  OPTIONAL result specifier
    //                OPTIONAL [CHARSET] specification
    //                searching criteria (one or more)
    //    Responses:  REQUIRED untagged response: SEARCH (*)
    //    Result:     OK - search completed
    //                NO - search error: cannot search that [CHARSET] or
    //                     criteria
    //                BAD - command unknown or arguments invalid

    // http://tools.ietf.org/html/rfc5256
    // BASE.6.4.SORT. SORT Command
    //    Arguments:  sort program
    //                charset specification
    //                searching criteria (one or more)
    //    Data:       untagged responses: SORT
    //    Result:     OK - sort completed
    //                NO - sort error: can't sort that charset or
    //                     criteria
    //                BAD - command unknown or arguments invalid
    private void ProcessSearch()
    {
      if (this is SearchTransaction) {
        // SEARCH / UID SEARCH
        var args = new List<ImapString>();

        if (Request.Arguments.ContainsKey("result specifier")) {
          args.Add("RETURN");
          args.Add(Request.Arguments["result specifier"]);
        }

        if (Request.Arguments.ContainsKey("charset specification")) {
          args.Add("CHARSET");
          args.Add(Request.Arguments["charset specification"]);
        }

        args.Add(Request.Arguments["searching criteria"]);

        SendCommand(commandString,
                    ProcessReceiveResponse,
                    args.ToArray());
      }
      else {
        // SORT / UID SORT
        if (Request.Arguments.ContainsKey("result specifier"))
          SendCommand(commandString,
                      ProcessReceiveResponse,
                      "RETURN",
                      Request.Arguments["result specifier"],
                      Request.Arguments["sort criteria"],
                      Request.Arguments["charset specification"],
                      Request.Arguments["searching criteria"]);
        else
          SendCommand(commandString,
                      ProcessReceiveResponse,
                      Request.Arguments["sort criteria"],
                      Request.Arguments["charset specification"],
                      Request.Arguments["searching criteria"]);
      }
    }

    protected override void OnDataResponseReceived(ImapDataResponse data)
    {
      // 7.2.5 SEARCH Response
      //   Contents:   zero or more numbers
      //   Example:    S: * SEARCH 2 3 6
      if (this is SearchTransaction && data.Type == ImapDataResponseType.Search)
        messages = ImapDataResponseConverter.FromSearch(data, uid);
      // http://tools.ietf.org/html/rfc5256
      // BASE.7.2.SORT. SORT Response
      //   Data:       zero or more numbers
      //   Example:    S: * SORT 2 3 6
      else if (this is SortTransaction && data.Type == ImapDataResponseType.Sort)
        messages = ImapDataResponseConverter.FromSort(data, uid);
      // http://tools.ietf.org/html/rfc4466
      // 2.6.2. ESEARCH untagged response
      //    Contents:   one or more search-return-data pairs
      else if (data.Type == ImapDataResponseType.ESearch)
        messages = ImapDataResponseConverter.FromESearch(data);

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok)
        Finish(new ImapTransactionResult<ImapMatchedSequenceSet>(ImapTransactionResultCode.Ok, messages, tagged.ResponseText.Text));
      else
        base.OnTaggedStatusResponseReceived(tagged);
    }

    private readonly bool uid;
    private /*readonly*/ string commandString;
    private /*readonly*/ ImapCapability resultSpecifierCapabilityRequirement;
    private ImapMatchedSequenceSet messages = null;
  }
}
