// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Transaction {
  // Layers:
  //   Session.ImapSession
  // * Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public abstract class ImapTransactionBase : ImapTransactionBase<ImapTransactionResult> {
    protected ImapTransactionBase(ImapConnection connection)
      : base(connection)
    {
    }
  }

  public abstract class ImapTransactionBase<TResult> : ImapTransactionBase<ImapTransactionRequest, TResult>
    where TResult : ImapTransactionResult, new()
  {
    protected ImapTransactionBase(ImapConnection connection)
      : base(connection, new ImapTransactionRequest())
    {
    }
  }

  public abstract class ImapTransactionBase<TRequest, TResult> : IImapTransaction, IDisposable
    where TRequest : ImapTransactionRequest
    where TResult : ImapTransactionResult, new()
  {
    protected delegate void ProcessTransactionDelegate();

    public ImapConnection Connection {
      get
      {
        CheckDisposed();
        return connection;
      }
    }

    public TRequest Request {
      get
      {
        CheckDisposed();
        return request;
      }
    }

    ImapTransactionRequest IImapTransaction.Request {
      get { return this.Request; }
    }

    public TResult Result {
      get
      {
        CheckDisposed();
        return result;
      }
    }

    ImapTransactionResult IImapTransaction.Result {
      get { return this.Result; }
    }

    public ImapTransactionState State {
      get
      {
        CheckDisposed();
        return state;
      }
    }

    protected string LastCommandTag {
      get { return lastCommandTag; }
    }

    protected ImapTransactionBase(ImapConnection connection, TRequest request)
    {
      if (connection == null)
        throw new ArgumentNullException("connection");
      if (request == null)
        throw new ArgumentNullException("request");

      this.connection = connection;
      this.request = request;
      this.result = null;
      this.receivedResponses = new Queue<ImapResponse>();
    }

    public virtual void Dispose()
    {
      if (disposed)
        return;

      // nothing to do in this class
      connection = null;
      request = null;
      result = null;
      receivedResponses = null;

      disposed = true;
    }

    private ImapConnection connection;
    private TRequest request;
    private TResult result;
    private bool disposed = false;

#region "processing transaction / state transition"
    void IImapTransaction.Start()
    {
      CheckDisposed();

      if (state == ImapTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");

      currentProcess = Reset();

      state = ImapTransactionState.Processing;
    }

    protected abstract ProcessTransactionDelegate Reset();

    void IImapTransaction.Process()
    {
      CheckDisposed();
      RejectFinished();

      try {
        if (currentProcess != null)
          currentProcess();
      }
      catch (Exception ex) {
        Finish(CreateResult(ex));
        Trace.Log(ex);
      }
    }

    /// <remarks>do not call directly</remarks>
    protected void ProcessReceiveResponse()
    {
      var response = Receive();

      if (response == null)
        return;

      try {
        OnResponseReceived(response);
      }
      catch (ImapMalformedDataException ex) {
        if (response is ImapTaggedStatusResponse) {
          // error in response data, abort transaction
          throw ex;
        }
        else {
          // error in response data, continue processing
          Trace.Log(ex);
        }
      }
    }

    protected void Transit(ProcessTransactionDelegate process)
    {
      if (state == ImapTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");
      if (process == null)
        throw new ArgumentNullException("process");

      currentProcess = process;
    }

    private TResult CreateResult(Exception exception)
    {
      var result = (exception is TimeoutException)
        ? CreateResult(ImapTransactionResultCode.SocketTimeout, "timed out")
        : CreateResult(ImapTransactionResultCode.InternalError, "internal error");

      result.Exception = exception;

      return result;
    }

    private TResult CreateResult(ImapTransactionResultCode code, string description)
    {
      var result = new TResult();

      result.Code = code;
      result.Description = description;

      return result;
    }

    protected void FinishError(ImapTaggedStatusResponse tagged)
    {
      switch (tagged.Condition) {
        case ImapResponseCondition.No:
          FinishError(ImapTransactionResultCode.No, tagged.ResponseText.Text);
          break;
        case ImapResponseCondition.Bad:
          FinishError(ImapTransactionResultCode.Bad, tagged.ResponseText.Text);
          break;
        case ImapResponseCondition.Bye:
          FinishError(ImapTransactionResultCode.Bye, tagged.ResponseText.Text);
          break;
        default: // OK, PREAUTH
          throw new InvalidOperationException("status is not NO/BAD/BYE");
      }
    }

    protected void FinishError(ImapTransactionResultCode code, string description)
    {
      if ((int)code < 300)
        throw new ArgumentException("status is not error", "status");

      Finish(CreateResult(code, description));
    }

    protected void FinishOk(ImapTaggedStatusResponse tagged)
    {
      switch (tagged.Condition) {
        case ImapResponseCondition.Ok:
          Finish(CreateResult(ImapTransactionResultCode.Ok, tagged.ResponseText.Text));
          break;
        case ImapResponseCondition.PreAuth:
          Finish(CreateResult(ImapTransactionResultCode.PreAuth, tagged.ResponseText.Text));
          break;
        default: // NO, BAD, BYE
          throw new InvalidOperationException("status is not OK/PREAUTH");
      }
    }

    protected internal void Finish(TResult result)
    {
      RejectFinished();

      this.result = result;
      this.result.ReceivedResponses = receivedResponses;
      this.result.TaggedStatusResponse = taggedStatusResponse;

      currentProcess = null;

      state = ImapTransactionState.Finished;
    }

    private void RejectFinished()
    {
      if (state == ImapTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");
    }

    private ImapTransactionState state = ImapTransactionState.NotStarted;
    private ProcessTransactionDelegate currentProcess;
#endregion

#region "sending / receiving"
    protected ImapResponse Receive()
    {
      RejectFinished();

      try {
        var resp = Connection.TryReceiveResponse();

        if (resp != null) {
          receivedResponses.Enqueue(resp);

          if (resp is ImapTaggedStatusResponse)
            taggedStatusResponse = (resp as ImapTaggedStatusResponse);
        }

        return resp;
      }
      catch (ImapMalformedResponseException ex) {
        // protocol error, abort transaction
        throw ex;
      }
    }

    protected void SendCommand(string command, ProcessTransactionDelegate nextProcess, params ImapString[] arguments)
    {
      RejectFinished();

      var comm = Connection.CreateCommand(command, arguments);

      commandContinuationContext = Connection.SendCommand(comm);

      lastCommandTag = comm.Tag;

      Transit(nextProcess);
    }

    protected void SendContinuation(params ImapString[] arguments)
    {
      RejectFinished();

      commandContinuationContext = Connection.SendCommand(Connection.CreateContinuingCommand(arguments));
    }

    protected bool SendContinuation()
    {
      RejectFinished();

      if (commandContinuationContext == null)
        return false;

      commandContinuationContext = Connection.SendCommand(commandContinuationContext);

      return true;
    }

    private string lastCommandTag = null;
    private ImapConnection.ICommandContinuationContext commandContinuationContext = null;
    private Queue<ImapResponse> receivedResponses;
    private ImapTaggedStatusResponse taggedStatusResponse = null;
#endregion

#region "response processing"
    private void OnResponseReceived(ImapResponse response)
    {
      if (response is ImapDataResponse)
        OnDataResponseReceived(response as ImapDataResponse);
      else if (response is ImapStatusResponse)
        OnStatusResponseReceived(response as ImapStatusResponse);
      else if (response is ImapCommandContinuationRequest)
        OnCommandContinuationRequestReceived(response as ImapCommandContinuationRequest);
    }

    protected virtual void OnDataResponseReceived(ImapDataResponse data)
    {
    }

    protected virtual void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      if (!SendContinuation())
        throw new ImapException("unexpected command continuation or sending literal failed");
    }

    protected virtual void OnStatusResponseReceived(ImapStatusResponse status)
    {
      if (status.ResponseText.Code == ImapResponseCode.Alert)
        // ALERT
        Trace.Info(status.ResponseText.Text);
      else if (status.ResponseText.Code == ImapResponseCode.ClientBug)
        // CLIENTBUG
        Trace.Info(status.ResponseText.Text);

      if (status is ImapUntaggedStatusResponse) {
        OnUntaggedStatusResponseReceived(status as ImapUntaggedStatusResponse);
      }
      else if (status is ImapTaggedStatusResponse) {
        var tagged = status as ImapTaggedStatusResponse;

        if (tagged.Tag == lastCommandTag)
          OnTaggedStatusResponseReceived(status as ImapTaggedStatusResponse);
        else
          // TODO: '5.5. Multiple Commands in Progress' or not matched tag
          throw new ImapException("unexpected response");
      }
    }

    protected virtual void OnUntaggedStatusResponseReceived(ImapUntaggedStatusResponse untagged)
    {
      if (untagged.Condition == ImapResponseCondition.Bye)
        // The BYE response is always untagged
        FinishError(ImapTransactionResultCode.Bye, untagged.ResponseText.Text);
    }

    protected virtual void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok || tagged.Condition == ImapResponseCondition.PreAuth)
        FinishOk(tagged);
      else
        FinishError(tagged);
    }
#endregion

    public override string ToString()
    {
      return string.Format("{{Connection={0}, State={1}, Request={2}, Result={3}}}", connection, state, request, result);
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }
  }
}
