// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.WebClients {
  public class ImapSessionManager : IWebRequestCreate {
    private static readonly ImapSessionManager instance = new ImapSessionManager();
    private static readonly NetworkCredential AnonymousCredential = new NetworkCredential("anonymous", "@anonymous");

    public static X509Certificate2Collection CertCollection {
      get { return ImapConnection.CertCollection; }
    }

    public static RemoteCertificateValidationCallback CertificateValidationCallback {
      get { return ImapConnection.CertificateValidationCallback; }
      set { ImapConnection.CertificateValidationCallback = value; }
    }

    public static LocalCertificateSelectionCallback CertificateSelectionCallback {
      get { return ImapConnection.CertificateSelectionCallback; }
      set { ImapConnection.CertificateSelectionCallback = value; }
    }

    private static UpgradeConnectionStreamCallback createSslStreamCallback = ImapConnection.CreateSslStream;

    public static UpgradeConnectionStreamCallback CreateSslStreamCallback {
      get { return createSslStreamCallback; }
      set
      {
        if (createSslStreamCallback == null)
          throw new ArgumentNullException("CreateSslStreamCallback");
        createSslStreamCallback = value;
      }
    }

    private static Dictionary<string, string> clientID;

    public static Dictionary<string, string> ClientID {
      get { return clientID; }
      set { clientID = value; }
    }

    static ImapSessionManager()
    {
      /*
       * 3.3. Defined Field Values
       *    Any string may be sent as a field, but the following are defined to
       *    describe certain values that might be sent.
       * 
       *      name            Name of the program
       *      version         Version number of the program
       *      os              Name of the operating system
       *      os-version      Version of the operating system
       *      vendor          Vendor of the client/server
       *      support-url     URL to contact for support
       *      address         Postal address of contact/vendor
       *      date            Date program was released, specified as a date-time
       *                        in IMAP4rev1
       *      command         Command used to start the program
       *      arguments       Arguments supplied on the command line, if any
       *                        if any
       *      environment     Description of environment, i.e., UNIX environment
       *                        variables or Windows registry settings
       */
      var assembly = System.Reflection.Assembly.GetExecutingAssembly();

      clientID = new Dictionary<string, string>();
      clientID.Add("name", assembly.GetName().Name);
      clientID.Add("version", assembly.GetName().Version.ToString());
      clientID.Add("environment", string.Format("{0} {1}", Runtime.Name, Environment.Version));
    }

    public static bool RegisterWebRequestPrefix()
    {
      if (WebRequest.RegisterPrefix(ImapUri.UriSchemeImap, instance) | WebRequest.RegisterPrefix(ImapUri.UriSchemeImaps, instance)) {
        // this will broke in windows
        // ImapUri.RegisterParser();
        return true;
      }
      else {
        return false;
      }
    }

    public static void DisconnectFrom(Uri serverUri)
    {
      if (!ImapUri.IsImap(serverUri))
        throw new ArgumentException("scheme must be imap or imaps", "serverUri");

      var port = serverUri.Port;

      if (port == -1)
        port = ImapUri.GetDefaultPortFromScheme(serverUri);

      DisconnectFrom(serverUri.Host, port);
    }

    public static void DisconnectFrom(string host, int port)
    {
      instance.InternalDisconnectFrom(host, port);
    }

    internal static ImapSession CreateSession(IImapSessionProfile profile)
    {
      try {
        ImapSession session = null;

        var result = CreateSession(profile, out session);

        if (result.Succeeded)
          return session;
        else
          throw new WebException(result.Description, null, WebExceptionStatus.ProtocolError, new ImapWebResponse(result));
      }
      catch (ImapConnectionException ex) {
        throw new WebException("connection error", ex, WebExceptionStatus.ConnectFailure, null);
      }
      catch (ImapException ex) {
        throw new WebException("internal error", ex, WebExceptionStatus.UnknownError, null);
      }
      catch (TimeoutException ex) {
        throw new WebException("timed out", ex, WebExceptionStatus.Timeout, null);
      }
      catch (WebException) {
        throw;
      }
    }

    private static ImapTransactionResult CreateSession(IImapSessionProfile profile, out ImapSession session)
    {
      ImapTransactionResult result;
      session = null;

      var authority = profile.Authority;
      var username = ImapStyleUriParser.GetUser(authority);
      var authMechanism = ImapStyleUriParser.GetAuthType(authority);

      for (var referral = 0;; referral++) {
        session = new ImapSession(authority.Host,
                                  authority.Port,
                                  true,
                                  profile.Timeout,
                                  (authority.Scheme == ImapUri.UriSchemeImaps)
                                    ? createSslStreamCallback
                                    : null);

        session.HandlesIncapableAsException = false;
        session.HandlesReferralAsException = true;
        session.TransactionTimeout  = profile.Timeout;
        session.SendTimeout         = profile.ReadWriteTimeout;
        session.ReceiveTimeout      = profile.ReadWriteTimeout;

        // try querying server capability (ignore error)
        session.Capability();

        if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
          throw new WebException("IMAP4rev1 not supported", WebExceptionStatus.RequestCanceled);

        if (profile.UseTlsIfAvailable && session.ServerCapabilities.Has(ImapCapability.StartTls)) {
          var r = session.StartTls(createSslStreamCallback, true);

          if (r.Failed)
            throw new WebException(r.Description, null, WebExceptionStatus.SecureChannelFailure, null);
          else if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
            throw new WebException("IMAP4rev1 not supported", WebExceptionStatus.RequestCanceled);
        }

        if (profile.UseDeflateIfAvailable && session.ServerCapabilities.Has(ImapCapability.CompressDeflate)) {
          var r = session.Compress(ImapCompressionMechanism.Deflate);

          if (r.Failed)
            throw new WebException(r.Description, null, WebExceptionStatus.RequestCanceled, null);
        }

        try {
          result = Authenticate(session,
                                profile.Credentials,
                                username,
                                authMechanism,
                                GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? new string[0]));
        }
        catch (ImapLoginReferralException ex) {
          if (profile.AllowLoginRedirect && referral == 0) {
            authority = ex.ReferToUri;
            continue;
          }
          else {
            throw new WebException("authentication failed", ex, WebExceptionStatus.RequestCanceled, null);
          }
        }

        if (result == null) {
          throw new WebException("appropriate authentication mechanism not found", WebExceptionStatus.RequestCanceled);
        }
        else if (result.Failed) {
          try {
            session.Disconnect(false);
          }
          catch (ImapConnectionException) {
            // ignore
          }

          return result;
        }

        break;
      }

      // update server info
      if (session.ServerCapabilities.Has(ImapCapability.Namespace))
        session.Namespace();

      if (session.ServerCapabilities.Has(ImapCapability.ID))
        session.ID(clientID);

      return result;
    }

#region "authentication"
    private static IEnumerable<ImapAuthenticationMechanism> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = Smdn.Security.Authentication.Sasl.Client.SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Exists(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.InvariantCultureIgnoreCase);
        }))
          yield return ImapAuthenticationMechanism.GetKnownOrCreate(mechanism);
      }
    }

    private static ImapTransactionResult Authenticate(ImapSession session, ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authMechanism, IEnumerable<ImapAuthenticationMechanism> usingSaslMechanisms)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       *
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = ImapAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = ImapAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate)
        return AuthenticateWithAppropriateMechanism(session, credentials, username, usingSaslMechanisms);
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous)
        return AuthenticateAsAnonymous(session, canFallback);
      else
        return AuthenticateWithSuppliedMechanism(session, credentials, username, authMechanism);
    }
    
    private static ImapTransactionResult AuthenticateWithAppropriateMechanism(ImapSession session, ICredentialsByHost credentials, string username, IEnumerable<ImapAuthenticationMechanism> usingSaslMechanisms)
    {
      ImapTransactionResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (session.ServerCapabilities.IsCapable(mechanism)) {
          result = session.Authenticate(credentials, username, mechanism);

          if (result.Succeeded)
            break;
        }
      }

      if ((result == null || result.Failed) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        result = session.Login(credentials, username);

      return result;
    }

    private static ImapTransactionResult AuthenticateWithSuppliedMechanism(ImapSession session, ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authMechanism)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       */
      return session.Authenticate(credentials, username, authMechanism);
    }

    private static ImapTransactionResult AuthenticateAsAnonymous(ImapSession session, bool canFallback)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */
      ImapTransactionResult result = null;

      if (session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Anonymous))
        // try AUTHENTICATE ANONYUMOUS
        result = session.Authenticate(AnonymousCredential, ImapAuthenticationMechanism.Anonymous);

      if ((result == null || (result.Failed && canFallback)) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        // try anonymous LOGIN
        result = session.Login(AnonymousCredential);

      return result;
    }
#endregion

    /*
     * instance members
     */
    private ImapSessionManager()
    {
    }

    WebRequest IWebRequestCreate.Create(Uri uri)
    {
      switch (ImapStyleUriParser.GetUriForm(uri)) {
        case ImapUriForm.Server:          return new ImapServerWebRequest(uri, this);
        case ImapUriForm.ListMessages:    return new ImapMailboxWebRequest(uri, this);
        case ImapUriForm.FetchMessage:    return new ImapFetchMessageWebRequest(uri, this);
        case ImapUriForm.SearchMessages:  return new ImapSearchMessageWebRequest(uri, this);
        default: throw new ArgumentException("invalid URI form", "uri");
      }
    }

    internal ImapSession GetExistSession(Uri requestUri)
    {
      var authority = ImapStyleUriParser.GetAuthority(requestUri);

      if (sessions.ContainsKey(authority))
        return sessions[authority];
      else
        return null;
    }

    internal void RegisterSession(Uri requestUri, ImapSession session)
    {
      var authority = ImapStyleUriParser.GetAuthority(requestUri);

      ImapSession existSession = null;

      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (sessions.ContainsKey(authority)) {
          existSession = sessions[authority];
          sessions.Remove(authority);
        }

        sessions.Add(authority, session);
      }

      if (existSession != null)
        existSession.Disconnect(true);
    }

    internal void UnregisterSession(Uri requestUri)
    {
      UnregisterSession(ImapStyleUriParser.GetAuthority(requestUri));
    }

    private void UnregisterSession(string authority)
    {
      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (!sessions.ContainsKey(authority))
          return;

        sessions.Remove(authority);
      }
    }

    private void InternalDisconnectFrom(string host, int port)
    {
      if (host == null)
        throw new ArgumentNullException("host", host);
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      var keys = new List<string>();

      foreach (var pair in sessions) {
        var session = pair.Value;

        if (session.IsDisposed || session.Connection == null) {
          keys.Add(pair.Key);
        }
        else if (string.Equals(session.Connection.Host, host, StringComparison.InvariantCultureIgnoreCase) &&
                 session.Connection.Port == port) {
          keys.Add(pair.Key);

          session.Disconnect(true);
        }
      }

      foreach (var key in keys) {
        UnregisterSession(key);
      }
    }

    private Dictionary<string, ImapSession> sessions = new Dictionary<string, ImapSession>(StringComparer.InvariantCultureIgnoreCase);
  }
}