// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

namespace Smdn.Net.Imap4 {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //   * ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.6. STORE Command
  //   message data item name
  //   value for message data item
  public class ImapStoreDataItem : ImapCombinableDataItem {

    // FLAGS <flag list>
    //     Replace the flags for the message (other than \Recent) with the
    //     argument.  The new value of the flags is returned as if a FETCH
    //     of those flags was done.
    public static ImapStoreDataItem ReplaceFlags(params ImapMessageFlag[] flags)
    {
      return ReplaceFlags(new string[] {}, flags);
    }

    public static ImapStoreDataItem ReplaceFlags(params string[] keywords)
    {
      return ReplaceFlags(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem ReplaceFlags(string[] keywords, params ImapMessageFlag[] flags)
    {
      return ReplaceFlags(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem ReplaceFlags(ImapMessageFlagList flags)
    {
      return FlagsCommon("FLAGS", flags);
    }

    // FLAGS.SILENT <flag list>
    //     Equivalent to FLAGS, but without returning a new value.
    public static ImapStoreDataItem ReplaceFlagsSilent(params ImapMessageFlag[] flags)
    {
      return ReplaceFlagsSilent(new string[] {}, flags);
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(params string[] keywords)
    {
      return ReplaceFlagsSilent(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(string[] keywords, params ImapMessageFlag[] flags)
    {
      return ReplaceFlagsSilent(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(ImapMessageFlagList flags)
    {
      return FlagsCommon("FLAGS.SILENT", flags);
    }

    // +FLAGS <flag list>
    //     Add the argument to the flags for the message.  The new value
    //     of the flags is returned as if a FETCH of those flags was done.
    public static ImapStoreDataItem AddFlags(params ImapMessageFlag[] flags)
    {
      return AddFlags(new string[] {}, flags);
    }

    public static ImapStoreDataItem AddFlags(params string[] keywords)
    {
      return AddFlags(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem AddFlags(string[] keywords, params ImapMessageFlag[] flags)
    {
      return AddFlags(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem AddFlags(ImapMessageFlagList flags)
    {
      return FlagsCommon("+FLAGS", flags);
    }

    // +FLAGS.SILENT <flag list>
    //     Equivalent to +FLAGS, but without returning a new value.
    public static ImapStoreDataItem AddFlagsSilent(params ImapMessageFlag[] flags)
    {
      return AddFlagsSilent(new string[] {}, flags);
    }

    public static ImapStoreDataItem AddFlagsSilent(params string[] keywords)
    {
      return AddFlagsSilent(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem AddFlagsSilent(string[] keywords, params ImapMessageFlag[] flags)
    {
      return AddFlagsSilent(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem AddFlagsSilent(ImapMessageFlagList flags)
    {
      return FlagsCommon("+FLAGS.SILENT", flags);
    }

    // -FLAGS <flag list>
    //     Remove the argument from the flags for the message.  The new
    //     value of the flags is returned as if a FETCH of those flags was
    //     done.
    public static ImapStoreDataItem RemoveFlags(params ImapMessageFlag[] flags)
    {
      return RemoveFlags(new string[] {}, flags);
    }

    public static ImapStoreDataItem RemoveFlags(params string[] keywords)
    {
      return RemoveFlags(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem RemoveFlags(string[] keywords, params ImapMessageFlag[] flags)
    {
      return RemoveFlags(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem RemoveFlags(ImapMessageFlagList flags)
    {
      return FlagsCommon("-FLAGS", flags);
    }

    // -FLAGS.SILENT <flag list>
    //     Equivalent to -FLAGS, but without returning a new value.
    public static ImapStoreDataItem RemoveFlagsSilent(params ImapMessageFlag[] flags)
    {
      return RemoveFlagsSilent(new string[] {}, flags);
    }

    public static ImapStoreDataItem RemoveFlagsSilent(params string[] keywords)
    {
      return RemoveFlagsSilent(keywords, new ImapMessageFlag[] {});
    }

    public static ImapStoreDataItem RemoveFlagsSilent(string[] keywords, params ImapMessageFlag[] flags)
    {
      return RemoveFlagsSilent(new ImapMessageFlagList(keywords, flags));
    }

    public static ImapStoreDataItem RemoveFlagsSilent(ImapMessageFlagList flags)
    {
      return FlagsCommon("-FLAGS.SILENT", flags);
    }

    // common construction method
    private static ImapStoreDataItem FlagsCommon(ImapString itemName, ImapMessageFlagList flags)
    {
      if (flags == null)
        throw new ArgumentNullException("flags");

      return new ImapStoreDataItem(itemName, flags.GetNonApplicableFlagsRemoved().ToArray());
    }

    public ImapString ItemName {
      get { return itemName; }
    }

    private ImapStoreDataItem(ImapString itemName, ImapString[] items)
      : base(items)
    {
      this.itemName = itemName;
    }

    protected override ImapStringList GetCombined()
    {
      return ToParenthesizedString();
    }

    public override string ToString()
    {
      return string.Format("{0} {1}", itemName, GetCombined());
    }

    private ImapString itemName;
  }
}