using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using NUnit.Framework;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol.Client {
  [TestFixture]
  public class DataResponseConverterTest {
    [SetUp]
    public void Setup()
    {
      baseStream = new MemoryStream();
      stream = new LineOrientedBufferedStream(baseStream);
      receiver = new ImapResponseReceiver(stream);
    }

    [TearDown]
    public void TearDown()
    {
      // nothing to do
    }

    private ImapDataResponse GetSingleDataResponse(string response)
    {
      var resp = Encoding.ASCII.GetBytes(response);

      baseStream.Seek(0, SeekOrigin.Begin);
      baseStream.Write(resp, 0, resp.Length);
      baseStream.Seek(0, SeekOrigin.Begin);

      var r = receiver.ReceiveResponse();

      Assert.IsTrue(r is ImapDataResponse, "response type");

      return r as ImapDataResponse;
    }

    [Test]
    public void TestCapabilityResponse()
    {
      var response =
        GetSingleDataResponse("* CAPABILITY IMAP4rev1 STARTTLS AUTH=GSSAPI XPIG-LATIN\r\n");

      var caps = ImapDataResponseConverter.FromCapability(response);

      Assert.AreEqual(4, caps.Count, "capability count");
      Assert.IsTrue(caps.Has(ImapCapability.Imap4Rev1));
      Assert.IsTrue(caps.Has(ImapCapability.StartTls));
      Assert.IsTrue(caps.Has("IMAP4rev1"));
      Assert.IsTrue(caps.Has("STARTTLS"));
      Assert.IsTrue(caps.Has("AUTH=GSSAPI"));
      Assert.IsTrue(caps.Has("XPIG-LATIN"));
    }

    [Test]
    public void TestCapabilityResponseDuplicated()
    {
      var response =
        GetSingleDataResponse("* CAPABILITY IMAP4rev1 IMAP4rev1 XPIG-LATIN XPIG-LATIN\r\n");

      var caps = ImapDataResponseConverter.FromCapability(response);

      Assert.AreEqual(2, caps.Count, "capability count");
      Assert.IsTrue(caps.Has(ImapCapability.Imap4Rev1));
      Assert.IsTrue(caps.Has("XPIG-LATIN"));
    }

    [Test]
    public void TestListResponse()
    {
      var response =
        GetSingleDataResponse("* LIST (\\Noselect) \"/\" ~/Mail/foo\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NoSelect));
      Assert.IsTrue(mailboxList.NameAttributes.Has(@"\Noselect"));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("~/Mail/foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListResponseDuplicated()
    {
      var response =
        GetSingleDataResponse("* LIST (\\Noselect \\Noselect) \"/\" ~/Mail/foo\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NoSelect));

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListExtendedResponseWithEmptyExtendedData()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" ()\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListExtendedResponseWithUnknownExtension()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"X-EXTENSION\" ())\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListExtendedResponseWithChildInfo()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNotNull(mailboxList.ChildInfo);
      Assert.IsTrue(mailboxList.ChildInfo.Subscribed);
    }

    [Test]
    public void TestListExtendedResponseWithUnknownChildInfoExtension()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"CHILDINFO\" (\"X-EXTENSION\"))\r\n");

      var mailboxList = ImapDataResponseConverter.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNotNull(mailboxList.ChildInfo);
      Assert.IsFalse(mailboxList.ChildInfo.Subscribed);
    }

    [Test]
    public void TestLsubResponse()
    {
      var response =
        GetSingleDataResponse("* LSUB () \".\" #news.comp.mail.misc\r\n");

      var mailboxList = ImapDataResponseConverter.FromLsub(response);

      Assert.AreEqual(0, mailboxList.NameAttributes.Count);
      Assert.AreEqual(".", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("#news.comp.mail.misc", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestStatusResponse()
    {
      var response =
        GetSingleDataResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)\r\n");

      string name;
      var statusAttribute = ImapDataResponseConverter.FromStatus(response, out name);

      Assert.AreEqual("blurdybloop", name);
      Assert.IsNotNull(statusAttribute.Messages);
      Assert.AreEqual(231, statusAttribute.Messages.Value);
      Assert.IsNull(statusAttribute.Recent);
      Assert.IsNull(statusAttribute.Unseen);
      Assert.IsNotNull(statusAttribute.UidNext);
      Assert.AreEqual(44292, statusAttribute.UidNext.Value);
      Assert.IsNull(statusAttribute.UidValidity);
    }

    [Test]
    public void TestSearchResponse()
    {
      var response =
        GetSingleDataResponse("* SEARCH 2 3 6\r\n");

      var messages = ImapDataResponseConverter.FromSearch(response, false);

      Assert.IsNull(messages.HighestModSeq);

      var arr = messages.ToArray();

      Assert.AreEqual(3, arr.Length);
      Assert.AreEqual(2, arr[0]);
      Assert.AreEqual(3, arr[1]);
      Assert.AreEqual(6, arr[2]);
    }

    [Test]
    public void TestSortResponse()
    {
      var response =
        GetSingleDataResponse("* SORT 2 84 882\r\n");

      var messages = ImapDataResponseConverter.FromSort(response, false);

      Assert.IsNull(messages.HighestModSeq);

      var arr = messages.ToArray();

      Assert.AreEqual(3, arr.Length);
      Assert.AreEqual(2, arr[0]);
      Assert.AreEqual(84, arr[1]);
      Assert.AreEqual(882, arr[2]);
    }

    [Test]
    public void TestSearchNoMatchResponse()
    {
      var response =
        GetSingleDataResponse("* SEARCH\r\n");

      var messages = ImapDataResponseConverter.FromSearch(response, false);

      Assert.IsTrue(messages.IsEmpty);
    }

    [Test]
    public void TestSortNoMatchResponse()
    {
      var response =
        GetSingleDataResponse("* SORT\r\n");

      var messages = ImapDataResponseConverter.FromSort(response, false);

      Assert.IsTrue(messages.IsEmpty);
    }

    [Test]
    public void TestSearchResponseWithModSeq()
    {
      var response =
        GetSingleDataResponse("* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n");

      var messages = ImapDataResponseConverter.FromSearch(response, false);

      Assert.AreEqual(917162500UL, messages.HighestModSeq);

      var arr = messages.ToArray();

      Assert.AreEqual(10, arr.Length);
      Assert.AreEqual(2, arr[0]);
      Assert.AreEqual(5, arr[1]);
      Assert.AreEqual(6, arr[2]);
      Assert.AreEqual(7, arr[3]);
      Assert.AreEqual(11, arr[4]);
      Assert.AreEqual(12, arr[5]);
      Assert.AreEqual(18, arr[6]);
      Assert.AreEqual(19, arr[7]);
      Assert.AreEqual(20, arr[8]);
      Assert.AreEqual(23, arr[9]);
    }

    [Test]
    public void TestThreadResponse1()
    {
      var response =
        GetSingleDataResponse("* THREAD (2)(3 6 (4 23)(44 7 96))\r\n");

      var threadList = ImapDataResponseConverter.FromThread(response, false);

      var traversed   = new[] {2, 3, 6, 4, 23, 44, 7, 96};
      var childCounts = new[] {0, 1, 2, 1, 0,  1,  1, 0};
      var index = 0;

      threadList.Traverse(delegate(ImapThreadList list) {
        Assert.AreEqual(traversed[index], list.Number, "traversed {0}", index);
        Assert.AreEqual(childCounts[index], list.Children.Length, "children {0}", index);

        index++;
      });

      Assert.AreEqual(8, index, "count");

      var sequenceSet = threadList.ToSequenceSet().ToArray();

      Assert.AreEqual(index, sequenceSet.Length, "sequence set length");

      for (index = 0; index < sequenceSet.Length; index++) {
        Assert.AreEqual(traversed[index], sequenceSet[index], "sequence set {0}", index);
      }
    }

    [Test]
    public void TestThreadResponse2()
    {
      var response =
        GetSingleDataResponse("* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)" +
                              "(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n");

      var threadList = ImapDataResponseConverter.FromThread(response, false);

      var traversed   = new[] {
        166, 167, 168, 169, 172, 170, 171, 173, 174, 175, 176, 178, 181, 180, 179, 177, 183, 182, 188, 184,
        185, 186, 187, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 202, 201, 203, 204, 205, 206, 207, 208,
      };
      var childCounts = new[] {
          0,   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   8,   0,   0,   0,   0,
          0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,
      };
      var index = 0;

      threadList.Traverse(delegate(ImapThreadList list) {
        Assert.AreEqual(traversed[index], list.Number, "traversed {0}", index);
        Assert.AreEqual(childCounts[index], list.Children.Length, "children {0}", index);

        index++;
      });

      Assert.AreEqual(43, index, "count");
    }

    [Test]
    public void TestThreadNoMatchResponse()
    {
      var response =
        GetSingleDataResponse("* THREAD\r\n");

      var threadList = ImapDataResponseConverter.FromThread(response, false);

      Assert.AreEqual(0, threadList.Children.Length);
    }

    [Test]
    public void TestFlagsResponse()
    {
      var response =
        GetSingleDataResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n");

      var flags = ImapDataResponseConverter.FromFlags(response);

      Assert.AreEqual(5, flags.Count);
      Assert.IsTrue(flags.Has(ImapMessageFlag.Answered));
      Assert.IsTrue(flags.Has(ImapMessageFlag.Flagged));
      Assert.IsTrue(flags.Has(ImapMessageFlag.Deleted));
      Assert.IsTrue(flags.Has(ImapMessageFlag.Seen));
      Assert.IsTrue(flags.Has(ImapMessageFlag.Draft));
    }

    [Test]
    public void TestFlagsResponseDuplicated()
    {
      var response =
        GetSingleDataResponse("* FLAGS (\\Answered \\Answered $label2 $label2)\r\n");

      var flags = ImapDataResponseConverter.FromFlags(response);

      Assert.AreEqual(2, flags.Count);
      Assert.IsTrue(flags.Has(ImapMessageFlag.Answered));
      Assert.IsTrue(flags.Has("$label2"));
    }

    [Test]
    public void TestExistsResponse()
    {
      var response =
        GetSingleDataResponse("* 23 EXISTS\r\n");

      Assert.AreEqual(23, ImapDataResponseConverter.FromExists(response));
    }

    [Test]
    public void TestRecentResponse()
    {
      var response =
        GetSingleDataResponse("* 5 RECENT\r\n");

      Assert.AreEqual(5, ImapDataResponseConverter.FromRecent(response));
    }

    [Test]
    public void TestExpungeResponse()
    {
      var response =
        GetSingleDataResponse("* 44 EXPUNGE\r\n");

      Assert.AreEqual(44, ImapDataResponseConverter.FromExpunge(response));
    }

    [Test]
    public void TestFetchResponse()
    {
      var response =
        GetSingleDataResponse("* 23 FETCH (FLAGS (\\Seen) RFC822.SIZE 44827)\r\n");

      var message = ImapDataResponseConverter.FromFetch(response);

      Assert.AreEqual(23, message.Sequence);
      Assert.AreEqual(1, message.Flags.Count);
      Assert.IsTrue(message.Flags.Has(ImapMessageFlag.Seen));
      Assert.AreEqual(44827, message.Rfc822Size);

      Assert.IsNull(message.Body);
      Assert.IsNull(message.BodyStructure);
      Assert.IsNull(message.Envelope);
      Assert.AreEqual(0, message.Uid);
      Assert.IsNull(message.InternalDate);
    }

    [Test]
    public void TestNamespaceResponse1()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" \"/\")) NIL ((\"Public Folders/\" \"/\"))\r\n");

      var namespaces = ImapDataResponseConverter.FromNamespace(response);

      Assert.AreEqual(1, namespaces.PersonalNamespaces.Length);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Length);

      Assert.AreEqual(1, namespaces.SharedNamespaces.Length);
      Assert.AreEqual("Public Folders/", namespaces.SharedNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.SharedNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.SharedNamespaces[0].Extensions.Count);
    }

    [Test]
    public void TestNamespaceResponse2()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" \"/\")(\"#mh/\" \"/\" \"X-PARAM\" (\"FLAG1\" \"FLAG2\"))) NIL NIL\r\n");

      var namespaces = ImapDataResponseConverter.FromNamespace(response);

      Assert.AreEqual(2, namespaces.PersonalNamespaces.Length);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);
      Assert.AreEqual("#mh/", namespaces.PersonalNamespaces[1].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[1].HierarchyDelimiter);
      Assert.AreEqual(1, namespaces.PersonalNamespaces[1].Extensions.Count);
      Assert.IsTrue(namespaces.PersonalNamespaces[1].Extensions.ContainsKey("X-PARAM"));
      Assert.AreEqual("FLAG1", namespaces.PersonalNamespaces[1].Extensions["X-PARAM"][0]);
      Assert.AreEqual("FLAG2", namespaces.PersonalNamespaces[1].Extensions["X-PARAM"][1]);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Length);
      Assert.AreEqual(0, namespaces.SharedNamespaces.Length);
    }

    [Test]
    public void TestEnabledResponse()
    {
      var response =
        GetSingleDataResponse("* ENABLED CONDSTORE X-GOOD-IDEA\r\n");

      var caps = ImapDataResponseConverter.FromEnabled(response);

      Assert.AreEqual(2, caps.Count, "capability count");
      Assert.IsTrue(caps.Has("CONDSTORE"));
      Assert.IsTrue(caps.Has("X-GOOD-IDEA"));
    }

    [Test]
    public void TestIdResponse()
    {
      var response =
        GetSingleDataResponse("* ID (\"name\" \"Cyrus\" \"version\" \"1.5\" \"os\" \"sunos\" \"os-version\" \"5.5\" \"support-url\" \"mailto:cyrus-bugs+@andrew.cmu.edu\")\r\n");

      var paramList = ImapDataResponseConverter.FromId(response);

      Assert.AreEqual(5, paramList.Count, "param count");
      Assert.AreEqual("Cyrus", paramList["name"]);
      Assert.AreEqual("1.5", paramList["version"]);
      Assert.AreEqual("sunos", paramList["os"]);
      Assert.AreEqual("5.5", paramList["os-version"]);
      Assert.AreEqual("mailto:cyrus-bugs+@andrew.cmu.edu", paramList["support-url"]);
    }

    [Test]
    public void TestIdResponseNil()
    {
      var response =
        GetSingleDataResponse("* ID NIL\r\n");

      var paramList = ImapDataResponseConverter.FromId(response);

      Assert.AreEqual(0, paramList.Count, "param count");
    }

    [Test]
    public void TestESearchResponse1()
    {
      var response =
        GetSingleDataResponse("* ESEARCH UID COUNT 5 ALL 4:19,21,28\r\n");

      var matched = ImapDataResponseConverter.FromESearch(response);

      Assert.IsTrue(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);

      Assert.IsNotNull(matched.Count);
      Assert.AreEqual(5, matched.Count);

      var arrActual = matched.ToArray();
      var arrExpected = new long[] {4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21, 28};

      Assert.AreEqual(arrExpected.Length, arrActual.Length);

      for (var i = 0; i < arrActual.Length; i++) {
        Assert.AreEqual(arrExpected[i], arrActual[i]);
      }
    }

    [Test]
    public void TestESearchResponse2()
    {
      var response =
        GetSingleDataResponse("* ESEARCH (TAG \"a567\") UID COUNT 5 ALL 4:19,21,28\r\n");

      var matched = ImapDataResponseConverter.FromESearch(response);

      Assert.IsTrue(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);

      Assert.IsNotNull(matched.Count);
      Assert.AreEqual(5, matched.Count);
      Assert.AreEqual("a567", matched.Tag);
    }

    [Test]
    public void TestESearchResponse3()
    {
      var response =
        GetSingleDataResponse("* ESEARCH COUNT 5 ALL 1:17,21\r\n");

      var matched = ImapDataResponseConverter.FromESearch(response);

      Assert.IsFalse(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);
      Assert.IsNull(matched.Tag);

      var arrActual = matched.ToArray();
      var arrExpected = new long[] {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 21};

      Assert.AreEqual(arrExpected.Length, arrActual.Length);

      for (var i = 0; i < arrActual.Length; i++) {
        Assert.AreEqual(arrExpected[i], arrActual[i]);
      }
    }

    [Test]
    public void TestLanguageResponse1()
    {
      var response =
        GetSingleDataResponse("* LANGUAGE (DE)");

      var langTagList = ImapDataResponseConverter.FromLanguage(response);

      Assert.IsNotNull(langTagList);
      Assert.AreEqual(1, langTagList.Length);
      Assert.AreEqual("DE", langTagList[0]);
    }

    [Test]
    public void TestLanguageResponse2()
    {
      var response =
        GetSingleDataResponse("* LANGUAGE (EN DE IT i-default)");

      var langTagList = ImapDataResponseConverter.FromLanguage(response);

      Assert.IsNotNull(langTagList);
      Assert.AreEqual(4, langTagList.Length);
      Assert.AreEqual("EN", langTagList[0]);
      Assert.AreEqual("DE", langTagList[1]);
      Assert.AreEqual("IT", langTagList[2]);
      Assert.AreEqual("i-default", langTagList[3]);
    }

    [Test]
    public void TestComparatorResponse()
    {
      var response =
        GetSingleDataResponse("* COMPARATOR i;basic");

      ImapCollationAlgorithm[] matchingComparators;

      var activeComparator = ImapDataResponseConverter.FromComparator(response, out matchingComparators);

      Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), activeComparator);
      Assert.IsNull(matchingComparators);
    }

    [Test]
    public void TestComparatorResponseWithMatchingComparators()
    {
      var response =
        GetSingleDataResponse("* COMPARATOR i;unicode-casemap (i;unicode-casemap i;ascii-casemap)");

      ImapCollationAlgorithm[] matchingComparators;

      var activeComparator = ImapDataResponseConverter.FromComparator(response, out matchingComparators);

      Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, activeComparator);

      Assert.IsNotNull(matchingComparators);
      Assert.AreEqual(2, matchingComparators.Length);
      Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, matchingComparators[0]);
      Assert.AreEqual(ImapCollationAlgorithm.AsciiCasemap, matchingComparators[1]);
    }

    [Test]
    public void TestXListResponse()
    {
      var response =
        GetSingleDataResponse("* XLIST (\\HasNoChildren \\AllMail) \"/\" \"[Gmail]/&MFkweTBmMG4w4TD8MOs-\"\r\n");

      var mailboxList = ImapDataResponseConverter.FromXList(response);

      Assert.AreEqual(2, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.HasNoChildren));
      Assert.IsTrue(mailboxList.NameAttributes.Has(ImapMailboxFlag.AllMail));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("[Gmail]/すべてのメール", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    private ImapResponseReceiver receiver;
    private MemoryStream baseStream;
    private LineOrientedBufferedStream stream;
  }
}
