using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

using NUnit.Framework;
using Smdn.Net.Imap4.Tests;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol.Client {
  [TestFixture]
  public class ImapCommandSenderTest {
    [SetUp]
    public void Setup()
    {
      baseStream = new MemoryStream();
      sender = new ImapCommandSender(new LineOrientedBufferedStream(baseStream), true);
    }

    [TearDown]
    public void TearDown()
    {
      // nothing to do
    }

    private string SendCommand(ImapCommand command)
    {
      return SendCommand(command, true);
    }

    private string SendCommand(ImapCommand command, bool synchronizeLiterals)
    {
      baseStream.Seek(0, SeekOrigin.Begin);

      sender.SynchronizeLiterals = synchronizeLiterals;
      sender.Send(command);

      while (sender.CommandContinuing) {
        sender.Send();
      }

      baseStream.Seek(0, SeekOrigin.Begin);

      var reader = new StreamReader(baseStream, NetworkTransferEncoding.Transfer8Bit);

      return reader.ReadToEnd();
    }

    [Test]
    public void TestFormatString()
    {
      var command = new ImapCommand("login", "0001", new ImapString("username"), new ImapString("password"));

      Assert.AreEqual("0001 login username password\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatStringImplicitConversion()
    {
      var command = new ImapCommand("login", "0001", "username", "password");

      Assert.AreEqual("0001 login username password\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatQuotedString()
    {
      var command = new ImapCommand("select", "0001", new ImapQuotedString("INBOX.sub"));

      Assert.AreEqual("0001 select \"INBOX.sub\"\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatQuotedStringContainsDQuote()
    {
      var command = new ImapCommand("id", "0001", new ImapParenthesizedString(new ImapString[] {
        new ImapQuotedString("name"),
        new ImapQuotedString("\"hoge\""),
      }));

      Assert.AreEqual("0001 id (\"name\" \"\\\"hoge\\\"\")\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatQuotedStringContainsBackSlash()
    {
      var command = new ImapCommand("id", "0001", new ImapParenthesizedString(new ImapString[] {
        new ImapQuotedString("name"),
        new ImapQuotedString("\\(^o^)/"),
      }));

      Assert.AreEqual("0001 id (\"name\" \"\\\\(^o^)/\")\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatQuotedStringContainsDQuoteAndBackSlash()
    {
      var command = new ImapCommand("id", "0001", new ImapParenthesizedString(new ImapString[] {
        new ImapQuotedString("name"),
        new ImapQuotedString("\\(\"o\")/"),
      }));

      Assert.AreEqual("0001 id (\"name\" \"\\\\(\\\"o\\\")/\")\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatStringList()
    {
      var command = new ImapCommand("search", "0001", new ImapStringList("charset", "utf-8", "UNDELETED", "UNFLAGGED"));

      Assert.AreEqual("0001 search charset utf-8 UNDELETED UNFLAGGED\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatParenthesizedString()
    {
      var command = new ImapCommand("store", "0021",
                                            "1,3,5",
                                            "-FLAGS",
                                            new ImapParenthesizedString("\\Seen", "\\Flagged"));

      Assert.AreEqual("0021 store 1,3,5 -FLAGS (\\Seen \\Flagged)\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatNestedParenthesizedString()
    {
      var command = new ImapCommand("search", "0021",
                                            "SMALLER",
                                            "18",
                                            new ImapParenthesizedString("OR",
                                                                        new ImapParenthesizedString("UNFLAGGED", "UNSEEN"),
                                                                        new ImapParenthesizedString("UNFLAGGED", "DRAFT")
                                                                        ));

      Assert.AreEqual("0021 search SMALLER 18 (OR (UNFLAGGED UNSEEN) (UNFLAGGED DRAFT))\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatLiteral()
    {
      var command = new ImapCommand("search", "0001",
                                            "1:3",
                                            new ImapParenthesizedString("TEXT",
                                                                        new ImapLiteralString("\u0080\u00a0\u00ff", Encoding.GetEncoding("latin1")),
                                                                        "SUBJECT",
                                                                        new ImapLiteralString("abc")));

      Assert.AreEqual("0001 search 1:3 (TEXT {3}\r\n\x80\xa0\xff SUBJECT {3}\r\nabc)\r\n",
                  SendCommand(command));
    }

    [Test]
    public void TestFormatLiteralNonSynchronizing()
    {
      var command = new ImapCommand("search", "0001",
                                            "1:3",
                                            new ImapParenthesizedString("TEXT",
                                                                        new ImapLiteralString("\u0080\u00a0\u00ff", Encoding.GetEncoding("latin1")),
                                                                        "SUBJECT",
                                                                        new ImapLiteralString("abc")));

      Assert.AreEqual("0001 search 1:3 (TEXT {3+}\r\n\x80\xa0\xff SUBJECT {3+}\r\nabc)\r\n",
                      SendCommand(command, false));
    }

    private MemoryStream baseStream;
    private ImapCommandSender sender;
  }
}
