using System;
using System.IO;
using System.Net;
using NUnit.Framework;

using Smdn.Net.Imap4.Tests;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.WebClients {
  [TestFixture]
  public class ImapSessionManagerTests {
    [SetUp]
    public void Setup()
    {
      ImapSessionManager.RegisterWebRequestPrefix();
    }

    [Test]
    public void TestRegisterWebRequestPrefix()
    {
      var imapRequest = WebRequest.Create("imap://localhost/");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapRequest);

      var imapsRequest = WebRequest.Create("imaps://localhost/");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapsRequest);
    }

    [Test]
    public void TestIWebRequestCreate()
    {
      var imapServerRequest = WebRequest.Create("imap://localhost/");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapServerRequest);
      StringAssert.Contains("ImapServerWebRequest", imapServerRequest.GetType().FullName);

      var imapMailboxRequest = WebRequest.Create("imap://localhost/INBOX");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapMailboxRequest);
      StringAssert.Contains("ImapMailboxWebRequest", imapMailboxRequest.GetType().FullName);

      var imapFetchMessageRequest = WebRequest.Create("imap://localhost/INBOX/;UID=1");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapFetchMessageRequest);
      StringAssert.Contains("ImapFetchMessageWebRequest", imapFetchMessageRequest.GetType().FullName);

      var imapSearchMessageRequest = WebRequest.Create("imap://localhost/INBOX?UID 1");

      Assert.IsInstanceOfType(typeof(ImapWebRequest), imapSearchMessageRequest);
      StringAssert.Contains("ImapSearchMessageWebRequest", imapSearchMessageRequest.GetType().FullName);

      foreach (var uri in new[] {
        "imap://localhost/?",
        "imap://localhost/;uid=1",
        "imap://localhost/;uidvalidity=1",
      }) {
        try {
          WebRequest.Create(uri);
          Assert.Fail("ArgumentException not thrown: {0}", uri);
        }
        catch (ArgumentException) {
        }
      }
    }

    private void WebExceptionAssertion(WebExceptionStatus expectedStatus, WebException actual)
    {
      if (expectedStatus != actual.Status)
        throw actual;

      if (expectedStatus == WebExceptionStatus.ProtocolError) {
        Assert.IsNotNull(actual.Response);
        Assert.IsInstanceOfType(typeof(ImapWebResponse), actual.Response);

        var response = actual.Response as ImapWebResponse;

        Assert.IsTrue(response.Result.Failed);
      }
    }

    [Test]
    public void TestConnectFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var request = WebRequest.Create(string.Format("imap://{0}@{1}/", credential.UserName, server.HostPort)) as ImapWebRequest;

        request.Timeout = 500;

        server.Stop();

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ConnectFailure, ex);
        }
      }
    }

    [Test]
    public void TestConnectUseTlsIfAvailable()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                               "0000 OK done\r\n");
        // STARTTLS response
        server.EnqueueResponse("0001 OK done\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0002 OK done\r\n");
        // LOGIN response
        server.EnqueueResponse("0003 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0004 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0005 OK done\r\n");

        var defaultCallback = ImapSessionManager.CreateSslStreamCallback;

        try {
          var streamUpgraded = false;

          ImapSessionManager.CreateSslStreamCallback = delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          };

          using (var response = request.GetResponse()) {
            Assert.AreEqual(response.ResponseUri, requestUri);
          }

          Assert.IsTrue(streamUpgraded, "stream upgraded");

          server.DequeueRequest(); // CAPABILITY
          StringAssert.StartsWith("0001 STARTTLS", server.DequeueRequest());
        }
        finally {
          ImapSessionManager.CreateSslStreamCallback = defaultCallback;
        }
      }
    }

    [Test]
    public void TestConnectStartTlsFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                               "0000 OK done\r\n");
        // STARTTLS response
        server.EnqueueResponse("0001 BAD done\r\n");

        var defaultCallback = ImapSessionManager.CreateSslStreamCallback;

        try {
          var streamUpgraded = false;

          ImapSessionManager.CreateSslStreamCallback = delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          };

          try {
            using (var response = request.GetResponse()) {
            }
          }
          catch (WebException ex) {
            WebExceptionAssertion(WebExceptionStatus.SecureChannelFailure, ex);
          }

          Assert.IsFalse(streamUpgraded, "stream upgraded");

          server.DequeueRequest(); // CAPABILITY
          StringAssert.StartsWith("0001 STARTTLS", server.DequeueRequest());
        }
        finally {
          ImapSessionManager.CreateSslStreamCallback = defaultCallback;
        }
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecified()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0};auth=plain@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0002 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0003 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0};auth=plain@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 NO done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecified()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://;auth=plain@{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 OK ready\r\n");
        // NOOP response
        server.EnqueueResponse("0002 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0003 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://;auth=plain@{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 NO done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanism(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0002 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0003 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE LOGIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication(string.Empty);
    }

    private void AuthenticateExceptionWhileAuthentication(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"DIGEST-MD5", "PLAIN"};
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE DIGEST-MD5 response
        server.EnqueueResponse("+ xxx-invalid-response-xxx\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.UnknownError, ex);
          Assert.IsInstanceOfType(typeof(ImapException), ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 NO done\r\n");
        // AUTHENTICATE PLAIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0002 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0003 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0004 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE LOGIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTHENTICATE LOGIN client response
        StringAssert.Contains("AUTHENTICATE PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLogin(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0000 OK done\r\n");
        // LOGIN response
        server.EnqueueResponse("0001 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0002 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0003 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("LOGIN user", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToLoginAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToLoginAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0000 OK done\r\n");
        // LOGIN response
        server.EnqueueResponse("0001 NO done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("LOGIN user", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string authArgs)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                               "0000 OK done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.RequestCanceled, ex);
        }

        server.DequeueRequest(); // CAPABILITY
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecified()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0};auth=anonymous@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0002 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0003 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0};auth=anonymous@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 NO done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecified()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                               "0000 OK done\r\n");
        // AUTHENTICATE response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0001 NO done\r\n");
        // LOGIN response
        server.EnqueueResponse("0002 OK done\r\n");
        // NOOP response
        server.EnqueueResponse("0003 OK done\r\n");
        // LOGOUT response
        server.EnqueueResponse("* BYE logging out\r\n" +
                               "0004 OK done\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("AUTHENTICATE ANONYMOUS", server.DequeueRequest());
        server.DequeueRequest(); // AUTH ANONYMOUS client response
        StringAssert.Contains("LOGIN anonymous", server.DequeueRequest()); // anonymous credential
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0000 OK done\r\n");
        // LOGIN response
        server.EnqueueResponse("0001 NO done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPABILITY
        StringAssert.Contains("LOGIN anonymous", server.DequeueRequest()); // anonymous credential
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("imap://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY response
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                               "0000 OK done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.RequestCanceled, ex);
        }

        server.DequeueRequest(); // CAPABILITY
      }
    }
  }
}
