// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Reflection;

namespace Smdn.Net.Imap4.Protocol {
  public static class ImapDataConverter {
    public static void ToMsgAtt(ImapMessage message, ImapData msgAtt)
    {
      /*
      msg-att         = "(" (msg-att-dynamic / msg-att-static)
                         *(SP (msg-att-dynamic / msg-att-static)) ")"
      msg-att-dynamic = "FLAGS" SP "(" [flag-fetch *(SP flag-fetch)] ")"
                          ; MAY change for a message
      msg-att-static  = "ENVELOPE" SP envelope / "INTERNALDATE" SP date-time /
                        "RFC822" [".HEADER" / ".TEXT"] SP nstring /
                        "RFC822.SIZE" SP number /
                        "BODY" ["STRUCTURE"] SP body /
                        "BODY" section ["<" number ">"] SP nstring /
                        "UID" SP uniqueid
                          ; MUST NOT change for a message

      section         = "[" [section-spec] "]"
      section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
                        "TEXT"
                          ; top-level or MESSAGE/RFC822 part
      section-part    = nz-number *("." nz-number)
                          ; body part nesting
      section-spec    = section-msgtext / (section-part ["." section-text])
      section-text    = section-msgtext / "MIME"
                          ; text other than actual body part (headers, etc.)

      header-fld-name = astring
      header-list     = "(" header-fld-name *(SP header-fld-name) ")"
      */
      ExceptionIfInvalidFormat(msgAtt, ImapDataFormat.List);

      var list = msgAtt.List;

      if (list.Length % 2 != 0) // must be 'att SP value'
        throw new ImapMalformedDataException(msgAtt);

      for (var index = 0; index < list.Length; index += 2) {
        var att = ToString(list[index]);

        switch (att.ToUpperInvariant()) {
          /*
           * msg-att-dynamic
           */
          case "FLAGS":
            ExceptionIfInvalidFormat(list[index + 1], ImapDataFormat.List);
            message.Flags = ToFlagFetch(list[index + 1].List);
            break;

          /*
           * msg-att-static
           */
          case "ENVELOPE":
            message.Envelope = ToEnvelope(list[index + 1]);
            break;

          case "INTERNALDATE":
            message.InternalDate = ToDateTime(list[index + 1]);
            break;

          case "RFC822.HEADER":
            message.Rfc822Header = ToNString(list[index + 1]);
            break;

          case "RFC822.SIZE":
            message.Rfc822Size = ToNumber(list[index + 1]);
            break;

          case "RFC822.TEXT":
            message.Rfc822Text = ToNString(list[index + 1]);
            break;

          case "BODY":
          case "BODYSTRUCTURE":
            message.BodyStructure = ToBody(list[index + 1]);
            break;

          case "RFC822":
            message.Body = ToNStringAsByteArray(list[index + 1]);
            break;

          case "UID":
            message.Uid = ToUniqueId(list[index + 1]);
            break;

          /*
           * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
           * http://tools.ietf.org/html/rfc4551
           *    fetch-mod-resp      = "MODSEQ" SP "(" permsg-modsequence ")"
           *    msg-att-dynamic     =/ fetch-mod-resp
           *    permsg-modsequence  = mod-sequence-value
           *                           ;; per message mod-sequence
           */
          case "MODSEQ":
            ExceptionIfInvalidFormat(list[index + 1], ImapDataFormat.List);
            ExceptionIfTooFewData(list[index + 1].List, 1);
            message.ModSeq = ToModSequenceValue(list[index + 1].List[0]);
            break;

          default:
            if (att.StartsWith("BODY["))
              goto case "RFC822";
            //else
              //Trace.Info("unsupported msg-att: {0}", att);
            break;
        }
      }
    }

    public static ImapStatusAttributeList ToStatusAttList(ImapData[] statusAttList)
    {
      /*
      status-att      = "MESSAGES" / "RECENT" / "UIDNEXT" / "UIDVALIDITY" /
                        "UNSEEN"
      status-att-list =  status-att SP number *(SP status-att SP number)
      */

      /*
       * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
       * http://tools.ietf.org/html/rfc4551
       * 
       *    status-att-val      =/ "HIGHESTMODSEQ" SP mod-sequence-valzer
       *                           ;; extends non-terminal defined in [IMAPABNF].
       *                           ;; Value 0 denotes that the mailbox doesn't
       *                           ;; support persistent mod-sequences
       *                           ;; as described in Section 3.1.2
       */
      var attList = new ImapStatusAttributeList();

      if (statusAttList.Length % 2 != 0) // must be 'att SP value'
        throw new ImapMalformedDataException();

      for (var index = 0; index < statusAttList.Length; index += 2) {
        var att = ToString(statusAttList[index]);

        switch (att.ToUpperInvariant()) {
          case "MESSAGES":      attList.Messages      = ToNumber(statusAttList[index + 1]); break;
          case "RECENT":        attList.Recent        = ToNumber(statusAttList[index + 1]); break;
          case "UIDNEXT":       attList.UidNext       = ToNumber(statusAttList[index + 1]); break;
          case "UIDVALIDITY":   attList.UidValidity   = ToNumber(statusAttList[index + 1]); break;
          case "UNSEEN":        attList.Unseen        = ToNumber(statusAttList[index + 1]); break;
          case "HIGHESTMODSEQ": attList.HighestModSeq = ToModSequenceValzer(statusAttList[index + 1]); break;
          default:
            //Trace.Info("unsupported status-att: {0}", att);
            break;
        }
      }

      return attList;
    }

    public static ImapCapability ToCapability(ImapData capability)
    {
      /*
      capability      = ("AUTH=" auth-type) / atom
                          ; New capabilities MUST begin with "X" or be
                          ; registered with IANA as standard or
                          ; standards-track
       */
      ExceptionIfInvalidFormat(capability, ImapDataFormat.Text);

      return ImapCapability.GetKnownOrCreate(capability.Text.ToString());
    }

    public static ImapCapabilityList ToCapability(ImapData[] list)
    {
      var capabilities = new ImapCapabilityList();

      foreach (var data in list) {
        var capa = ToCapability(data);

        if (!capabilities.Has(capa))
          capabilities.Add(capa);
      }

      return capabilities;
    }

    public static ImapMessageFlagList ToFlagList(ImapData flagList)
    {
      /*
      flag            = "\Answered" / "\Flagged" / "\Deleted" /
                        "\Seen" / "\Draft" / flag-keyword / flag-extension
                          ; Does not include "\Recent"
      flag-extension  = "\" atom
                          ; Future expansion.  Client implementations
                          ; MUST accept flag-extension flags.  Server
                          ; implementations MUST NOT generate
                          ; flag-extension flags except as defined by
                          ; future standard or standards-track
                          ; revisions of this specification.
      flag-fetch      = flag / "\Recent"
      flag-keyword    = atom
      flag-list       = "(" [flag *(SP flag)] ")"
      flag-perm       = flag / "\*"
      */
      ExceptionIfInvalidFormat(flagList, ImapDataFormat.List);

      return ToFlags(flagList.List);
    }

    public static ImapMessageFlagList ToFlagFetch(ImapData[] flags)
    {
      return ToFlags(flags, ImapMessageFlag.FetchFlags);
    }

    public static ImapMessageFlagList ToFlagPerm(ImapData[] flags)
    {
      return ToFlags(flags, ImapMessageFlag.PermittedFlags);
    }

    public static ImapMessageFlagList ToFlags(ImapData[] flags)
    {
      return ToFlags(flags, null);
    }

    private static ImapMessageFlagList ToFlags(ImapData[] flags, ImapMessageFlagList permittedSystemFlags)
    {
      var flagsList = new ImapMessageFlagList();

      foreach (var flag in flags) {
        ExceptionIfInvalidFormat(flag, ImapDataFormat.Text);

        var f = ImapMessageFlag.GetKnownOrCreate(flag.Text.ToString());

#pragma warning disable 642
        if (permittedSystemFlags != null && f.IsSystemFlag && !permittedSystemFlags.Has(f))
          //Trace.Info("ignored invalid system flag: {0}", f);
          ;
        else if (!flagsList.Has(f))
          flagsList.Add(f);
#pragma warning restore 642
      }

      return flagsList;
    }

    public static ImapMailboxList ToMailboxList(ImapData[] mailboxList)
    {
      /*
      mailbox-list    = "(" [mbx-list-flags] ")" SP
                         (DQUOTE QUOTED-CHAR DQUOTE / nil) SP mailbox
      */
      ExceptionIfTooFewData(mailboxList, 3);

      var list = new ImapMailboxList(ToMbxListFlags(mailboxList[0].List),
                                     ToNString(mailboxList[1]),
                                     ToMailbox(mailboxList[2]));

      if (3 < mailboxList.Length) {
        /*
         * http://tools.ietf.org/html/rfc5258
         * RFC 5258 - Internet Message Access Protocol version 4 - LIST Command Extensions
         * 6. Formal Syntax
         * 
         *    mailbox-list =  "(" [mbx-list-flags] ")" SP
         *                (DQUOTE QUOTED-CHAR DQUOTE / nil) SP mailbox
         *                [SP mbox-list-extended]
         *                ; This is the list information pointed to by the ABNF
         *                ; item "mailbox-data", which is defined in [IMAP4]
         */
        var mboxListExtended = ToMboxListExtended(mailboxList[3]);

        if (mboxListExtended.ContainsKey("CHILDINFO"))
          list.ChildInfo = ToChildinfoExtendedItem(mboxListExtended["CHILDINFO"]);
      }

      return list;
    }

    public static Dictionary<string, ImapData> ToMboxListExtended(ImapData mboxListExtended)
    {
      /*
       * http://tools.ietf.org/html/rfc5258
       * RFC 5258 - Internet Message Access Protocol version 4 - LIST Command Extensions
       * 6. Formal Syntax
       * 
       *    mbox-list-extended =  "(" [mbox-list-extended-item
       *                *(SP mbox-list-extended-item)] ")"
       *    mbox-list-extended-item =  mbox-list-extended-item-tag SP
       *                tagged-ext-val
       *    mbox-list-extended-item-tag =  astring
       *                ; The content MUST conform to either "eitem-vendor-tag"
       *                ; or "eitem-standard-tag" ABNF productions.
       *                ; A tag registration template is described in this
       *                ; document in Section 9.5.
       * 
       *    tagged-ext-val =  tagged-ext-simple /
       *                "(" [tagged-ext-comp] ")"
       *    tagged-ext-simple =  sequence-set / number
       *    tagged-ext-comp =  astring /
       *                tagged-ext-comp *(SP tagged-ext-comp) /
       *                "(" tagged-ext-comp ")"
       *                ; Extensions that follow this general
       *                ; syntax should use nstring instead of
       *                ; astring when appropriate in the context
       *                ; of the extension.
       *                ; Note that a message set or a "number"
       *                ; can always be represented as an "atom".
       *                ; A URL should be represented as
       *                ; a "quoted" string.
       */
      ExceptionIfInvalidFormat(mboxListExtended, ImapDataFormat.List);

      var list = mboxListExtended.List;

      if (list.Length % 2 != 0)
        throw new ImapMalformedDataException("must be list of tag-value pairs", mboxListExtended);

      var taggedItems = new Dictionary<string, ImapData>();

      for (var i = 0; i < list.Length; i += 2) {
        var tag = ToAString(list[i]);

        if (taggedItems.ContainsKey(tag))
          throw new ImapMalformedDataException("tag already contained", mboxListExtended);

        taggedItems.Add(tag, list[i + 1]);
      }

      return taggedItems;
    }

    private static ImapMailboxList.ListExtendedChildInfo ToChildinfoExtendedItem(ImapData childinfoExtendedItem)
    {
      /*
       * http://tools.ietf.org/html/rfc5258
       * RFC 5258 - Internet Message Access Protocol version 4 - LIST Command Extensions
       * 6. Formal Syntax
       * 
       *    childinfo-extended-item =  "CHILDINFO" SP "("
       *                list-select-base-opt-quoted
       *                *(SP list-select-base-opt-quoted) ")"
       *                ; Extended data item (mbox-list-extended-item)
       *                ; returned when the RECURSIVEMATCH
       *                ; selection option is specified.
       *                ; Note 1: the CHILDINFO tag can be returned
       *                ; with and without surrounding quotes, as per
       *                ; mbox-list-extended-item-tag production.
       *                ; Note 2: The selection options are always returned
       *                ; quoted, unlike their specification in
       *                ; the extended LIST command.
       * 
       *    list-select-base-opt =  "SUBSCRIBED" / option-extension
       *                ; options that can be used by themselves
       */
      ExceptionIfInvalidFormat(childinfoExtendedItem, ImapDataFormat.List);
      ExceptionIfTooFewData(childinfoExtendedItem.List, 1);

      var childInfo = new ImapMailboxList.ListExtendedChildInfo();

      foreach (var extval in childinfoExtendedItem.List) {
        if (extval.Format == ImapDataFormat.Text && IsTextEqualsTo(extval, "SUBSCRIBED"))
          childInfo.Subscribed = true;
      }

      return childInfo;
    }

    public static string ToMailbox(ImapData data)
    {
      /*
      mailbox         = "INBOX" / astring
                          ; INBOX is case-insensitive.  All case variants of
                          ; INBOX (e.g., "iNbOx") MUST be interpreted as INBOX
                          ; not as an astring.  An astring which consists of
                          ; the case-insensitive sequence "I" "N" "B" "O" "X"
                          ; is considered to be INBOX and not an astring.
                          ;  Refer to section 5.1 for further
                          ; semantic details of mailbox names.
      */
      return Smdn.Formats.TextConvert.FromModifiedUTF7String(ToAString(data));
    }

    public static ImapMailboxFlagList ToMbxListFlags(ImapData[] list)
    {
      /*
      mbx-list-flags  = *(mbx-list-oflag SP) mbx-list-sflag
                        *(SP mbx-list-oflag) /
                        mbx-list-oflag *(SP mbx-list-oflag)
      mbx-list-oflag  = "\Noinferiors" / flag-extension
                          ; Other flags; multiple possible per LIST response
      mbx-list-sflag  = "\Noselect" / "\Marked" / "\Unmarked"
                          ; Selectability flags; only one per LIST response
      */

      var flags = new ImapMailboxFlagList();

      foreach (var data in list) {
        ExceptionIfInvalidFormat(data, ImapDataFormat.Text);

        var flag = ImapMailboxFlag.GetKnownOrCreate(data.Text.ToString());

        if (!flags.Has(flag))
          flags.Add(flag);
      }

      return flags;
    }

    public static ImapEnvelope ToEnvelope(ImapData envelope)
    {
      /*
      envelope        = "(" env-date SP env-subject SP env-from SP
                        env-sender SP env-reply-to SP env-to SP env-cc SP
                        env-bcc SP env-in-reply-to SP env-message-id ")"
      env-bcc         = "(" 1*address ")" / nil
      env-cc          = "(" 1*address ")" / nil
      env-date        = nstring
      env-from        = "(" 1*address ")" / nil
      env-in-reply-to = nstring
      env-message-id  = nstring
      env-reply-to    = "(" 1*address ")" / nil
      env-sender      = "(" 1*address ")" / nil
      env-subject     = nstring
      env-to          = "(" 1*address ")" / nil
      */
      ExceptionIfInvalidFormat(envelope, ImapDataFormat.List);

      if (envelope.List.Length != 10)
        throw new ImapMalformedDataException("data counts in envelope must be 10", envelope);

      var index = 0;
      var list = envelope.List;
      var envData = new object[list.Length];

      foreach (var format in new[] {
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [0] env-date
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [1] env-subject
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [2] env-from
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [3] env-sender
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [4] env-reply-to
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [5] env-to
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [6] env-cc
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [7] env-bcc
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [8] env-in-reply-to
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [9] env-message-id
      }) {
        var converted = false;

        foreach (var f in format) {
          if (list[index].Format != f)
            continue;

          if (f == ImapDataFormat.List) {
            if (list[index].List.Length <= 0)
              throw new ImapMalformedDataException(envelope);
            envData[index] = Array.ConvertAll<ImapData, ImapAddress>(list[index].List, ToAddress);
          }
          else {
            envData[index] = ToNString(list[index]);
          }

          // converted
          converted = true;
          break;
        }

        if (!converted)
          throw new ImapMalformedDataException(envelope);

        index++;
      }

      return new ImapEnvelope((string)envData[0], // env-date
                              (string)envData[1], // env-subject
                              (ImapAddress[])envData[2] ?? new ImapAddress[] {}, // env-from
                              (ImapAddress[])envData[3] ?? new ImapAddress[] {}, // env-sender
                              (ImapAddress[])envData[4] ?? new ImapAddress[] {}, // env-reply-to
                              (ImapAddress[])envData[5] ?? new ImapAddress[] {}, // env-to
                              (ImapAddress[])envData[6] ?? new ImapAddress[] {}, // env-cc
                              (ImapAddress[])envData[7] ?? new ImapAddress[] {}, // env-bcc
                              (string)envData[8], // env-in-reply-to
                              (string)envData[9] // env-message-id
                              );
    }

    public static ImapAddress ToAddress(ImapData address)
    {
      /*
      address         = "(" addr-name SP addr-adl SP addr-mailbox SP
                        addr-host ")"
      addr-adl        = nstring
                          ; Holds route from [RFC-2822] route-addr if
                          ; non-NIL
      addr-host       = nstring
                          ; NIL indicates [RFC-2822] group syntax.
                          ; Otherwise, holds [RFC-2822] domain name
      addr-mailbox    = nstring
                          ; NIL indicates end of [RFC-2822] group; if
                          ; non-NIL and addr-host is NIL, holds
                          ; [RFC-2822] group name.
                          ; Otherwise, holds [RFC-2822] local-part
                          ; after removing [RFC-2822] quoting
      addr-name       = nstring
                          ; If non-NIL, holds phrase from [RFC-2822]
                          ; mailbox after removing [RFC-2822] quoting
      */
      ExceptionIfInvalidFormat(address, ImapDataFormat.List);
      ExceptionIfTooFewData(address.List, 4);

      return new ImapAddress(ToNString(address.List[0]),
                             ToNString(address.List[1]),
                             ToNString(address.List[2]),
                             ToNString(address.List[3]));
    }

    public static ImapBodyStructureBase ToBody(ImapData body)
    {
      /*
      body            = "(" (body-type-1part / body-type-mpart) ")"
      body-extension  = nstring / number /
                         "(" body-extension *(SP body-extension) ")"
                          ; Future expansion.  Client implementations
                          ; MUST accept body-extension fields.  Server
                          ; implementations MUST NOT generate
                          ; body-extension fields except as defined by
                          ; future standard or standards-track
                          ; revisions of this specification.
      body-ext-1part  = body-fld-md5 [SP body-fld-dsp [SP body-fld-lang
                        [SP body-fld-loc *(SP body-extension)]]]
                          ; MUST NOT be returned on non-extensible
                          ; "BODY" fetch
      body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
                        [SP body-fld-loc *(SP body-extension)]]]
                          ; MUST NOT be returned on non-extensible
                          ; "BODY" fetch
      body-fields     = body-fld-param SP body-fld-id SP body-fld-desc SP
                        body-fld-enc SP body-fld-octets
      body-fld-desc   = nstring
      body-fld-dsp    = "(" string SP body-fld-param ")" / nil
      body-fld-enc    = (DQUOTE ("7BIT" / "8BIT" / "BINARY" / "BASE64"/
                        "QUOTED-PRINTABLE") DQUOTE) / string
      body-fld-id     = nstring
      body-fld-lang   = nstring / "(" string *(SP string) ")"
      body-fld-loc    = nstring
      body-fld-lines  = number
      body-fld-md5    = nstring
      body-fld-octets = number
      body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
      body-type-1part = (body-type-basic / body-type-msg / body-type-text)
                        [SP body-ext-1part]
      body-type-basic = media-basic SP body-fields
                          ; MESSAGE subtype MUST NOT be "RFC822"
      body-type-mpart = 1*body SP media-subtype
                        [SP body-ext-mpart]
      body-type-msg   = media-message SP body-fields SP envelope
                        SP body SP body-fld-lines
      body-type-text  = media-text SP body-fields SP body-fld-lines

      media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" /
                        "MESSAGE" / "VIDEO") DQUOTE) / string) SP
                        media-subtype
                          ; Defined in [MIME-IMT]
      media-message   = DQUOTE "MESSAGE" DQUOTE SP DQUOTE "RFC822" DQUOTE
                          ; Defined in [MIME-IMT]
      media-subtype   = string
                          ; Defined in [MIME-IMT]
      media-text      = DQUOTE "TEXT" DQUOTE SP media-subtype
                          ; Defined in [MIME-IMT]
       */
      ExceptionIfInvalidFormat(body, ImapDataFormat.List);
      ExceptionIfTooFewData(body.List, 1);

      if (body.List[0].Format == ImapDataFormat.Text)
        return ToBodyType1Part(body.List);
      else if (body.List[0].Format == ImapDataFormat.List)
        return ToBodyTypeMPart(body.List);
      else
        throw new ImapMalformedDataException(body);
    }

    private static ImapSinglePartBodyStructure ToBodyType1Part(ImapData[] bodyType1Part)
    {
      /*
      body            = "(" (body-type-1part / body-type-mpart) ")"
      body-type-1part = (body-type-basic / body-type-msg / body-type-text)
                        [SP body-ext-1part]
      body-type-basic = media-basic SP body-fields
                          ; MESSAGE subtype MUST NOT be "RFC822"
      body-type-msg   = media-message SP body-fields SP envelope
                        SP body SP body-fld-lines
      body-type-text  = media-text SP body-fields SP body-fld-lines

      media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" /
                        "MESSAGE" / "VIDEO") DQUOTE) / string) SP
                        media-subtype
                          ; Defined in [MIME-IMT]
      media-message   = DQUOTE "MESSAGE" DQUOTE SP DQUOTE "RFC822" DQUOTE
                          ; Defined in [MIME-IMT]
      media-text      = DQUOTE "TEXT" DQUOTE SP media-subtype
                          ; Defined in [MIME-IMT]
      */

      // basic fields
      //   type
      //     [0] type:        media-basic / media-text / media-message
      //     [1] subtype:     media-subtype
      //   body-fields
      //     [2] parameters:  body-fld-param
      //     [3] id:          body-fld-id
      //     [4] description: body-fld-desc
      //     [5] encoding:    body-fld-enc
      //     [6] size:        body-fld-octets
      //   <body-type-msg>       <body-type-text>       <body-type-basic>
      //     [7] envelope          [7] body-fld-lines     [7] (body-ext-1part)
      //     [8] body              [8] (body-ext-1part)
      //     [9] body-fld-lines
      //     [10] (body-ext-1part)
      ExceptionIfTooFewData(bodyType1Part, 7);

      var basicFields = new object[7];

      // type
      string bodyType;

      if (IsTextEqualsToCaseInsensitive(bodyType1Part[0], "MESSAGE")) {
        if (IsTextEqualsToCaseInsensitive(bodyType1Part[1], "RFC822"))
          // body-type-msg
          bodyType = "msg";
        else
          // body-type-basic
          bodyType = "basic";
      }
      else if (IsTextEqualsToCaseInsensitive(bodyType1Part[0], "TEXT")) {
         // body-type-text
        bodyType = "text";
      }
      else {
        // body-type-basic
        bodyType = "basic";
      }

      basicFields[0] = ToString(bodyType1Part[0]); // [0] type
      basicFields[1] = ToString(bodyType1Part[1]); // [1] subtype

      // body-fields
      basicFields[2] = ToBodyFldParam(bodyType1Part[2]); // [2] parameters
      basicFields[3] = ToNString(bodyType1Part[3]);      // [3] id
      basicFields[4] = ToNString(bodyType1Part[4]);      // [4] description
      basicFields[5] = ToNString(bodyType1Part[5]);      // [5] encoding
      basicFields[6] = ToNumber(bodyType1Part[6]);       // [6] size

      ImapEnvelope encapsulatedEnvelope = null;
      ImapBodyStructureBase encapsulatedBodyStructure = null;

      if (bodyType == "msg") {
        // [7] envelope
        encapsulatedEnvelope = ToEnvelope(bodyType1Part[7]);

        // [8] body
        encapsulatedBodyStructure = ToBody(bodyType1Part[8]);
      }

      long lines = 0;

      if (bodyType == "msg")
        lines = ToNumber(bodyType1Part[9]); // [9] body-fld-lines
      else if (bodyType == "text")
        lines = ToNumber(bodyType1Part[7]); // [7] body-fld-lines

      return new ImapSinglePartBodyStructure(basicFields[0] as string,
                                             basicFields[1] as string,
                                             basicFields[2] as Dictionary<string, string>,
                                             basicFields[3] as string,
                                             basicFields[4] as string,
                                             basicFields[5] as string,
                                             (long)basicFields[6],
                                             encapsulatedEnvelope,
                                             encapsulatedBodyStructure,
                                             lines);
    }

    private static Dictionary<string, string> ToBodyFldParam(ImapData bodyFldParam)
    {
      /*
      body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
      */
      if (bodyFldParam.Format == ImapDataFormat.Nil)
        return null;

      ExceptionIfInvalidFormat(bodyFldParam, ImapDataFormat.List);

      if (bodyFldParam.List.Length % 2 != 0)
        throw new ImapMalformedDataException("must be list of attribute/value pairs", bodyFldParam);

      var list = bodyFldParam.List;
      var pairs = new Dictionary<string, string>();

      for (var param = 0; param < list.Length; param += 2) {
        pairs.Add(ToString(list[param]), ToString(list[param + 1]));
      }

      return pairs;
    }

    private static ImapMultiPartBodyStructure ToBodyTypeMPart(ImapData[] bodyTypeMPart)
    {
      /*
      body            = "(" (body-type-1part / body-type-mpart) ")"
      body-type-mpart = 1*body SP media-subtype
                        [SP body-ext-mpart]
      media-subtype   = string
                          ; Defined in [MIME-IMT]
      body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
                        [SP body-fld-loc *(SP body-extension)]]]
                          ; MUST NOT be returned on non-extensible
                          ; "BODY" fetch
      */
      ExceptionIfTooFewData(bodyTypeMPart, 2);

      var index = 0;

      // body
      var bodyNested = new List<ImapBodyStructureBase>();

      for (; index < bodyTypeMPart.Length; index++) {
        if (bodyTypeMPart[index].Format == ImapDataFormat.List)
          bodyNested.Add(ToBody(bodyTypeMPart[index]));
        else if (bodyTypeMPart[index].Format == ImapDataFormat.Text)
          break;
        else
          throw new ImapMalformedDataException(bodyTypeMPart[index]);
      }

      if (bodyTypeMPart.Length <= index)
        throw new ImapMalformedDataException("missing media-subtype");

      // media-subtype
      var mediaSubType = ToString(bodyTypeMPart[index]);

      // TODO: parse body-ext-mpart

      return new ImapMultiPartBodyStructure(bodyNested.ToArray(), mediaSubType);
    }

    public static ImapSequenceSet ToUidSet(ImapData sequenceSet)
    {
      return ToSequenceSet(true, sequenceSet);
    }

    public static ImapSequenceSet ToSequenceSet(ImapData sequenceSet)
    {
      return ToSequenceSet(false, sequenceSet);
    }

    public static ImapSequenceSet ToSequenceSet(bool uid, ImapData sequenceSet)
    {
      // seq-number      = nz-number / "*"
      // seq-range       = seq-number ":" seq-number
      // sequence-set    = (seq-number / seq-range) *("," sequence-set)
      ExceptionIfInvalidFormat(sequenceSet, ImapDataFormat.Text);

      var sets = sequenceSet.Text.ToString().Split(new char[] {','}, StringSplitOptions.None);
      var numbers = new List<long>();
      ImapSequenceSet result = null;

      for (var i = 0;;) {
        ImapSequenceSet current = null;

        var sep = sets[i].IndexOf(":");

        if (0 <= sep) {
          var from = sets[i].Substring(0, sep);
          var to = sets[i].Substring(sep + 1);

          if (from == "*") // *:number
            current = ImapSequenceSet.CreateToSet(uid, ToNonZeroNumber(to));
          else if (to == "*") // number:*
            current = ImapSequenceSet.CreateFromSet(uid, ToNonZeroNumber(from));
          else // number:number
            current = ImapSequenceSet.CreateRangeSet(uid, ToNonZeroNumber(from), ToNonZeroNumber(to));
        }
        else if (sets[i] == "*") {
          // *
          current = ImapSequenceSet.CreateAllSet(uid);
        }
        else {
          // number
          numbers.Add(ToNonZeroNumber(sets[i]));
        }

        i++;

        if (current != null || i == sets.Length) {
          if (0 < numbers.Count) {
            result = ImapSequenceSet.Combine(result, ImapSequenceSet.CreateSet(uid, numbers.ToArray()));
            numbers.Clear();
          }

          if (current != null)
            result = ImapSequenceSet.Combine(result, current);
        }

        if (i == sets.Length)
          break;
      }

      return result;
    }

    public static ImapThreadList ToThreadList(bool uid, ImapData[] threadLists)
    {
      var threads = new List<ImapThreadList>();

      foreach (var threadList in threadLists) {
        threads.AddRange(ToThreadList(uid, threadList));
      }

      return ImapThreadList.CreateRootedList(uid, threads.ToArray());
    }

    public static ImapThreadList[] ToThreadList(bool uid, ImapData threadList)
    {
      /*
         thread-list     = "(" (thread-members / thread-nested) ")"
         thread-members  = nz-number *(SP nz-number) [SP thread-nested]
         thread-nested   = 2*thread-list
      */
      ExceptionIfInvalidFormat(threadList, ImapDataFormat.List);
      ExceptionIfTooFewData(threadList.List, 1);

      if (threadList.List[0].Format == ImapDataFormat.Text)
        return new ImapThreadList[] {ToThreadMembers(uid, threadList.List)};
      else if (threadList.List[0].Format == ImapDataFormat.List)
        return ToThreadNested(uid, threadList.List, 0);
      else
        throw new ImapMalformedDataException(threadList);
    }

    public static ImapThreadList ToThreadMembers(bool uid, ImapData[] threadMembers)
    {
      /*
         thread-list     = "(" (thread-members / thread-nested) ")"
         thread-members  = nz-number *(SP nz-number) [SP thread-nested]
         thread-nested   = 2*thread-list
      */
      var index = 0;

      // thread-members
      var members = new Stack<long>();

      for (; index < threadMembers.Length; index++) {
        if (threadMembers[index].Format == ImapDataFormat.List)
          break;

        ExceptionIfInvalidFormat(threadMembers[index], ImapDataFormat.Text);

        members.Push(ToNonZeroNumber(threadMembers[index]));
      }

      var threadList = (index < threadMembers.Length) ?
        new ImapThreadList(uid, members.Pop(), ToThreadNested(uid, threadMembers, index)) : // thread-nested
        new ImapThreadList(uid, members.Pop());

      while (0 < members.Count) {
        threadList = new ImapThreadList(uid, members.Pop(), threadList);
      }

      return threadList;
    }

    public static ImapThreadList[] ToThreadNested(bool uid, ImapData[] list)
    {
      return ToThreadNested(uid, list, 0);
    }

    private static ImapThreadList[] ToThreadNested(bool uid, ImapData[] threadNested, int index)
    {
      /*
         thread-list     = "(" (thread-members / thread-nested) ")"
         thread-members  = nz-number *(SP nz-number) [SP thread-nested]
         thread-nested   = 2*thread-list
      */
      if (threadNested.Length - index < 2)
        throw new ImapMalformedDataException("at least 2 thread-lists is required");

      var nested = new List<ImapThreadList>();

      for (var i = 0; index < threadNested.Length; i++, index++) {
        nested.AddRange(ToThreadList(uid, threadNested[index]));
      }

      return nested.ToArray();
    }

    /*
     * RFC 2971 IMAP4 ID extension
     * http://tools.ietf.org/html/rfc2971
     */
    public static Dictionary<string, string> ToIdParamsList(ImapData idParamsList)
    {
      // 4. Formal Syntax
      //      id_params_list ::= "(" #(string SPACE nstring) ")" / nil
      //          ;; list of field value pairs
      if (idParamsList.Format == ImapDataFormat.Nil)
        return new Dictionary<string, string>();

      ExceptionIfInvalidFormat(idParamsList, ImapDataFormat.List);

      var list = idParamsList.List;

      if (list.Length % 2 != 0)
        throw new ImapMalformedDataException("must be list of field value pairs", idParamsList);

      var paramsList = new Dictionary<string, string>();

      for (var i = 0; i < list.Length; i += 2) {
        var field = ToString(list[i]);

        if (paramsList.ContainsKey(field))
          throw new ImapMalformedDataException("field already contained", idParamsList);

        paramsList.Add(field, ToNString(list[i + 1]));
      }

      return paramsList;
    }

    /*
     * RFC 2342 IMAP4 Namespace
     * http://tools.ietf.org/html/rfc2342
     */
    public static ImapNamespaceDesc[] ToNamespace(ImapData @namespace)
    {
      /*
       Namespace = nil / "(" 1*( "(" string SP  (<"> QUOTED_CHAR <"> /
          nil) *(Namespace_Response_Extension) ")" ) ")"
       Namespace_Response_Extension = SP string SP "(" string *(SP string)
          ")"

       string = <string>
          ; <string> as defined in [RFC-2060]
          ; Note that  the namespace prefix is to a mailbox and following
          ; IMAP4 convention, any international string in the NAMESPACE
          ; response MUST be of modified UTF-7 format as described in
          ;  [RFC-2060].
      */
      if (@namespace.Format == ImapDataFormat.Nil)
        return new ImapNamespaceDesc[] {};

      ExceptionIfInvalidFormat(@namespace, ImapDataFormat.List);
      ExceptionIfTooFewData(@namespace.List, 1);

      var list = @namespace.List;

      var namespaces = new List<ImapNamespaceDesc>();

      foreach (var nslist in list) {
        ExceptionIfInvalidFormat(nslist, ImapDataFormat.List);
        ExceptionIfTooFewData(nslist.List, 1);

        var prefix = Smdn.Formats.TextConvert.FromModifiedUTF7String(ToString(nslist.List[0]));

        if (2 <= nslist.List.Length) {
          var delimiter = ToString(nslist.List[1]);

          if (3 <= nslist.List.Length)
            namespaces.Add(new ImapNamespaceDesc(prefix, delimiter, ToNamespaceResponseExtension(nslist.List, 2, nslist.List.Length - 2)));
          else
            namespaces.Add(new ImapNamespaceDesc(prefix, delimiter));
        }
        else {
          namespaces.Add(new ImapNamespaceDesc(prefix));
        }
      }

      return namespaces.ToArray();
    }

    private static Dictionary<string, string[]> ToNamespaceResponseExtension(ImapData[] extensions, int index, int count)
    {
      if (count % 2 != 0)
        throw new ImapMalformedDataException("must be list of name value pairs");

      var exts = new Dictionary<string, string[]>();

      for (var i = 0; i < count; i += 2, index += 2) {
        ExceptionIfInvalidFormat(extensions[index + 1], ImapDataFormat.List);

        exts.Add(ToString(extensions[index]), Array.ConvertAll<ImapData, string>(extensions[index + 1].List, ToString));
      }

      return exts;
    }

    /*
     * RFC 5255 - Internet Message Access Protocol Internationalization
     * http://tools.ietf.org/html/rfc5255
     */
    public static string ToLangTagQuoted(ImapData langTagQuoted)
    {
      /*
       *     language-data     = "LANGUAGE" SP "(" lang-tag-quoted *(SP
       *                       lang-tag-quoted) ")"
       *     lang-range-quoted = astring
       *         ; Once any literal wrapper or quoting is removed, this
       *         ; follows the language-range rule in [RFC4647]
       *     lang-tag-quoted   = astring
       *         ; Once any literal wrapper or quoting is removed, this follows
       *         ; the Language-Tag rule in [RFC4646]
       */
      return ToAString(langTagQuoted);
    }

    public static ImapCollationAlgorithm ToCompSelQuoted(ImapData compSelQuoted)
    {
      /*
       *     comp-sel-quoted   = astring
       *         ; Once any literal wrapper or quoting is removed, this
       *         ; follows the collation-selected rule from [RFC4790]
       */
      return ImapCollationAlgorithm.GetKnownOrCreate(ToAString(compSelQuoted));
    }

    public static ImapCollationAlgorithm ToCompOrderQuoted(ImapData compOrderQuoted)
    {
      /*
       *     comp-order-quoted = astring
       *         ; Once any literal wrapper or quoting is removed, this
       *         ; follows the collation-order rule from [RFC4790]
       */
      return ImapCollationAlgorithm.GetKnownOrCreate(ToAString(compOrderQuoted));
    }

    /*
     * RFC 4466 - Collected Extensions to IMAP4 ABNF
     * http://tools.ietf.org/html/rfc4466
     */
    public static string ToSearchCorrelator(ImapData searchCorrelator)
    {
      /*
       *    search-correlator  = SP "(" "TAG" SP tag-string ")"
       */
      ExceptionIfInvalidFormat(searchCorrelator, ImapDataFormat.List);
      ExceptionIfTooFewData(searchCorrelator.List, 2);

      if (!IsTextEqualsToCaseInsensitive(searchCorrelator.List[0], "TAG"))
        throw new ImapMalformedDataException(searchCorrelator.List[0]);

      return ToString(searchCorrelator.List[1]);
    }

    public static ImapMatchedSequenceSet ToSearchReturnData(ImapData[] searchReturnData, string tag, bool uid, int startIndex)
    {
      /*
       *    search-return-data = search-modifier-name SP search-return-value
       *                         ;; Note that not every SEARCH return option
       *                         ;; is required to have the corresponding
       *                         ;; ESEARCH return data.
       *    search-return-value = tagged-ext-val
       *                         ;; Data for the returned search option.
       *                         ;; A single "nz-number"/"number" value
       *                         ;; can be returned as an atom (i.e., without
       *                         ;; quoting).  A sequence-set can be returned
       *                         ;; as an atom as well.
       *    search-modifier-name = tagged-ext-label
       */
      if ((searchReturnData.Length - startIndex) % 2 != 0)
        throw new ImapMalformedDataException("must be list of name/value pairs");

      long? min = null;
      long? max = null;
      long? count = null;
      ulong? modSeq = null;
      ImapSequenceSet all = null;

      for (var index = startIndex; index < searchReturnData.Length; index += 2) {
        switch (ToString(searchReturnData[index]).ToUpperInvariant()) {
          /*
           * RFC 4731 - IMAP4 Extension to SEARCH Command for Controlling What Kind of Information Is Returned
           * http://tools.ietf.org/html/rfc4731
           *      search-return-data = "MIN" SP nz-number /
           *                           "MAX" SP nz-number /
           *                           "ALL" SP sequence-set /
           *                           "COUNT" SP number
           *                           ;; conforms to the generic
           *                           ;; search-return-data syntax defined
           *                           ;; in [IMAPABNF]
           * 
           *      When the CONDSTORE [CONDSTORE] IMAP extension is also supported,
           *      the ABNF is updated as follows:
           * 
           *      search-return-data =/ "MODSEQ" SP mod-sequence-value
           *                           ;; mod-sequence-value is defined
           *                           ;; in [CONDSTORE]
           */
          case "MIN":
            min = ToNonZeroNumber(searchReturnData[index + 1]);
            break;

          case "MAX":
            max = ToNonZeroNumber(searchReturnData[index + 1]);
            break;

          case "ALL":
            all = ToSequenceSet(uid, searchReturnData[index + 1]);
            break;

          case "COUNT":
            count = ToNonZeroNumber(searchReturnData[index + 1]);
            break;

          case "MODSEQ":
            modSeq = ToModSequenceValue(searchReturnData[index + 1]);
            break;
        }
      }

      var returnData = (all == null)
        ? ImapMatchedSequenceSet.CreateEmpty(uid)
        : new ImapMatchedSequenceSet(all);

      returnData.Tag = tag;
      returnData.Max = max;
      returnData.Min = min;
      returnData.Count = count;
      returnData.HighestModSeq = modSeq;

      return returnData;
    }

    public static DateTimeOffset ToDateTime(ImapData dateTime)
    {
      ExceptionIfInvalidFormat(dateTime, ImapDataFormat.Text);

      return ImapDateTimeFormat.FromDateTimeString(dateTime.Text.ToString());
    }

    public static string ToString(ImapData @string)
    {
      ExceptionIfInvalidFormat(@string, ImapDataFormat.Text);

      return @string.Text.ToString(); // XXX
    }

    public static string ToAString(ImapData astring)
    {
      /*
      astring         = 1*ASTRING-CHAR / string
      ASTRING-CHAR   = ATOM-CHAR / resp-specials
      atom            = 1*ATOM-CHAR
      ATOM-CHAR       = <any CHAR except atom-specials>
      atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
                        quoted-specials / resp-specials
      */
      ExceptionIfInvalidFormat(astring, ImapDataFormat.Text);

      return astring.Text.ToString(); // XXX
    }

    public static string ToNString(ImapData nstring)
    {
      /*
      nstring         = string / nil
      */
      if (nstring.Format == ImapDataFormat.Text)
        return nstring.Text.ToString();
      else if (nstring.Format == ImapDataFormat.Nil)
        return null;
      else
        throw new ImapMalformedDataException("nstring must be nil or text", nstring);
    }

    public static byte[] ToNStringAsByteArray(ImapData nstring)
    {
      /*
      nstring         = string / nil
      */
      if (nstring.Format == ImapDataFormat.Text)
        return nstring.Text.ByteArray;
      else if (nstring.Format == ImapDataFormat.Nil)
        return null;
      else
        throw new ImapMalformedDataException("nstring must be nil or text", nstring);
    }

    [CLSCompliant(false)]
    public static ulong ToModSequenceValue(ImapData modSequenceValue)
    {
      /*
       mod-sequence-value  = 1*DIGIT
                              ;; Positive unsigned 64-bit integer
                              ;; (mod-sequence)
                              ;; (1 <= n < 18,446,744,073,709,551,615)
      */
      var val = ToModSequenceValzer(modSequenceValue);

      if (val == 0)
        throw new ImapMalformedDataException(string.Format("mod-sequence-value must be non-zero positive number, but was {0}", val));

      return val;
    }

    [CLSCompliant(false)]
    public static ulong ToModSequenceValzer(ImapData modSequenceValzer)
    {
      /*
        mod-sequence-valzer = "0" / mod-sequence-value
      */
      ExceptionIfInvalidFormat(modSequenceValzer, ImapDataFormat.Text);

      var num = modSequenceValzer.Text.ByteArray;
      ulong val = 0;

      for (var index = 0; index < num.Length; index++) {
        if (0x30 <= num[index] && num[index] <= 0x39)
          val = (val * 10) + (ulong)(num[index] - 0x30);
        else
          throw new ImapMalformedDataException(modSequenceValzer);
      }

      return val;
    }

    public static long ToUniqueId(ImapData uniqueId)
    {
      /*
      uniqueid        = nz-number
                          ; Strictly ascending
      */
      return ToNonZeroNumber(uniqueId);
    }

    public static long ToNonZeroNumber(ImapData nzNumber)
    {
      var val = ToNumber(nzNumber);

      if (val == 0)
        throw new ImapMalformedDataException(string.Format("nz-number must be non-zero positive number, but was {0}", val));

      return val;
    }

    internal static long ToNonZeroNumber(string nzNumber)
    {
      /*
      nz-number       = digit-nz *DIGIT
                          ; Non-zero unsigned 32-bit integer
                          ; (0 < n < 4,294,967,296)
      */
      var val = ToNumber(nzNumber);

      if (val == 0)
        throw new ImapMalformedDataException(string.Format("nz-number must be non-zero positive number, but was {0}", val));

      return val;
    }

    public static long ToNumber(ImapData number)
    {
      ExceptionIfInvalidFormat(number, ImapDataFormat.Text);

      var num = number.Text.ByteArray;
      long val = 0;

      for (var index = 0; index < num.Length; index++) {
        if (0x30 <= num[index] && num[index] <= 0x39)
          val = (val * 10) + (num[index] - 0x30);
        else
          throw new ImapMalformedDataException(number);
      }

      return val;
    }

    internal static long ToNumber(string number)
    {
      /*
      number          = 1*DIGIT
                          ; Unsigned 32-bit integer
                          ; (0 <= n < 4,294,967,296)
      */
      try {
        var val = long.Parse(number);

        if (val < 0)
          throw new ImapMalformedDataException(string.Format("number must be zero or positive number, but was {0}", val));

        return val;
      }
      catch (FormatException) {
        throw new ImapMalformedDataException(number);
      }
    }

    public static Uri ToUri(ImapData uri)
    {
      ExceptionIfInvalidFormat(uri, ImapDataFormat.Text);

      try {
        return new Uri(uri.Text.ToString());
      }
      catch {
        throw new ImapMalformedDataException(uri);
      }
    }

    /*
     * utility methods
     */

    private static bool IsTextEqualsTo(ImapData data, string text)
    {
      ExceptionIfInvalidFormat(data, ImapDataFormat.Text);

      return data.Text.Equals(text);
    }

    private static bool IsTextEqualsToCaseInsensitive(ImapData data, string text)
    {
      ExceptionIfInvalidFormat(data, ImapDataFormat.Text);

      return data.Text.EqualsIgnoreCase(text);
    }

    private static void ExceptionIfInvalidFormat(ImapData data, ImapDataFormat requiredFormat)
    {
      if (data.Format != requiredFormat)
        throw new ImapMalformedDataException(string.Format("invalid data format; required format is {0} but was {1}", requiredFormat, data.Format), data);
    }

    private static void ExceptionIfTooFewData(ImapData[] data, int requiredCount)
    {
      if (data.Length < requiredCount)
        throw new ImapMalformedDataException(string.Format("too few data counts; required is {0} but was {1}", requiredCount, data.Length));
    }
  }
}