// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Smdn.Formats;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol {
  public abstract class ImapReceiver {
    private LineOrientedBufferedStream stream;

    protected ImapReceiver(LineOrientedBufferedStream stream)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      this.stream = stream;
    }

    protected ByteString Receive(int requiredDataLengthAtLeast)
    {
      ByteString line;

      if (0 == stream.ReadLine(requiredDataLengthAtLeast, out line))
        throw new ImapConnectionException("read failed");

      return line;
    }

    private static readonly ByteString NIL = new ByteString("NIL");

    protected bool ParseData(DataReader dataReader, out ImapData[] data, ref int requiredBytes)
    {
      data = null;

      var parsed = new List<ImapData>();
      var builder = new ByteStringBuilder();

      for (;;) {
        if (dataReader.IsEndOfData)
          break;

        var b = dataReader.PeekByte();

        if (b == Octets.SP || b == Octets.CR) {
          dataReader.Read();
          continue;
        }

        if (b == Octets.LF) {
          dataReader.Read();
          break;
        }

        ImapData d;

        builder.Length = 0;

        if (b == ImapOctets.DQuote) {
          if (ParseQuotedString(dataReader, builder, false))
            d = ImapData.CreateTextData(builder.ToByteString());
          else
            return false;
        }
        else if (b == ImapOctets.OpenParen) {
          if (ParseParenthesizedData(dataReader, builder, ref requiredBytes)) {
            var r = new DataReader(builder);

            ImapData[] list = null;

            if (!ParseData(r, out list, ref requiredBytes))
              return false;

            d = ImapData.CreateListData(list);
          }
          else {
            return false;
          }
        }
        else if (b == ImapOctets.OpenBrace) {
          if (ParseLiteralString(dataReader, builder, false, ref requiredBytes))
            d = ImapData.CreateTextData(builder.ToByteString());
          else
            return false;
        }
        else {
          if (ParseNonQuotedText(dataReader, builder)) {
            var text = builder.ToByteString();

            if (NIL.EqualsIgnoreCase(text))
              d = ImapData.CreateNilData();
            else
              d = ImapData.CreateTextData(text);
          }
          else {
            return false;
          }
        }

        parsed.Add(d);
      }

      data = parsed.ToArray();

      return true;
    }

    private bool ParseQuotedString(DataReader reader, ByteStringBuilder builder, bool keepDelimiter)
    {
      // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
      // QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> / 
      //                   "\" quoted-specials
      // TEXT-CHAR       = <any CHAR except CR and LF>
      // CHAR            = <any 7-bit US-ASCII character except NUL,
      //                   0x01 - 0x7f>
      // quoted-specials = DQUOTE / "\"
      // date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year

      var b = reader.ReadByte(); // DQUOTE

      if (keepDelimiter)
        builder.Append(b);

      var escape = false;

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        b = reader.ReadByte();

        escape = (b == ImapOctets.BackSlash);

        if (escape) {
          if (keepDelimiter)
            builder.Append(b);
          b = reader.ReadByte(); // quoted-specials
        }

        if (b == Octets.CR || b == Octets.LF || (!escape && b == ImapOctets.DQuote)) {
          if (keepDelimiter)
            builder.Append(b);

          return true; // success
        }
        else { // QUOTED-CHAR
          builder.Append(b);
        }
      }
    }

    private bool ParseLiteralString(DataReader reader, ByteStringBuilder builder, bool keepDelimiter, ref int requiredBytes)
    {
      // literal         = "{" number "}" CRLF *CHAR8
      //                ; Number represents the number of CHAR8s
      // CHAR8           = %x01-ff ; any OCTET except NUL, %x00
      // number          = 1*DIGIT
      //                     ; Unsigned 32-bit integer
      //                     ; (0 <= n < 4,294,967,296)

      reader.Read(); // "{"

      var number = new byte[10];
      var numberLength = 0;
      long octet = 0;

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        var b = reader.ReadByte();

        if (Octets.IsDecimalNumber(b) && numberLength < number.Length) {
          number[numberLength++] = b; // number
          octet = (octet * 10) + (b - 0x30/*'0'*/);
        }
        else if (b == ImapOctets.CloseBrace) {
          break; // "}"
        }
        else {
          throw new ImapMalformedDataException("malformed literal (invalid number)");
        }
      }

      if (keepDelimiter) {
        builder.Append(ImapOctets.OpenBrace);
        builder.Append(number, 0, numberLength);
        builder.Append(ImapOctets.CloseBrace);
      }

      // CRLF
      if (reader.IsEndOfData)
        return false; // abort
      else if (Octets.CR != reader.ReadByte())
        throw new ImapMalformedDataException("malformed literal (missing CR)");

      if (keepDelimiter)
        builder.Append(Octets.CR);

      if (reader.IsEndOfData)
        return false; // abort
      else if (Octets.LF != reader.ReadByte())
        throw new ImapMalformedDataException("malformed literal (missing LF)");

      if (keepDelimiter)
        builder.Append(Octets.LF);

      // *CHAR8
      if (octet <= reader.Length - reader.Position) {
        reader.ReadBytes((int)octet, builder); // XXX: long

        return true; // success
      }
      else {
        var remain = octet - (reader.Length - reader.Position);

        if (requiredBytes < remain)
          requiredBytes = (int)remain; // XXX: long

        return false; // abort
      }
    }

    private bool ParseParenthesizedData(DataReader reader, ByteStringBuilder builder, ref int requiredBytes)
    {
      var nested = 0;

      reader.Read(); // "("

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        var b = reader.PeekByte();

        if (b == ImapOctets.DQuote) {
          // quoted
          if (ParseQuotedString(reader, builder, true))
            continue;
          else
            return false;
        }
        else if (b == ImapOctets.OpenBrace) {
          // literal
          if (ParseLiteralString(reader, builder, true, ref requiredBytes))
            continue;
          else
            return false;
        }

        b = reader.ReadByte();

        if (b == ImapOctets.OpenParen) {
          nested++;
        }
        else if (b == ImapOctets.CloseParen) {
          nested--;
          if (nested == -1) { // not nested
            return true; // success
          }
        }

        builder.Append(b);
      }
    }

    private bool ParseNonQuotedText(DataReader reader, ByteStringBuilder builder)
    {
      // atom            = 1*ATOM-CHAR
      // ATOM-CHAR       = <any CHAR except atom-specials>
      // atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
      //                   quoted-specials / resp-specials
      // number          = 1*DIGIT
      // list-wildcards  = "%" / "*"
      // quoted-specials = DQUOTE / "\"
      // resp-specials   = "]"

      // section         = "[" [section-spec] "]"
      // section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
      //                   "TEXT"
      //                     ; top-level or MESSAGE/RFC822 part
      // section-part    = nz-number *("." nz-number)
      //                     ; body part nesting
      // section-spec    = section-msgtext / (section-part ["." section-text])
      // section-text    = section-msgtext / "MIME"
      //                     ; text other than actual body part (headers, etc.)
      var section = false;

      for (;;) {
        if (reader.IsEndOfData || reader.IsNextChar(Octets.CR) || reader.IsNextChar(Octets.LF))
          break;

        var b = reader.ReadByte();

        if (!section && b == Octets.SP) {
          break;
        }
        else {
          builder.Append(b);

          if (b == ImapOctets.OpenBracket)
            section = true;
          else if (b == ImapOctets.CloseBracket)
            section = false;
        }
      }

      return true; // success
    }

#region "class DataReader"
    protected class DataReader {
      public int Position {
        get { return pos - offset; }
      }

      public int Length {
        get { return len - offset; }
      }

      public bool IsEndOfData {
        get { return len <= pos; }
      }

      public DataReader(ByteString s)
        : this(s, 0)
      {
      }

      public DataReader(ByteString s, int offset)
      {
        if (s == null)
          throw new ArgumentNullException("s");

        this.source = s.ByteArray;
        this.len = source.Length;
        this.pos = offset;
        this.offset = offset;
      }

      public DataReader(ByteStringBuilder b)
      {
        if (b == null)
          throw new ArgumentNullException("b");

        var seg = b.GetSegment();

        this.source = seg.Array;
        this.len = seg.Count;
        this.pos = seg.Offset;
        this.offset = seg.Offset;
      }

      public int Peek()
      {
        if (len <= pos)
          return -1;
        else
          return (int)source[pos];
      }

      public byte PeekByte()
      {
        if (len <= pos)
          throw new EndOfStreamException();
        else
          return source[pos];
      }

      public int Read()
      {
        if (len <= pos)
          return -1;
        else
          return (int)source[pos++];
      }

      public byte ReadByte()
      {
        if (len <= pos)
          throw new EndOfStreamException();
        else
          return source[pos++];
      }

      public bool IsNextChar(byte ch)
      {
        if (len <= pos)
          return false;
        else
          return source[pos] == ch;
      }

      public byte[] ReadBytes(int count)
      {
        var lengthToRead = (len <= pos + count) ? len - pos : count;
        var ret = new byte[lengthToRead];

        Buffer.BlockCopy(source, pos, ret, 0, count);

        pos += lengthToRead;

        return ret;
      }

      public int ReadBytes(int count, ByteStringBuilder builder)
      {
        var lengthToRead = (len <= pos + count) ? len - pos : count;

        builder.Append(source, pos, lengthToRead);

        pos += lengthToRead;

        return lengthToRead;
      }

      public ByteString ReadToEnd()
      {
        var ret = new ByteString(source, pos, len - pos);

        pos = len;

        return ret;
      }

      private byte[] source;
      private readonly int len;
      private readonly int offset;
      private int pos;
    }
#endregion
  }
}
