// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;

using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol {
  public abstract class ImapSender {
    public virtual bool SynchronizeLiterals {
      get { return false; }
      set {}
    }

    internal protected int UnsentFragments {
      get { return fragments.Count; }
    }

    protected LineOrientedBufferedStream Stream {
      get { return stream; }
    }

    protected ImapSender(LineOrientedBufferedStream stream)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      this.stream = stream;
    }

    public void Send()
    {
      for (;;) {
        if (fragments.Count == 0)
          return;

        var fragment = fragments.Dequeue();

        if (fragment is BytesFragment)
          stream.WriteToCrLf((fragment as BytesFragment).Bytes);
        else if (fragment is StreamFragment)
          stream.Write((fragment as StreamFragment).Stream);
        else if (fragment is WaitContinuationFragment)
          return;
      }
    }

    protected void ClearUnsent()
    {
      fragments.Clear();
    }

    protected void Enqueue(byte[] bytes)
    {
      fragments.Enqueue(new BytesFragment(bytes));
    }

    protected void Enqueue(IEnumerable<ImapString> args)
    {
      var delim = false;

      foreach (var arg in args) {
        if (delim)
          fragments.Enqueue(new BytesFragment(new byte[] {0x20})); // SP
        else
          delim = true;

        var str = (arg is ImapCombinableDataItem) ? (arg as ImapCombinableDataItem).GetCombined() : arg;

        if (str is ImapStringList) {
          var parenthesize = (str is ImapParenthesizedString);

          if (parenthesize)
            fragments.Enqueue(new BytesFragment(new byte[] {0x28})); // '('

          Enqueue(str as ImapStringList);

          if (parenthesize)
            fragments.Enqueue(new BytesFragment(new byte[] {0x29})); // ')'
        }
        else {
          if (str is ImapQuotedString) {
            // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
            var quoted = str as ImapQuotedString;

            fragments.Enqueue(new BytesFragment(string.Format("\"{0}\"", quoted.Escaped)));
          }
          else if (str is IImapLiteralString) {
            // literal         = "{" number "}" CRLF *CHAR8
            var literal = str as IImapLiteralString;
            var stream = literal.GetLiteralStream();

            if ((literal.SyncMode == ImapLiteralSyncMode.Synchronous) || SynchronizeLiterals) {
              // "{" number "}" CRLF
              fragments.Enqueue(new BytesFragment(string.Format("{{{0}}}\x0d\x0a", stream.Length)));
              fragments.Enqueue(new WaitContinuationFragment());
            }
            else {
              /*
               * IMAP4 non-synchronizing literals
               * http://tools.ietf.org/html/rfc2088
               */
              // "{" number "}" CRLF
              fragments.Enqueue(new BytesFragment(string.Format("{{{0}+}}\x0d\x0a", stream.Length)));
            }

            // *CHAR8
            fragments.Enqueue(new StreamFragment(stream));
          }
          else if (str is ImapPreformattedString) {
            fragments.Enqueue(new BytesFragment((str as ImapPreformattedString).Octets));
          }
          else {
            // atom            = 1*ATOM-CHAR
            fragments.Enqueue(new BytesFragment((string)str));
          }
        }
      }
    }

    private /*readonly*/ LineOrientedBufferedStream stream;
    private /*readonly*/ Queue<IFragment> fragments = new Queue<IFragment>();

    private interface IFragment {
    }

    private class WaitContinuationFragment : IFragment {
    }

    private class BytesFragment : IFragment {
      public byte[] Bytes;

      public BytesFragment(string str)
        : this(NetworkTransferEncoding.Transfer7Bit.GetBytes(str))
      {
      }

      public BytesFragment(byte[] bytes)
      {
        this.Bytes = bytes;
      }
    }

    private class StreamFragment : IFragment {
      public Stream Stream;

      public StreamFragment(Stream stream)
      {
        this.Stream = stream;
      }
    }
  }
}