// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.IO;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;

namespace Smdn.Net.MessageAccessProtocols {
  public abstract class ConnectionBase : IDisposable {
    public const int DefaultTimeoutMilliseconds = 16000;

    private static int connectionIdCount = 0;

    public bool IsDisposed {
      get { return disposed; }
    }

    public int Id {
      get { return id; }
    }

    public string Host {
      get { CheckDisposed(); return host; }
    }

    public int Port {
      get { CheckDisposed(); return remoteEndPoint.Port; }
    }

    public IPEndPoint RemoteEndPoint {
      get { CheckConnected(); return remoteEndPoint; }
    }

    public IPEndPoint LocalEndPoint {
      get { CheckConnected(); return localEndPoint; }
    }

    public int SendTimeout {
      get { return Socket.SendTimeout; }
      set { Socket.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { return Socket.ReceiveTimeout; }
      set { Socket.ReceiveTimeout = value; }
    }

    public bool IsSecurePortConnection {
      get { CheckConnected(); return isSecurePortConnection; }
    }

    public bool IsSecureConnection {
      get { return RawStream is SslStream; }
    }

    protected virtual Stream RawStream {
      get
      {
        CheckConnected();

        if (stream.InnerStream is Trace.TracingStream)
          return (stream.InnerStream as Trace.TracingStream).InnerStream;
        else
          return stream.InnerStream;
      }
    }

    protected LineOrientedBufferedStream Stream {
      get { CheckConnected(); return stream; }
    }

    protected Socket Socket {
      get { CheckConnected(); return client.Client; }
    }

    protected ConnectionBase()
    {
      this.id = Interlocked.Increment(ref connectionIdCount);

      Interlocked.CompareExchange(ref connectionIdCount, 0, int.MaxValue);
    }

    ~ConnectionBase()
    {
      Dispose(false);
    }

    void IDisposable.Dispose()
    {
      Close();
    }

    public void Close()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool disposing)
    {
      if (disposed)
        return;

      if (disposing) {
        try {
          if (client != null) {
            client.Close();
            Trace.Log(this, "disconnected: {0} ({1})", remoteEndPoint, host);
          }
        }
        finally {
          client = null;
          stream = null;
        }
      }

      disposed = true;
    }

    protected virtual void Connect(string host, int port, UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (host == null)
        throw new ArgumentNullException("host");
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      TcpClient c = null;

      try {

        c = new TcpClient();

        c.ReceiveTimeout = DefaultTimeoutMilliseconds;
        c.SendTimeout    = DefaultTimeoutMilliseconds;

        Trace.Log(this, "connecting to {0}:{1}", host, port);

        c.Connect(host, port);

        this.host = host;
        this.client = c;
        this.remoteEndPoint = c.Client.RemoteEndPoint as IPEndPoint;
        this.localEndPoint  = c.Client.LocalEndPoint  as IPEndPoint;
        this.stream = CreateBufferedStream(Trace.GetTracingStream(this, client.GetStream()));

        Trace.Log(this, "connected: {0} ({1})", this.remoteEndPoint, host);
      }
      catch (Exception ex) {
        Trace.Log(this, ex);

        if (c != null)
          c.Close();

        this.client = null;

        throw new ConnectionException("connect failed", ex);
      }

      try {
        if (createAuthenticatedStreamCallback != null)
          UpgradeStream(createAuthenticatedStreamCallback);

        isSecurePortConnection = RawStream is SslStream;
      }
      catch {
        this.client.Close();
        this.client = null;

        throw;
      }
    }

    public static Stream CreateClientSslStream(ConnectionBase connection, Stream baseStream, X509Certificate2Collection certCollection, RemoteCertificateValidationCallback certificateValidationCallback, LocalCertificateSelectionCallback certificateSelectionCallback)
    {
      var sslStream = new SslStream(baseStream, false, certificateValidationCallback, certificateSelectionCallback);

      sslStream.AuthenticateAsClient(connection.host, certCollection, SslProtocols.Default, true);

      return sslStream;
    }

    public virtual void UpgradeStream(UpgradeConnectionStreamCallback upgradeStreamCallback)
    {
      if (upgradeStreamCallback == null)
        throw new ArgumentNullException("upgradeStreamCallback");

      try {
        var upgradedStream = upgradeStreamCallback(this, RawStream);

        if (upgradedStream is SslStream) {
          var sslStream = upgradedStream as SslStream;

          Trace.Log(this, "TLS started: {0}, {1}, {2}, {3}{4}",
                    sslStream.SslProtocol,
                    sslStream.CipherAlgorithm,
                    sslStream.HashAlgorithm,
                    sslStream.KeyExchangeAlgorithm,
                    sslStream.IsMutuallyAuthenticated ? ", mutually authenticated" : string.Empty);
        }

        stream = CreateBufferedStream(Trace.GetTracingStream(this, upgradedStream));
      }
      catch (AuthenticationException ex) {
        Trace.Log(this, ex);

        throw new ConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
      catch (IOException ex) {
        Trace.Log(this, ex);

        throw new ConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
    }

    protected virtual LineOrientedBufferedStream CreateBufferedStream(Stream stream)
    {
      return new LineOrientedBufferedStream(stream);
    }

    public override string ToString()
    {
      return string.Format("{{Host={0}, RemoteEndPoint={1}, LocalEndPoint={2}, Connected={3}}}",
                           Host,
                           remoteEndPoint,
                           localEndPoint,
                           Socket.Connected);
    }

    protected void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    protected void CheckConnected()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
      if (client == null)
        throw new InvalidOperationException("not connected");
    }

    private bool disposed = false;
    private int id;
    private string host = null;
    private IPEndPoint remoteEndPoint;
    private IPEndPoint localEndPoint;
    private bool isSecurePortConnection;
    private TcpClient client = null;
    private LineOrientedBufferedStream stream = null;
  }
}
