// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

using Smdn.Collections;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;
using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.Pop3.Client.Transaction;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Pop3.WebClients {
  public class PopSessionManager : IWebRequestCreate {
    private static readonly PopSessionManager instance = new PopSessionManager();
    private static readonly NetworkCredential AnonymousCredential = new NetworkCredential("anonymous", "@anonymous");

    public static X509Certificate2Collection CertCollection {
      get { return PopConnection.CertCollection; }
    }

    public static RemoteCertificateValidationCallback CertificateValidationCallback {
      get { return PopConnection.CertificateValidationCallback; }
      set { PopConnection.CertificateValidationCallback = value; }
    }

    public static LocalCertificateSelectionCallback CertificateSelectionCallback {
      get { return PopConnection.CertificateSelectionCallback; }
      set { PopConnection.CertificateSelectionCallback = value; }
    }

    private static UpgradeConnectionStreamCallback createSslStreamCallback = PopConnection.CreateSslStream;

    public static UpgradeConnectionStreamCallback CreateSslStreamCallback {
      get { return createSslStreamCallback; }
      set
      {
        if (createSslStreamCallback == null)
          throw new ArgumentNullException("CreateSslStreamCallback");
        createSslStreamCallback = value;
      }
    }

    public static bool RegisterWebRequestPrefix()
    {
      if (WebRequest.RegisterPrefix(PopUri.UriSchemePop, instance) | WebRequest.RegisterPrefix(PopUri.UriSchemePops, instance)) {
        // this will broke in windows
        // PopUri.RegisterParser();
        return true;
      }
      else {
        return false;
      }
    }

    public static void DisconnectFrom(Uri serverUri)
    {
      if (!PopUri.IsPop(serverUri))
        throw new ArgumentException("scheme must be pop or pops", "serverUri");

      var port = serverUri.Port;

      if (port == -1)
        port = PopUri.GetDefaultPortFromScheme(serverUri);

      DisconnectFrom(serverUri.Host, port);
    }

    public static void DisconnectFrom(string host, int port)
    {
      instance.InternalDisconnectFrom(host, port);
    }

    internal static PopSession CreateSession(IPopSessionProfile profile)
    {
      try {
        PopSession session = null;

        var result = CreateSession(profile, out session);

        if (result.Succeeded)
          return session;
        else
          throw new WebException(result.Description, null, WebExceptionStatus.ProtocolError, new PopWebResponse(result));
      }
      catch (PopConnectionException ex) {
        throw new WebException("connection error", ex, WebExceptionStatus.ConnectFailure, null);
      }
      catch (PopException ex) {
        throw new WebException("internal error", ex, WebExceptionStatus.UnknownError, null);
      }
      catch (TimeoutException ex) {
        throw new WebException("timed out", ex, WebExceptionStatus.Timeout, null);
      }
      catch (WebException) {
        throw;
      }
    }

    private static PopTransactionResult CreateSession(IPopSessionProfile profile, out PopSession session)
    {
      PopTransactionResult result;
      session = null;

      var authority = profile.Authority;

      session = new PopSession(authority.Host,
                               authority.Port,
                               profile.Timeout,
                               (authority.Scheme == PopUri.UriSchemePops)
                                 ? createSslStreamCallback
                                 : null);

      session.HandlesIncapableAsException = false;
      session.TransactionTimeout  = profile.Timeout;
      session.SendTimeout         = profile.ReadWriteTimeout;
      session.ReceiveTimeout      = profile.ReadWriteTimeout;

      // try querying server capability (ignore error; POP3 Extension Mechanism might not supported)
      session.Capa();

      if (profile.UseTlsIfAvailable && session.ServerCapabilities.IsCapable(PopCapability.Stls)) {
        var r = session.Stls(createSslStreamCallback, false);

        if (r.Failed)
          throw new WebException(r.Description, null, WebExceptionStatus.SecureChannelFailure, null);

        // try re-querying server capability (ignore error; POP3 Extension Mechanism might not supported)
        session.Capa();
      }

      result = Authenticate(session,
                            profile.Credentials,
                            PopStyleUriParser.GetUser(authority),
                            PopStyleUriParser.GetAuthType(authority),
                            GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? new string[0]));

      if (result == null) {
        throw new WebException("appropriate authentication mechanism not found", WebExceptionStatus.RequestCanceled);
      }
      else if (result.Failed) {
        try {
          session.Disconnect(false);
        }
        catch (PopConnectionException) {
          // ignore
        }
      }

      return result;
    }

#region "authentication"
    private static IEnumerable<PopAuthenticationMechanism> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = Smdn.Security.Authentication.Sasl.Client.SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Exists(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.InvariantCultureIgnoreCase);
        }))
          yield return PopAuthenticationMechanism.GetKnownOrCreate(mechanism);
      }
    }

    private static PopTransactionResult Authenticate(PopSession session, ICredentialsByHost credentials, string username, PopAuthenticationMechanism authMechanism, IEnumerable<PopAuthenticationMechanism> usingSaslMechanisms)
    {
      /*
       * http://tools.ietf.org/html/rfc2384
       * 4. POP User Name and Authentication Mechanism
       * 
       *    The string ";AUTH=*" indicates that the client SHOULD select an
       *    appropriate authentication mechanism.  It MAY use any mechanism
       *    supported by the POP server.
       * 
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = PopAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = PopAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == PopAuthenticationMechanism.SelectAppropriate)
        return AuthenticateWithAppropriateMechanism(session, credentials, username, usingSaslMechanisms);
      else if (authMechanism == PopAuthenticationMechanism.Anonymous)
        return AuthenticateAsAnonymous(session, canFallback);
      else
        return AuthenticateWithSuppliedMechanism(session, credentials, username, authMechanism);
    }

    private static PopTransactionResult AuthenticateWithAppropriateMechanism(PopSession session, ICredentialsByHost credentials, string username, IEnumerable<PopAuthenticationMechanism> usingSaslMechanisms)
    {
      PopTransactionResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (session.ServerCapabilities.IsCapable(mechanism)) {
          result = session.Auth(credentials, username, mechanism);

          if (result.Succeeded)
            break;
        }
      }

      if (result == null || result.Failed) {
        if (session.ApopAvailable)
          result = session.Apop(credentials, username);
      }

      if (result == null || result.Failed)
        result = session.Login(credentials, username);

      return result;
    }

    private static PopTransactionResult AuthenticateWithSuppliedMechanism(PopSession session, ICredentialsByHost credentials, string username, PopAuthenticationMechanism authMechanism)
    {
      /*
       * http://tools.ietf.org/html/rfc2384
       * 4. POP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism can be expressed by adding ";AUTH=<enc-
       *    auth-type>" to the end of the user name.  If the authentication
       *    mechanism name is not preceded by a "+", it is a SASL POP [SASL]
       *    mechanism.  If it is preceded by a "+", it is either "APOP" or an
       *    extension mechanism.
       * 
       *    When an <enc-auth-type> is specified, the client SHOULD request
       *    appropriate credentials from that mechanism and use the "AUTH",
       *    "APOP", or extension command instead of the "USER" command.  If no
       *    user name is specified, one SHOULD be obtained from the mechanism or
       *    requested from the user as appropriate.
       * 
       * 
       *    If an <enc-auth-type> other than ";AUTH=*" is specified, the client
       *    SHOULD NOT use a different mechanism without explicit user
       *    permission.
       */
      return session.Auth(credentials, username, authMechanism);
    }

    private static PopTransactionResult AuthenticateAsAnonymous(PopSession session, bool canFallback)
    {
      PopTransactionResult result = null;

      if (session.ServerCapabilities.IsCapable(PopAuthenticationMechanism.Anonymous))
        // try AUTH ANONYUMOUS
        result = session.Auth(AnonymousCredential, PopAuthenticationMechanism.Anonymous);

      if (result == null || (result.Failed && canFallback))
        // try anonymous LOGIN
        result = session.Login(AnonymousCredential);

      return result;
    }
#endregion

    /*
     * instance members
     */
    private PopSessionManager()
    {
    }

    WebRequest IWebRequestCreate.Create(Uri uri)
    {
      switch (ExtendedPopStyleUriParser.GetUriForm(uri)) {
        case ExtendedPopUriForm.Mailbox: return new PopMailboxWebRequest(uri, this);
        case ExtendedPopUriForm.Message: return new PopMessageWebRequest(uri, this);
        default: throw new ArgumentException("invalid URI form", "uri");
      }
    }

    internal PopSession GetExistSession(Uri requestUri)
    {
      var authority = PopStyleUriParser.GetAuthority(requestUri);

      if (sessions.ContainsKey(authority))
        return sessions[authority];
      else
        return null;
    }

    internal void RegisterSession(Uri requestUri, PopSession session)
    {
      var authority = PopStyleUriParser.GetAuthority(requestUri);

      PopSession existSession = null;

      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (sessions.ContainsKey(authority)) {
          existSession = sessions[authority];
          sessions.Remove(authority);
        }

        sessions.Add(authority, session);
      }

      if (existSession != null)
        existSession.Disconnect(true);
    }

    internal void UnregisterSession(Uri requestUri)
    {
      UnregisterSession(PopStyleUriParser.GetAuthority(requestUri));
    }

    private void UnregisterSession(string authority)
    {
      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (!sessions.ContainsKey(authority))
          return;

        sessions.Remove(authority);
      }
    }

    private void InternalDisconnectFrom(string host, int port)
    {
      if (host == null)
        throw new ArgumentNullException("host", host);
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      var keys = new List<string>();

      foreach (var pair in sessions) {
        var session = pair.Value;

        if (session.IsDisposed || session.Connection == null) {
          keys.Add(pair.Key);
        }
        else if (string.Equals(session.Connection.Host, host, StringComparison.InvariantCultureIgnoreCase) &&
                 session.Connection.Port == port) {
          keys.Add(pair.Key);

          session.Disconnect(true);
        }
      }

      foreach (var key in keys) {
        UnregisterSession(key);
      }
    }

    private Dictionary<string, PopSession> sessions = new Dictionary<string, PopSession>(StringComparer.InvariantCultureIgnoreCase);
  }
}