using System;
using System.IO;
using System.Net;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Pop3.WebClients {
  [TestFixture]
  public class PopSessionManagerTests {
    [SetUp]
    public void Setup()
    {
      PopSessionManager.RegisterWebRequestPrefix();
    }

    [Test]
    public void TestRegisterWebRequestPrefix()
    {
      var popRequest = WebRequest.Create("pop://localhost/");

      Assert.IsInstanceOfType(typeof(PopWebRequest), popRequest);

      var popsRequest = WebRequest.Create("pops://localhost/");

      Assert.IsInstanceOfType(typeof(PopWebRequest), popsRequest);
    }

    [Test]
    public void TestIWebRequestCreate()
    {
      var popMailboxRequest = WebRequest.Create("pop://localhost/");

      Assert.IsInstanceOfType(typeof(PopWebRequest), popMailboxRequest);
      StringAssert.Contains("PopMailboxWebRequest", popMailboxRequest.GetType().FullName);

      var popMessageRequest = WebRequest.Create("pop://localhost/;MSG=1");

      Assert.IsInstanceOfType(typeof(PopWebRequest), popMessageRequest);
      StringAssert.Contains("PopMessageWebRequest", popMessageRequest.GetType().FullName);
    }

    private void WebExceptionAssertion(WebExceptionStatus expectedStatus, WebException actual)
    {
      if (expectedStatus != actual.Status)
        throw actual;

      if (expectedStatus == WebExceptionStatus.ProtocolError) {
        Assert.IsNotNull(actual.Response);
        Assert.IsInstanceOfType(typeof(PopWebResponse), actual.Response);

        var response = actual.Response as PopWebResponse;

        Assert.IsTrue(response.Result.Failed);
      }
    }

    [Test]
    public void TestConnectFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var request = WebRequest.Create(string.Format("pop://{0}@{1}/", credential.UserName, server.HostPort)) as PopWebRequest;

        request.Timeout = 500;

        server.Stop();

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ConnectFailure, ex);
        }
      }
    }

    [Test]
    public void TestConnectUseTlsIfAvailable()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        // STLS response
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        var defaultCallback = PopSessionManager.CreateSslStreamCallback;

        try {
          var streamUpgraded = false;

          PopSessionManager.CreateSslStreamCallback = delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          };

          using (var response = request.GetResponse()) {
            Assert.AreEqual(response.ResponseUri, requestUri);
          }

          Assert.IsTrue(streamUpgraded, "stream upgraded");

          server.DequeueRequest(); // CAPA
          StringAssert.StartsWith("STLS", server.DequeueRequest());
        }
        finally {
          PopSessionManager.CreateSslStreamCallback = defaultCallback;
        }
      }
    }

    [Test]
    public void TestConnectStlsFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        // STLS response
        server.EnqueueResponse("-ERR\r\n");

        var defaultCallback = PopSessionManager.CreateSslStreamCallback;

        try {
          var streamUpgraded = false;

          PopSessionManager.CreateSslStreamCallback = delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          };

          try {
            using (var response = request.GetResponse()) {
            }
          }
          catch (WebException ex) {
            WebExceptionAssertion(WebExceptionStatus.SecureChannelFailure, ex);
          }

          Assert.IsFalse(streamUpgraded, "stream upgraded");

          server.DequeueRequest(); // CAPA
          StringAssert.StartsWith("STLS", server.DequeueRequest());
        }
        finally {
          PopSessionManager.CreateSslStreamCallback = defaultCallback;
        }
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0};auth=plain@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0};auth=plain@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://;auth=plain@{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://;auth=plain@{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanism(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication(string.Empty);
    }

    private void AuthenticateExceptionWhileAuthentication(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"DIGEST-MD5", "PLAIN"};
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH DIGEST-MD5 response
        server.EnqueueResponse("+ xxx-invalid-response-xxx\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.UnknownError, ex);
          Assert.IsInstanceOfType(typeof(PopException), ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // AUTH PLAIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH LOGIN client response
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToApop()
    {
      AuthenticateWithAppropriateMechanismFallbackToApop(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToApop()
    {
      AuthenticateWithAppropriateMechanismFallbackToApop(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToApop(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.UsingSaslMechanisms = new[] {"PLAIN", "LOGIN"};
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK <timestamp@localhost>\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH PLAIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // APOP response
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH PLAIN client response
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH LOGIN client response
        StringAssert.StartsWith("APOP user", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPass(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPass(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToUserPass(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER user", server.DequeueRequest());
        StringAssert.StartsWith("PASS pass", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToUserPassAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure(";auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToUserPassAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("-ERR\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER user", server.DequeueRequest());
      }
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(";auth=*");
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string.Empty);
    }

    private void AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string authArgs)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}{1}@{2}/", credential.UserName, authArgs, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // TODO: LOGINDISABLE

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.RequestCanceled, ex);
        }

        server.DequeueRequest(); // CAPA
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0};auth=anonymous@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0};auth=anonymous@{1}/", credential.UserName, server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Method = "NOOP";
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");
        // NOOP response
        server.EnqueueResponse("+OK\r\n");
        // QUIT response
        server.EnqueueResponse("+OK\r\n");

        using (var response = request.GetResponse()) {
          Assert.AreEqual(response.ResponseUri, requestUri);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
        server.DequeueRequest(); // AUTH ANONYMOUS client response
        StringAssert.StartsWith("USER anonymous", server.DequeueRequest()); // anonymous credential
        StringAssert.StartsWith("PASS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("-ERR\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.ProtocolError, ex);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER anonymous", server.DequeueRequest()); // anonymous credential
        StringAssert.StartsWith("PASS", server.DequeueRequest());
      }
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var credential = new NetworkCredential("user", "pass");
        var requestUri = new Uri(string.Format("pop://{0}/", server.HostPort));
        var request = WebRequest.Create(requestUri) as PopWebRequest;

        request.Credentials = credential;
        request.UseTlsIfAvailable = true;
        request.KeepAlive = false;
        request.Timeout = 1000;

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // TODO: LOGINDISABLE

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          WebExceptionAssertion(WebExceptionStatus.RequestCanceled, ex);
        }

        server.DequeueRequest(); // CAPA
      }
    }
  }
}
