using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.IO;
using Smdn.Formats;
using Smdn.Formats.Ini;
using Smdn.Formats.Mime;
using Smdn.Net.Imap4;
using Smdn.Net.Imap4.WebClients;
using Smdn.Net.Pop3;
using Smdn.Net.Pop3.WebClients;
using Smdn.Windows.Forms;

namespace Smdn.Applications.TundereBird {
  public partial class MainWindow {
    public MainWindow()
    {
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();

      ImapSessionManager.CertificateValidationCallback += ValidateServerCertificate;
      PopSessionManager.CertificateValidationCallback += ValidateServerCertificate;

      ImapSessionManager.ClientID["name"] = assemblyName.Name;
      ImapSessionManager.ClientID["version"] = assemblyName.Version.ToString();

      InitializeComponent();
      Initialize();
    }

    private static Font CreateMonospaceFont()
    {
      return new Font(FontFamily.GenericMonospace, Control.DefaultFont.SizeInPoints, FontStyle.Regular, GraphicsUnit.Point);
    }

    private void Initialize()
    {
      InitializeContextMenu();

      toolStripMenuItemFile_Quit.Click += delegate { Close(); };

      toolStripMenuItemMailbox_Update.Click       += UpdateMailboxTree;
      toolStripMenuItemMailbox_Append.Click       += delegate(object sender, EventArgs e) { AppendMessage((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Create.Click       += delegate(object sender, EventArgs e) { CreateMailbox((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Delete.Click       += delegate(object sender, EventArgs e) { DeleteMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_MoveToTrash.Click  += delegate(object sender, EventArgs e) { MoveMailboxToTrash((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Rename.Click       += delegate(object sender, EventArgs e) { RenameMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Subscribe.Click    += delegate(object sender, EventArgs e) { ShowSubscribeDialog((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Unsubscribe.Click  += delegate(object sender, EventArgs e) { UnsubscribeMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };

      toolStripMenuItemMessage_Fetch.Click        += delegate(object sender, EventArgs e) { FetchAndSaveMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_ToggleSeen.Click   += delegate(object sender, EventArgs e) { ToggleMessageSeen((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_Expunge.Click      += delegate(object sender, EventArgs e) { ExpungeMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_MoveToTrash.Click  += delegate(object sender, EventArgs e) { MoveMessageToTrash((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_ToggleKeywords_Append.Click
                                                  += delegate(object sender, EventArgs e) { StoreMessageKeyword((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };

      toolStripMenuItemMessage_CopyTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;
      toolStripMenuItemMessage_MoveTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;

      toolStripMenuItemMessage_ToggleKeywords.DropDownOpening += ToggleMessageKeywordsDropDownOpening;

      treeViewMailbox.NodeMouseClick += AccountOrMailboxSelected;

      listViewMessages.View = View.Details;
      listViewMessages.FullRowSelect = true;
      listViewMessages.Columns.Add("タイトル", 160);
      listViewMessages.Columns.Add("サイズ", 60);
      listViewMessages.Columns.Add("受信日", 100);
      listViewMessages.Columns.Add("送信者", 100);
      listViewMessages.Columns.Add("フラグ", 200);
      listViewMessages.SelectedIndexChanged += MessageSelected;

      textBoxMessage.Multiline = true;
      textBoxMessage.ScrollBars = ScrollBars.Both;
      textBoxMessage.Font = CreateMonospaceFont();

      textBoxStatus.ReadOnly = true;
      textBoxStatus.Multiline = true;
      textBoxStatus.ScrollBars = ScrollBars.Both;
      textBoxStatus.Font = CreateMonospaceFont();

      textBoxServerInfo.ReadOnly = true;
      textBoxServerInfo.Multiline = true;
      textBoxServerInfo.ScrollBars = ScrollBars.Both;
      textBoxServerInfo.Font = CreateMonospaceFont();

      timerUpdate.Interval = 5 * 60 * 1000; // 5 min
      timerUpdate.Tick += delegate {
        NoOp();
      };
      timerUpdate.Enabled = true;
    }

    private ToolStripItem[] mailboxToolStripItems;
    private ToolStripItem[] messageToolStripItems;

    private void SetMailboxContextMenuProperty()
    {
      var mailboxSelected = treeViewMailbox.SelectedNode is ImapMailboxNode;

      toolStripMenuItemMailbox_Append.Enabled =
      toolStripMenuItemMailbox_MoveToTrash.Enabled =
      toolStripMenuItemMailbox_Delete.Enabled =
      toolStripMenuItemMailbox_Rename.Enabled =
      toolStripMenuItemMailbox_Unsubscribe.Enabled = mailboxSelected;

      toolStripMenuItemMailbox_Append.ContextNode =
      toolStripMenuItemMailbox_MoveToTrash.ContextNode =
      toolStripMenuItemMailbox_Delete.ContextNode =
      toolStripMenuItemMailbox_Rename.ContextNode =
      toolStripMenuItemMailbox_Unsubscribe.ContextNode = treeViewMailbox.SelectedNode as ImapMailboxNode;

      var accountSelected = treeViewMailbox.SelectedNode is ImapAccountNode;

      toolStripMenuItemMailbox_Update.Enabled = treeViewMailbox.SelectedNode is IAccountContainer ||
                                                treeViewMailbox.SelectedNode is IMailboxContainer;

      toolStripMenuItemMailbox_Create.Enabled =
      toolStripMenuItemMailbox_Subscribe.Enabled = mailboxSelected || accountSelected;

      toolStripMenuItemMailbox_Update.ContextNode =
      toolStripMenuItemMailbox_Create.ContextNode =
      toolStripMenuItemMailbox_Subscribe.ContextNode = treeViewMailbox.SelectedNode;
    }

    private void SetMessageContextMenuProperty()
    {
      var contextMessage     = (0 < listViewMessages.SelectedItems.Count) ? listViewMessages.SelectedItems[0] as IMessageContainer : null;
      var contextImapMessage = contextMessage as ImapMessageItem;

      toolStripMenuItemMessage_Fetch.Enabled = 
      toolStripMenuItemMessage_Expunge.Enabled = (contextMessage != null);

      toolStripMenuItemMessage_MoveTo.Enabled = 
      toolStripMenuItemMessage_CopyTo.Enabled = 
      toolStripMenuItemMessage_ToggleSeen.Enabled = 
      toolStripMenuItemMessage_MoveToTrash.Enabled =
      toolStripMenuItemMessage_ToggleKeywords.Enabled = (contextImapMessage != null);

      toolStripMenuItemMessage_Fetch.ContextMessage = 
      toolStripMenuItemMessage_MoveTo.ContextMessage = 
      toolStripMenuItemMessage_CopyTo.ContextMessage = 
      toolStripMenuItemMessage_ToggleSeen.ContextMessage = 
      toolStripMenuItemMessage_Expunge.ContextMessage = 
      toolStripMenuItemMessage_MoveToTrash.ContextMessage = 
      toolStripMenuItemMessage_ToggleKeywords.ContextMessage = contextMessage;

      toolStripMenuItemMessage_MoveTo.DropDownItems.Clear();
      toolStripMenuItemMessage_MoveTo.DropDownItems.Add(contextMenuDummyItem);

      toolStripMenuItemMessage_CopyTo.DropDownItems.Clear();
      toolStripMenuItemMessage_CopyTo.DropDownItems.Add(contextMenuDummyItem);

      if (contextImapMessage == null)
        toolStripMenuItemMessage_ToggleSeen.Checked = false;
      else
        toolStripMenuItemMessage_ToggleSeen.Checked = contextImapMessage.Message.Flags.Has(ImapMessageFlag.Seen);
    }

    private void InitializeContextMenu()
    {
      mailboxToolStripItems = new ToolStripItem[toolStripMenuItemMailbox.DropDownItems.Count];

      toolStripMenuItemMailbox.DropDownItems.CopyTo(mailboxToolStripItems, 0);

      messageToolStripItems = new ToolStripItem[toolStripMenuItemMessage.DropDownItems.Count];

      toolStripMenuItemMessage.DropDownItems.CopyTo(messageToolStripItems, 0);

      contextMenuMailbox.Items.Add(contextMenuDummyItem);

      contextMenuMailbox.Opening += delegate {
        SetMailboxContextMenuProperty();

        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };

      contextMenuMailbox.Closed += delegate {
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownOpening += delegate {
        SetMailboxContextMenuProperty();

        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownClosed += delegate {
        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };


      contextMenuMessage.Items.Add(contextMenuDummyItem);

      contextMenuMessage.Opening += delegate {
        SetMessageContextMenuProperty();

        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };

      contextMenuMessage.Closed += delegate {
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownOpening += delegate {
        SetMessageContextMenuProperty();

        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownClosed += delegate {
        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };
    }

    protected override void OnShown (EventArgs e)
    {
      base.OnShown (e);

      Application.DoEvents();

      LoadConfig();

      InitializeMailboxTree();
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing (e);

      if (!e.Cancel) {
        LogoutAll();
      }
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;

      var message = new StringBuilder();

      message.AppendLine("サーバ証明書に問題があります。　続行しますか?");
      message.AppendLine();
      message.AppendFormat("SslPolicyErrors: {0}", sslPolicyErrors);
      message.AppendLine();
      message.AppendFormat("X509Certificate: {0}{1}", Environment.NewLine, certificate.ToString(false));
      message.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          message.AppendFormat("発行先: {0}", element.Certificate.Subject);
          message.AppendLine();
          message.AppendFormat("発行者: {0}", element.Certificate.Issuer);
          message.AppendLine();
          message.AppendFormat("有効期間: {0} から {1}", element.Certificate.GetEffectiveDateString(), element.Certificate.GetExpirationDateString());
          message.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            message.AppendFormat("Status: {0}", status.Status);
            message.AppendLine();
            message.AppendFormat("StatusInformation: {0}", status.StatusInformation);
            message.AppendLine();
            message.AppendLine();
          }
        }
      }

      var result = MessageBox.Show(message.ToString(),
                                   "ValidateServerCertificate",
                                   MessageBoxButtons.YesNo);

      return (result == DialogResult.Yes);
    }

    private void LoadConfig()
    {
      var config = IniDocument.Load("tunderebird.ini");
      var accounts = new List<AccountBase>();

      var defaultSection = config.DefaultSection;

      mimeTypesFile = defaultSection.Get("MimeTypesFile", "/etc/mime.types");

      // 証明書ファイルのセクション
      var certFilesSection = config["CertFiles"];

      foreach(var entry in certFilesSection) {
        try {
          ImapSessionManager.CertCollection.Add(new X509Certificate2(entry.Value));
        }
        catch (Exception ex) {
          AppendStatus("証明書{0}を読み込めませんでした", entry.Value);
          AppendStatus("Exception: {0}", ex);
        }
      }

      config.Remove(certFilesSection.Name);

      // アカウントのセクション
      foreach (var section in config) {
        if (section.IsDefaultSection)
          continue;

        try {
          var serverUri = section.GetThrowException<Uri>("ServerURL", ConvertUtils.ToUri);
          Converter<string, bool> convertYesNo = (val) => val != "no";

          AccountBase account;

          if (ImapUri.IsImap(serverUri)) {
            var username = ImapStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentials)new NetworkCredential(username, section["Password"])
              : (ICredentials)new PasswordInputCredential(username, this);

            account = new ImapAccount(section.Name,
                                      serverUri,
                                      credential,
                                      section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                      section.Get("TryTLS", true, convertYesNo),
                                      section.Get("ReadOnly", true, convertYesNo));
          }
          else if (PopUri.IsPop(serverUri)) {
            var username = PopStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentials)new NetworkCredential(username, section["Password"])
              : (ICredentials)new PasswordInputCredential(username, this);

            account = new PopAccount(section.Name,
                                     serverUri,
                                     credential,
                                     section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                     section.Get("TryTLS", true, convertYesNo));
          }
          else {
            throw new UriFormatException("unsupported uri scheme");
          }

          accounts.Add(account);
        }
        catch (Exception ex) {
          AppendStatus("{0}のアカウント設定を読み込めませんでした", section.Name);
          AppendStatus("Exception: {0}", ex);
        }
      }

      this.imapAccounts = accounts.FindAll(delegate(AccountBase account) { return account is ImapAccount; })
                                  .ConvertAll(delegate(AccountBase account) { return account as ImapAccount; }).ToArray();
      this.popAccounts = accounts.FindAll(delegate(AccountBase account) { return account is PopAccount; })
                                 .ConvertAll(delegate(AccountBase account) { return account as PopAccount; }).ToArray();
      this.allAccounts = accounts.ToArray();

      if (this.allAccounts.Length == 0)
        AppendStatus("表示できるアカウント設定がありません。");
    }

    private void InitializeMailboxTree()
    {
      foreach (var account in imapAccounts) {
        UpdateMailboxTree(account);
      }

      foreach (var account in popAccounts) {
        UpdateMailboxTree(account);
      }
    }

    private void UpdateMailboxTree(object sender, EventArgs e)
    {
      if (sender is TreeViewContextMenuItem) {
        var node = (sender as TreeViewContextMenuItem).ContextNode;

        if (node is IAccountContainer)
          UpdateMailboxTree((node as IAccountContainer).Account);
      }
    }

    private void UpdateMailboxTree(AccountBase account)
    {
      if (account is ImapAccount)
        UpdateMailboxTree(account as ImapAccount);
      else if (account is PopAccount)
        UpdateMailboxTree(account as PopAccount);
    }

    private void UpdateMailboxTree(ImapAccount account)
    {
      ImapMailbox[] mailboxes;

      try {
        Cursor = Cursors.WaitCursor;

        try {
          AppendStatus("{0}のメールボックスを取得しています", account.Name);

          // retrieve server capability
          account.Request(ImapWebRequestMethods.NoOp);

          if (account.ServerCapabilities.Has(ImapCapability.GimapXlist)) {
            // FIXME: list subscribed mailbox
            mailboxes = account.Request(account.ServerUri, ImapWebRequestMethods.XList, delegate(ImapWebResponse response) {
              return response.Mailboxes;
            });

            foreach (var mailbox in mailboxes) {
              if (mailbox.Flags.Has(ImapMailboxFlag.Trash)) {
                account.TrashMailboxName = mailbox.Name;
                break;
              }
            }
          }
          else {
            mailboxes = account.Request(delegate(ImapWebResponse response) {
              return response.Mailboxes;
            });
          }
        }
        catch (WebException ex) {
          AppendStatus("{0}のメールボックス一覧を取得できませんでした", account.Name);
          AppendStatus(ex);
          return;
        }
      }
      finally {
        Cursor = Cursors.Default;
      }

      Application.DoEvents();

      // アカウントとメールボックスのノードを作成する
      try {
        Cursor = Cursors.WaitCursor;

        treeViewMailbox.BeginUpdate();

        var removingNodes = new List<ImapMailboxNode>();

        foreach (var pair in imapMailboxNodeMap) {
          if (pair.Value.Account == account)
            removingNodes.Add(pair.Value);
        }

        foreach (var removingNode in removingNodes) {
          imapMailboxNodeMap.Remove(removingNode.Mailbox);
        }

        ImapAccountNode rootNode = null;

        foreach (TreeNode node in treeViewMailbox.Nodes) {
          if ((node is ImapAccountNode) && (node as ImapAccountNode).Account == account) {
            rootNode = node as ImapAccountNode;
            break;
          }
        }

        if (rootNode == null) {
          rootNode = new ImapAccountNode(account);
          treeViewMailbox.Nodes.Add(rootNode);
        }
        else {
          rootNode.Nodes.Clear();
        }

        var sortedMailboxes = new List<ImapMailbox>(mailboxes);

        sortedMailboxes.Sort(delegate(ImapMailbox x, ImapMailbox y) {
          return string.Compare(x.Name, y.Name);
        });

        foreach (var mailbox in mailboxes) {
          var node = new ImapMailboxNode(account, mailbox);

          if (mailbox.Name.Contains(mailbox.HierarchyDelimiter)) {
            // find parent
            var hierarchy = mailbox.Name.Split(new [] {mailbox.HierarchyDelimiter}, StringSplitOptions.None);
            TreeNode parent = rootNode;
            var found = false;

            for (var i = 0; i < hierarchy.Length - 1; i++) {
              found = false;

              foreach (TreeNode n in parent.Nodes) {
                if (n is ImapMailboxNode) {
                  if ((n as ImapMailboxNode).Mailbox.Name == string.Join(mailbox.HierarchyDelimiter, hierarchy, 0, i + 1)) {
                    found = true;
                    parent = n;
                    break;
                  }
                }
              }

              if (!found)
                break;
            }

            if (!found) {
              parent = rootNode;

              for (var i = 0; i < hierarchy.Length - 1; i++) {
                var name = string.Join(mailbox.HierarchyDelimiter, hierarchy, 0, i + 1);
                var n = rootNode.Nodes.Find(name, true).FirstOrDefault();

                if (n == null) {
                  n = new ImapNonExistentMailboxNode(hierarchy[i], name);
                  parent.Nodes.Add(n);
                  parent = n;
                }
                else {
                  parent = n;
                }
              }
            }

            parent.Nodes.Add(node);
          }
          else {
            rootNode.Nodes.Add(node);
          }

          imapMailboxNodeMap.Add(mailbox, node);
        }

        rootNode.ExpandAll();
      }
      finally {
        treeViewMailbox.EndUpdate();

        Cursor = Cursors.Default;
      }

      Application.DoEvents();

      try {
        Cursor = Cursors.WaitCursor;

        foreach (var pair in imapMailboxNodeMap) {
          var node = pair.Value;

          if (node.Account != account)
            continue;

          var mailbox = pair.Key;

          if (mailbox.Flags.Has(ImapMailboxFlag.NoSelect))
            continue;

          try {
            AppendStatus("メールボックス'{0}'のステータスを取得しています", mailbox.Name);

            account.Request(mailbox.Url, ImapWebRequestMethods.Examine);
          }
          catch (WebException ex) {
            AppendStatus("メールボックス'{0}'のステータスを取得できませんでした", mailbox.Name);
            AppendStatus(ex);
          }

          node.UpdateText();

          Application.DoEvents();
        }
      }
      finally {
        Cursor = Cursors.Default;
      }

      treeViewMailbox.Refresh();
    }

    private void UpdateMailboxTree(PopAccount account)
    {
      // アカウントのノードを作成する
      PopAccountNode rootNode = null;

      try {
        Cursor = Cursors.WaitCursor;

        treeViewMailbox.BeginUpdate();

        foreach (TreeNode node in treeViewMailbox.Nodes) {
          if ((node is PopAccountNode) && (node as PopAccountNode).Account == account) {
            rootNode = node as PopAccountNode;
            break;
          }
        }

        if (rootNode == null) {
          rootNode = new PopAccountNode(account);
          treeViewMailbox.Nodes.Add(rootNode);
        }
        else {
          rootNode.Nodes.Clear();
        }

        rootNode.ExpandAll();
      }
      finally {
        treeViewMailbox.EndUpdate();

        Cursor = Cursors.Default;
      }

      Application.DoEvents();

      try {
        Cursor = Cursors.WaitCursor;

        try {
          AppendStatus("メールボックス'{0}'のステータスを取得しています", account.Name);

          var node = account.Request(null, PopWebRequestMethods.Stat, delegate(PopWebResponse response) {
            return new PopMailboxNode(account, response.DropList);
          });

          rootNode.Nodes.Add(node);

          node.UpdateText();
        }
        catch (WebException ex) {
          AppendStatus("メールボックス'{0}'のステータスを取得できませんでした", account.Name);
          AppendStatus(ex);

          treeViewMailbox.Nodes.Remove(rootNode);
        }

        Application.DoEvents();
      }
      finally {
        Cursor = Cursors.Default;
      }

      treeViewMailbox.Refresh();
    }

    private void AccountOrMailboxSelected(object sender, TreeNodeMouseClickEventArgs e)
    {
      if (e.Node is IMailboxContainer) {
        var mailboxContainer = e.Node as IMailboxContainer;

        ViewServerInfo(mailboxContainer);

        UpdateMessageList(mailboxContainer);
      }
      else if (e.Node is IAccountContainer) {
        ViewServerInfo(e.Node as IAccountContainer);
      }
    }

    private void UpdateMessageList()
    {
      var mailboxContainer = treeViewMailbox.SelectedNode as IMailboxContainer;

      if (mailboxContainer == null)
        return;
      else
        UpdateMessageList(mailboxContainer);
    }

    private void UpdateMessageList(IMailboxContainer container)
    {
      if (container.Account is ImapAccount)
        UpdateMessageList(container.Account as ImapAccount, container.GetMailbox() as ImapMailbox);
      else
        UpdateMessageList(container.Account as PopAccount);
    }

    private void UpdateMessageList(ImapAccount account, ImapMailbox mailbox)
    {
      var mailboxNode = imapMailboxNodeMap[mailbox];

      if (mailbox.Flags.Has(ImapMailboxFlag.NoSelect)) {
        AppendStatus("{0}は選択できません", mailbox.Name);
        return;
      }

      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", mailbox.Name);

      try {
        var messageCount = account.Request(mailbox.Url, delegate(ImapWebResponse response) {
          CreateMessageList(account, response.Messages);

          mailboxNode.UnseenMessage = 0;

          foreach (var message in response.Messages) {
            if (!message.Flags.Has(ImapMessageFlag.Seen))
              mailboxNode.UnseenMessage++;
          }

          mailboxNode.UpdateText();

          return response.Messages.Length;
        });

        AppendStatus("全{0}件のメッセージがあります", messageCount);
      }
      catch (WebException ex) {
        AppendStatus("{0}のメッセージ一覧を取得できませんでした", mailbox.Name);
        AppendStatus(ex);
      }
    }

    private void CreateMessageList(ImapAccount account, ImapMessage[] messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new ImapMessageItem(account, message));
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private void UpdateMessageList(PopAccount account)
    {
      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", account.Name);

      try {
        var scanLists = account.Request(account.ServerUri, PopWebRequestMethods.List, delegate(PopWebResponse response) {
          return response.ScanLists;
        });

        var messages = new Dictionary<PopScanListing, MimeMessage>();

        foreach (var scanList in scanLists) {
          var pair = account.Request(new Uri(account.ServerUri, string.Format("./;MSG={0}", scanList.MessageNumber)), PopWebRequestMethods.Top, delegate(PopWebResponse response) {
            using (var responseStream = response.GetResponseStream()) {
              return new KeyValuePair<PopScanListing, MimeMessage>(scanList, MimeMessage.Load(responseStream));
            }
          });

          messages.Add(pair.Key, pair.Value);
        }

        CreateMessageList(account, messages);

        AppendStatus("全{0}件のメッセージがあります", messages.Count());
      }
      catch (WebException ex) {
        AppendStatus("{0}のメッセージ一覧を取得できませんでした", account.Name);
        AppendStatus(ex);
      }
    }

    private void CreateMessageList(PopAccount account, IDictionary<PopScanListing, MimeMessage> messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new PopMessageItem(account, message.Key, message.Value));
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private void MessageSelected(object sender, EventArgs e)
    {
      if (listViewMessages.SelectedItems.Count <= 0)
        return;

      MimeMessage message = null;

      if (listViewMessages.SelectedItems[0] is IMessageContainer) {
        var container = listViewMessages.SelectedItems[0] as IMessageContainer;
        var messageUrl = container.MessageUrl;

        AppendStatus("{0}の本文を取得しています", messageUrl);

        try {
          message = container.Account.Request(messageUrl, delegate(WebResponse response) {
            try {
              return MimeMessage.Load(response.GetResponseStream());
            }
            catch (Exception ex) {
              AppendStatus("本文の解析に失敗しました");
              AppendStatus("Exception: {0}", ex);
              return null;
            }
          });
        }
        catch (WebException ex) {
          AppendStatus("{0}の本文を取得できませんでした", messageUrl);
          AppendStatus(ex);
          return;
        }
      }

      if (message != null)
        ViewMessage(message);

      AppendStatus("完了");
    }

    private void ViewMessage(MimeMessage message)
    {
      var sb = new StringBuilder(4096);

      if (message.MimeType.TypeEquals("text"))
        sb.AppendLine(message.ReadContentAsText());
      else if (message.MimeType.TypeEquals("multipart"))
        sb.AppendLine("(マルチパートメッセージです)");
      else
        sb.AppendLine("(非テキストメッセージです)");

      if (0 < message.SubParts.Count) {
        sb.AppendLine("-- multipart message --");
        sb.AppendLine();

        foreach (var subPart in message.SubParts) {
          if (subPart.MimeType.TypeEquals("text"))
            sb.AppendLine(subPart.ReadContentAsText());
          else if (subPart.MimeType.TypeEquals("multipart"))
            sb.AppendLine("(マルチパートメッセージです)");
          else
            sb.AppendLine("(非テキストメッセージです)");

          sb.AppendLine("-- end of part --");
        }
      }

      textBoxMessage.Text = sb.ToString();
    }

    private void ViewServerInfo(IAccountContainer container)
    {
      if (container.Account is ImapAccount)
        ViewServerInfo(container.Account as ImapAccount);
      else if (container.Account is PopAccount)
        ViewServerInfo(container.Account as PopAccount);
    }

    private void ViewServerInfo(ImapAccount account)
    {
      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      sb.AppendLine("[server ID]");

      foreach (var pair in account.ServerID) {
        sb.AppendFormat("{0}: {1}", pair.Key, pair.Value);
        sb.AppendLine();
      }
      sb.AppendLine();

      foreach (var ns in new[] {
        new {Namespaces = account.ServerNamespaces.PersonalNamespaces, Title = "personal"},
        new {Namespaces = account.ServerNamespaces.SharedNamespaces, Title = "shared"},
        new {Namespaces = account.ServerNamespaces.OtherUsersNamespaces, Title = "other users"},
      }) {
        sb.AppendFormat("[namespaces ({0})]", ns.Title);
        sb.AppendLine();
        foreach (var desc in ns.Namespaces) {
          sb.AppendFormat("  prefix: \"{0}\" delimiter: \"{1}\"", desc.Prefix, desc.HierarchyDelimiter);
          sb.AppendLine();
        }
        sb.AppendLine();
      }

      textBoxServerInfo.Text = sb.ToString();
    }

    private void ViewServerInfo(PopAccount account)
    {
      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      textBoxServerInfo.Text = sb.ToString();
    }

    private readonly string fileDialogMessageFilterString = FileDialogFilter.CreateFilterString(new[] {
      new[] {"emlファイル", "*.eml"},
      new[] {"すべてのファイル", "*.*"},
    });

    private void AppendMessage(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;
      var account = node.Account;

      using (var openFileDialog = new OpenFileDialog()) {
        openFileDialog.Filter = fileDialogMessageFilterString;
        openFileDialog.Title = string.Format("{0}へファイルをアップロード", mailbox.Name);
        openFileDialog.Multiselect = false;

        if (openFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        var request = account.CreateRequest(mailbox.Url, ImapWebRequestMethods.Append);

        using (var requestStream = request.GetRequestStream()) {
          if (PathUtils.AreExtensionEqual(openFileDialog.FileName, ".eml")) {
            using (var stream = openFileDialog.OpenFile()) {
              stream.WriteToEnd(requestStream, 0x4000);
            }
          }
          else {
            WriteMimeCapsulatedMessageTo(openFileDialog.FileName, requestStream);
          }
        }

        Uri appendedMessageUri = null;

        using (var response = request.GetResponse() as ImapWebResponse) {
          appendedMessageUri = response.ResponseUri;
        }

        if (appendedMessageUri != null) {
          var appendedMessage = account.Request(new Uri(appendedMessageUri, string.Format("?UID {0}", ImapStyleUriParser.GetUid(appendedMessageUri))), delegate(ImapWebResponse response) {
            return response.Messages[0];
          });

          listViewMessages.Items.Add(new ImapMessageItem(account, appendedMessage));
        }

        MessageBox.Show(string.Format("{0}を{1}へアップロードしました", openFileDialog.FileName, mailbox.Name));
      }
    }

    private void WriteMimeCapsulatedMessageTo(string path, Stream stream)
    {
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();
      var description = string.Format("{0}\r\nuploaded by {1}", path, assemblyName.Name);

      // TODO: Content-Type: *;format=flowed
      var part1 = new MimeMessage();

      part1.MimeType = MimeType.TextPlain;
      part1.WriteContent(description, MimeMessageTransferEncoding.SevenBit, Encoding.ASCII);

      MimeMessage part2;

      using (var attachmentStream = File.OpenRead(path)) {
        part2 = MimeMessage.CreateAttachment(attachmentStream,
                                             MimeType.GetMimeTypeByExtension(path, mimeTypesFile) ?? MimeType.ApplicationOctetStream,
                                             Path.GetFileName(path),
                                             description,
                                             true /*inline*/);
      }

      var message = MimeMessage.CreateMultipart(new[] {part1, part2});

      message.Headers.Add(new MimeHeader("Date", DateTimeConvert.ToRFC822DateTimeString(DateTimeOffset.Now)));
      message.Headers.Add(new MimeHeader("Subject", path, MimeHeaderEncoding.BEncoding, Encoding.UTF8));
      message.Headers.Add(new MimeHeader("User-Agent", string.Format("{0} {1}", assemblyName.Name, assemblyName.Version)));

      message.Save(stream);
    }

    private void CreateMailbox(TreeNode node)
    {
      ImapAccount account;
      ImapMailbox parentMailbox;

      if (node is ImapMailboxNode) {
        account = (node as ImapMailboxNode).Account;
        parentMailbox = (node as ImapMailboxNode).Mailbox;
      }
      else if (node is ImapAccountNode) {
        account = (node as ImapAccountNode).Account;
        parentMailbox = null;
      }
      else {
        return;
      }

      string createMailboxName;

      using (var dialog = new InputDialog()) {
        dialog.Text = "メールボックスの作成";
        dialog.Description = string.Format("新しく{0}に作成するメールボックスの名前を指定してください。",
                                           (parentMailbox == null) ? account.Name : parentMailbox.Name);
        dialog.InputText = "新しいメールボックス";

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        createMailboxName = dialog.InputText;
      }

      if (parentMailbox != null)
        createMailboxName = parentMailbox.Name + parentMailbox.HierarchyDelimiter + createMailboxName;

      try {
        account.Request(new Uri(account.ServerUri, createMailboxName), ImapWebRequestMethods.Create);

        AppendStatus("{0}を作成しました", createMailboxName);
      }
      catch (WebException ex) {
        AppendStatus("{0}の作成に失敗しました", createMailboxName);
        AppendStatus(ex);
      }

      UpdateMailboxTree(account);
    }

    private void DeleteMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;
      var result = MessageBox.Show(this,
                                   string.Format("{0}を削除します。　この操作は元に戻せません。", mailbox.Name),
                                   "メールボックスの削除", 
                                   MessageBoxButtons.OKCancel,
                                   MessageBoxIcon.Warning,
                                   MessageBoxDefaultButton.Button2);

      if (result != DialogResult.OK)
        return;

      try {
        node.Account.Request(mailbox.Url, ImapWebRequestMethods.Delete);

        AppendStatus("{0}を削除しました", mailbox.Name);
      }
      catch (WebException ex) {
        AppendStatus("{0}の削除に失敗しました", mailbox.Name);
        AppendStatus(ex);
      }

      UpdateMailboxTree(node.Account);
    }

    private void MoveMailboxToTrash(ImapMailboxNode node)
    {
      var account = node.Account;
      var trashMailbox = GetOrCreateTrashMailbox(account);

      if (trashMailbox == null)
        return;

      var newMailboxName = node.Mailbox.Name;
      var delim = newMailboxName.LastIndexOf(node.Mailbox.HierarchyDelimiter);

      if (0 < delim)
        newMailboxName = newMailboxName.Substring(delim + 1);

      RenameMailbox(node.Account, node.Mailbox, trashMailbox.Name + trashMailbox.HierarchyDelimiter + newMailboxName);
    }

    private ImapMailbox GetOrCreateTrashMailbox(ImapAccount account)
    {
      foreach (var n in imapMailboxNodeMap.Values) {
        if (n.Account != account)
          continue;

        if (n.Account.TrashMailboxName == n.Mailbox.Name)
          return n.Mailbox;
        else if (n.Mailbox.Name == "Trash")
          return n.Mailbox;
      }

      try {
        return account.Request(new Uri(account.ServerUri, "Trash"), ImapWebRequestMethods.Create, delegate(ImapWebResponse response) {
          return response.Mailboxes[0];
        });
      }
      catch (WebException ex) {
        AppendStatus("メールボックス'Trash'を作成できませんでした");
        AppendStatus(ex);
      }

      return null; // failed
    }

    private void RenameMailbox(ImapMailboxNode node)
    {
      string newMailboxName;

      using (var dialog = new InputDialog()) {
        dialog.Text = "メールボックスの名前の変更";
        dialog.Description = "新しいメールボックスの名前を入力してください。";
        dialog.InputText = node.Mailbox.Name;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;
        else if (dialog.InputText == node.Mailbox.Name)
          return;

        newMailboxName = dialog.InputText;
      }

      RenameMailbox(node.Account, node.Mailbox, newMailboxName);
    }

    private void RenameMailbox(ImapAccount account, ImapMailbox mailbox, string newname)
    {
      try {
        var request = account.CreateRequest(mailbox.Url, ImapWebRequestMethods.Rename);

        request.AllowCreateMailbox = true;
        request.DestinationUri = new Uri(account.ServerUri, newname);

        using (var response = request.GetResponse()) {
        }

        AppendStatus("{0}の名前を{1}に変更しました", mailbox.Name, newname);
      }
      catch (WebException ex) {
        AppendStatus("{0}の名前を{1}に変更できませんでした", mailbox.Name, newname);
        AppendStatus(ex);
      }

      UpdateMailboxTree(account);
    }

    private void ShowSubscribeDialog(TreeNode node)
    {
      ImapAccount account;

      if (node is ImapAccountNode)
        account = (node as ImapAccountNode).Account;
      else if (node is ImapMailboxNode)
        account = (node as ImapMailboxNode).Account;
      else
        return;

      using (var dialog = new SubscribeMailboxDialog(account)) {
        dialog.ShowDialog(this);
      }

      UpdateMailboxTree(account);
    }

    private void UnsubscribeMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;

      try {
        node.Account.Request(mailbox.Url, ImapWebRequestMethods.Unsubscribe);

        AppendStatus("{0}を購読解除しました", mailbox.Name);
      }
      catch (WebException ex) {
        AppendStatus("{0}の購読解除に失敗しました", mailbox.Name);
        AppendStatus(ex);
      }

      UpdateMailboxTree(node.Account);
    }

    private void FetchAndSaveMessage(IMessageContainer message)
    {
      var messageUrl = message.MessageUrl;

      var subject = (message is ImapMessageItem) 
        ? TextConvert.FromMimeEncodedString((message as ImapMessageItem).Message.Envelope.Subject)
        : GetHeaderValue((message as PopMessageItem).Message, "Subject");

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.Filter = fileDialogMessageFilterString;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", messageUrl.LocalPath);
        saveFileDialog.FileName = string.Format("{0}.eml", PathUtils.ReplaceInvalidFileNameChars(subject, " "));

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        message.Account.Request(messageUrl, delegate(WebResponse response) {
          using (var stream = saveFileDialog.OpenFile()) {
            using (var responseStream = response.GetResponseStream()) {
              responseStream.WriteToEnd(stream, 0x4000);
            }
          }

          return response;
        });

        MessageBox.Show(string.Format("{0}を{1}に保存しました", messageUrl.LocalPath, saveFileDialog.FileName));
      }
    }

    private void ToggleMessageSeen(ImapMessageItem item)
    {
      StoreMessage(item,
                   item.Message.Flags.Has(ImapMessageFlag.Seen)
                     ? ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen)
                     : ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));
    }

    private void ToggleMessageKeywordsDropDownOpening(object sender, EventArgs e)
    {
      var item = sender as ListViewContextMenuItem;

      item.DropDownItems.Clear();

      var mailbox = (treeViewMailbox.SelectedNode as ImapMailboxNode).Mailbox;

      if (mailbox == null)
        return;

      foreach (ImapMessageFlag flag in mailbox.ApplicableFlags) {
        if (flag.IsSystemFlag)
          continue;

        var flagItem = new ListViewTargetFlagContextMenuItem();

        flagItem.Text = (string)flag;
        flagItem.ContextMessage = item.ContextMessage;
        flagItem.TargetFlag = flag;
        flagItem.Checked = (item.ContextMessage as ImapMessageItem).Message.Flags.Has(flag);
        flagItem.Click += delegate(object clickSender, EventArgs clickEventArgs) {
          var i = clickSender as ListViewTargetFlagContextMenuItem;
          var storeDataItem = i.Checked
            ? ImapStoreDataItem.RemoveFlags(i.TargetFlag)
            : ImapStoreDataItem.AddFlags(i.TargetFlag);

          StoreMessage(i.ContextMessage as ImapMessageItem, storeDataItem);
        };

        item.DropDownItems.Add(flagItem);
      }

      if (!mailbox.PermanentFlags.Has(ImapMessageFlag.AllowedCreateKeywords))
        return;

      toolStripMenuItemMessage_ToggleKeywords_Append.ContextMessage = item.ContextMessage;

      item.DropDownItems.Add(toolStripMenuItemMessage_ToggleKeywords_Append);
    }

    private void ExpungeMessage(IMessageContainer message)
    {
      try {
        var method = (message is ImapMessageItem)
          ? ImapWebRequestMethods.Expunge
          : PopWebRequestMethods.Dele;

        message.Account.Request(message.MessageUrl, method);

        var item = message as ListViewItem;

        if (item != null)
          listViewMessages.Items.Remove(item);

        AppendStatus("{0}を削除しました", message.MessageUrl);
      }
      catch (WebException ex) {
        AppendStatus("{0}を削除できませんでした", message.MessageUrl);
        AppendStatus(ex);
      }
    }

    private void StoreMessageKeyword(ImapMessageItem item)
    {
      using (var dialog = new InputDialog()) {
        dialog.Text = "フラグの追加";
        dialog.Description = "新しく設定するフラグの名前を指定してください。";

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        StoreMessage(item, ImapStoreDataItem.AddFlags(dialog.InputText));
      }
    }

    private void StoreMessage(ImapMessageItem item, ImapStoreDataItem storeDateItem)
    {
      var account = item.Account;
      var message = item.Message;

      try {
        var request = account.CreateRequest(message.Url, ImapWebRequestMethods.Store);

        request.StoreDataItem = storeDateItem;

        using (var response = request.GetResponse()) {
        }

        AppendStatus("{0}に'{1}'を設定しました", message.Url, storeDateItem);
      }
      catch (WebException ex) {
        AppendStatus("{0}に'{1}'を設定できませんでした", message.Url, storeDateItem);
        AppendStatus(ex);
      }

      try {
        var searchUrl = new Uri(message.Url, string.Format("?UID {0}", message.Uid));
        var updatedMessage = account.Request(searchUrl, delegate(ImapWebResponse response) {
          return response.Messages[0];
        });

        item.UpdateMessage(updatedMessage);
      }
      catch (WebException ex) {
        AppendStatus("{0}の属性を更新できませんでした", message.Url);
        AppendStatus(ex);
      }
    }

    private void CopyOrMoveMessageToDropDownOpening(object sender, EventArgs e)
    {
      var listViewContextMenuItem = sender as ListViewContextMenuItem;

      listViewContextMenuItem.DropDownItems.Clear();

      var messageItem = listViewContextMenuItem.ContextMessage as ImapMessageItem;

      if (messageItem == null)
        return;

      var copy = (sender == toolStripMenuItemMessage_CopyTo);

      foreach (var pair in imapMailboxNodeMap) {
        if (pair.Value.Account == messageItem.Account) {
          var dropDownItem = new ListViewTargetMailboxContextMenuItem();

          dropDownItem.Text = pair.Key.Name;
          dropDownItem.ContextMessage = messageItem;
          dropDownItem.TargetMailbox = pair.Key;

          if (copy)
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              CopyMessageTo(contextMessageItem, item.TargetMailbox);
            };
          else
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              MoveMessageTo(contextMessageItem, item.TargetMailbox);
            };

          listViewContextMenuItem.DropDownItems.Add(dropDownItem);
        }
      }
    }

    private void MoveMessageToTrash(ImapMessageItem item)
    {
      var trashMailbox = GetOrCreateTrashMailbox(item.Account);

      if (trashMailbox == null)
        return;

      MoveMessageTo(item, trashMailbox);
    }

    private void MoveMessageTo(ImapMessageItem item, ImapMailbox mailbox)
    {
      CopyMessageTo(item, mailbox, true);
    }

    private void CopyMessageTo(ImapMessageItem item, ImapMailbox mailbox)
    {
      CopyMessageTo(item, mailbox, false);
    }

    private void CopyMessageTo(ImapMessageItem item, ImapMailbox mailbox, bool expungeAfterCopy)
    {
      var message = item.Message;

      try {
        var request = item.Account.CreateRequest(message.Url, ImapWebRequestMethods.Copy);

        request.DestinationUri = mailbox.Url;

        using (var response = request.GetResponse()) {
        }

        AppendStatus("{0}を{1}にコピーしました", message.Url, mailbox.Name);
      }
      catch (WebException ex) {
        AppendStatus("{0}を{1}にコピーできませんでした", message.Url, mailbox.Name);
        AppendStatus(ex);
        return;
      }

      if (expungeAfterCopy)
        ExpungeMessage(item);
    }

    private void NoOp()
    {
      foreach (var account in imapAccounts) {
        account.Request(ImapWebRequestMethods.NoOp);
      }

      foreach (var account in popAccounts) {
        account.Request(PopWebRequestMethods.NoOp);
      }
    }

    private void LogoutAll()
    {
      foreach (var account in imapAccounts) {
        try {
          ImapSessionManager.DisconnectFrom(account.ServerUri);
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }

      foreach (var account in popAccounts) {
        try {
          PopSessionManager.DisconnectFrom(account.ServerUri);
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }
    }

    private void AppendStatus(WebException ex)
    {
      AppendStatus("WebExceptionStatus: {0}{1}{2}", ex.Status, Environment.NewLine, ex);
    }

    private void AppendStatus(string format, params object[] args)
    {
      statusStringBuffer.AppendFormat(format, args);
      statusStringBuffer.AppendLine();

      textBoxStatus.Text = statusStringBuffer.ToString();
      textBoxStatus.Select(textBoxStatus.TextLength - 1, 0);
      textBoxStatus.ScrollToCaret();
      textBoxStatus.Update();

      Console.Error.WriteLine(format, args);
    }

    private class TreeViewContextMenuItem : ToolStripMenuItem {
      public TreeNode ContextNode {
        get; set;
      }
    }

    private class ListViewContextMenuItem : ToolStripMenuItem {
      public IMessageContainer ContextMessage {
        get; set;
      }
    }

    private class ListViewTargetMailboxContextMenuItem : ListViewContextMenuItem {
      public ImapMailbox TargetMailbox {
        get; set;
      }
    }

    private class ListViewTargetFlagContextMenuItem : ListViewContextMenuItem {
      public ImapMessageFlag TargetFlag {
        get; set;
      }
    }

    private interface IAccountContainer {
      AccountBase Account { get; }
    }

    private class ImapAccountNode : TreeNode, IAccountContainer {
      public ImapAccount Account{
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapAccountNode(ImapAccount account)
      {
        this.Account = account;

        Text = string.Format("{0} - {1}", account.Name, account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private class PopAccountNode : TreeNode, IAccountContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopAccountNode(PopAccount account)
      {
        this.Account = account;

        Text = string.Format("{0} - {1}", account.Name, account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private interface IMailboxContainer : IAccountContainer {
      object GetMailbox();
    }

    private class ImapNonExistentMailboxNode : TreeNode {
      public ImapNonExistentMailboxNode(string text, string mailboxName)
      {
        this.Text = text;
        this.Name = mailboxName;
        this.ForeColor = SystemColors.GrayText;
      }
    }

    private class ImapMailboxNode : TreeNode, IMailboxContainer {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMailbox Mailbox {
        get; set;
      }

      public long? UnseenMessage {
        get; set;
      }

      public ImapMailboxNode(ImapAccount account, ImapMailbox mailbox)
      {
        this.Account = account;
        this.Mailbox = mailbox;
        this.UnseenMessage = null;

        this.Name = mailbox.Name;

        if (mailbox.Flags.Has(ImapMailboxFlag.NoSelect))
          this.ForeColor = SystemColors.GrayText;

        UpdateText();
      }

      public void UpdateText()
      {
        var nameHierarchy = Mailbox.Name.Split(new [] {Mailbox.HierarchyDelimiter}, StringSplitOptions.None);
        var name = nameHierarchy[nameHierarchy.Length - 1];

        if (UnseenMessage == null) {
          Text = string.Format("{0} 新着{1} 全{2}",
                               name,
                               Mailbox.RecentMessage,
                               Mailbox.ExistsMessage);
        }
        else {
          Text = string.Format("{0} 新着{1} 未読{2} 全{3}",
                               name,
                               Mailbox.RecentMessage,
                               UnseenMessage,
                               Mailbox.ExistsMessage);
        }
      }

      object IMailboxContainer.GetMailbox()
      {
        return Mailbox;
      }
    }

    private class PopMailboxNode : TreeNode, IMailboxContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopMailboxNode(PopAccount account, PopDropListing dropList)
      {
        this.Account = account;
        this.dropList = dropList;

        UpdateText();
      }

      public void UpdateText()
      {
        Text = string.Format("全{0}",
                             dropList.MessageCount);
      }

      object IMailboxContainer.GetMailbox()
      {
        return null;
      }

      private PopDropListing dropList;
    }

    private interface IMessageContainer : IAccountContainer {
      Uri MessageUrl { get; }
    }

    private class ImapMessageItem : ListViewItem, IMessageContainer {
      private static Font BoldFont;

      static ImapMessageItem()
      {
        BoldFont = new Font(Control.DefaultFont, FontStyle.Bold);
      }

      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMessage Message {
        get; private set;
      }

      Uri IMessageContainer.MessageUrl {
        get { return Message.Url; }
      }

      public ImapMessageItem(ImapAccount account, ImapMessage message)
      {
        this.Account = account;

        UpdateMessage(message);
      }

      public void UpdateMessage(ImapMessage message)
      {
        this.Message = message;

        this.SubItems.Clear();

        if (message.Envelope == null) {
          this.Text = string.Empty;
          this.SubItems.Add(message.Rfc822Size.ToString());
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(message.Flags.ToString());
        }
        else {
          this.Text = Smdn.Formats.TextConvert.FromMimeEncodedString(message.Envelope.Subject ?? string.Empty);
          this.SubItems.Add(message.Rfc822Size.ToString());
          this.SubItems.Add(message.Envelope.Date);
          this.SubItems.Add(FromAddress(message.Envelope.From));
          this.SubItems.Add(message.Flags.ToString());
        }

        if (message.Flags.Has(ImapMessageFlag.Seen))
          this.Font = Control.DefaultFont;
        else
          this.Font = BoldFont;
      }

      private string FromAddress(ImapAddress[] addresses)
      {
        if (addresses == null || addresses.Length == 0)
          return string.Empty;

        return string.Join(", ", Array.ConvertAll(addresses, delegate(ImapAddress address) {
          return string.Format("{0} <{1}@{2}>",
                               TextConvert.FromMimeEncodedString(address.Name ?? string.Empty),
                               address.Mailbox,
                               address.Host);
        }));
      }
    }

    private class PopMessageItem : ListViewItem, IMessageContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopScanListing ScanList {
        get; private set;
      }

      public MimeMessage Message {
        get; private set;
      }

      Uri IMessageContainer.MessageUrl {
        get { return new Uri(Account.ServerUri, string.Format("./;MSG={0}", ScanList.MessageNumber)); }
      }

      public PopMessageItem(PopAccount account, PopScanListing scanList, MimeMessage message)
      {
        this.Account = account;
        this.ScanList = scanList;

        UpdateMessage(message);
      }

      public void UpdateMessage(MimeMessage message)
      {
        this.Message = message;

        this.Text = MainWindow.GetHeaderValue(message, "Subject");
        this.SubItems.Add(ScanList.SizeInOctets.ToString());
        this.SubItems.Add(MainWindow.GetHeaderValue(message, "Date"));
        this.SubItems.Add(MainWindow.GetHeaderValue(message, "From"));
        this.SubItems.Add("-");
      }
    }

    private static string GetHeaderValue(MimeMessage message, string headerName)
    {
      var header = message.Headers.GetHeaders(headerName).FirstOrDefault();

      if (header == null)
        return string.Empty;
      else
        return header.Value;
    }

    private string mimeTypesFile;

    private ImapAccount[] imapAccounts;
    private PopAccount[] popAccounts;
    private AccountBase[] allAccounts;
    private readonly Dictionary<ImapMailbox, ImapMailboxNode> imapMailboxNodeMap = new Dictionary<ImapMailbox, ImapMailboxNode>();
    private StringBuilder statusStringBuffer = new StringBuilder(10240);
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
  }
}
