using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Tests;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionTest {
    class NullCredential : ICredentialsByHost {
      public NetworkCredential GetCredential(string host, int port, string authType)
      {
        return null;
      }
    }

    [SetUp]
    public void Setup()
    {
      server = new ImapPseudoServer();
      server.Start();

      host = server.ServerEndPoint.Address.ToString();
      port = server.ServerEndPoint.Port;

      credential = new NetworkCredential(username, password, host);
      uri = new Uri(string.Format("{0}://{1}@{2}:{3}",
                                  ImapUri.UriSchemeImap,
                                  username,
                                  host,
                                  port));
    }

    [TearDown]
    public void TearDown()
    {
      server.EnqueueResponse("* BYE Logging out\r\n");
      server.Stop();
    }

    private void AssertEqualCaseInsensitive(string expected, string actual)
    {
      AssertEqualCaseInsensitive(expected, actual, "sent command");
    }

    private void AssertEqualCaseInsensitive(string expected, string actual, string message)
    {
      Assert.IsTrue(string.Equals(expected, actual, StringComparison.InvariantCultureIgnoreCase),
                    message + "\r\nexpected: " + expected.Replace("\r", "\\r").Replace("\n", "\\n") + "\r\nactual: " + actual.Replace("\r", "\\r").Replace("\n", "\\n"));
    }

    private ImapSession Connect()
    {
      server.EnqueueResponse("* OK ImapPseudoServer ready\r\n");

      return new ImapSession(host, port);
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestChangeReadOnlyPropertyServerCapabilities()
    {
      using (var session = Connect()) {
        session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestChangeReadOnlyPropertyHierarchyDelimiters()
    {
      using (var session = Connect()) {
        session.HierarchyDelimiters.Add("#refname", ".");
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestChangeReadOnlyPropertyServerID()
    {
      using (var session = Connect()) {
        session.ServerID.Add("key", "value");
      }
    }

    private void Idle(Action<ImapSession> preprocessSession, Action<ImapSession> processSession)
    {
      using (var session = Authenticate("IDLE")) {
        preprocessSession(session);

        server.EnqueueResponse("+ idling\r\n");

        var asyncResult = session.BeginIdle();

        Assert.IsTrue(session.IsIdling);
        Assert.IsTrue(session.IsTransactionProceeding);

        processSession(session);

        server.EnqueueResponse("0002 OK done\r\n");

        Assert.IsTrue((bool)session.EndIdle(asyncResult));

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);
      }
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestSetTransactionTimeoutWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.TransactionTimeout = 1000;
        Assert.AreEqual(1000, session.TransactionTimeout);
      },
      delegate(ImapSession session) {
        session.TransactionTimeout = 1000;
      });
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestSetSendTimeoutWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.SendTimeout = 1000;
        Assert.AreEqual(1000, session.SendTimeout);
      },
      delegate(ImapSession session) {
        session.SendTimeout = 1000;
      });
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestSetReceiveTimeoutWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.ReceiveTimeout = 1000;
        Assert.AreEqual(1000, session.ReceiveTimeout);
      },
      delegate(ImapSession session) {
        session.ReceiveTimeout = 1000;
      });
    }

    [Test]
    public void TestConnect()
    {
      using (var session = Connect()) {
        Assert.IsFalse(session.IsIdling);
        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", host, port)), session.Authority);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      }
    }

    [Test, ExpectedException(typeof(ImapConnectionException))]
    public void TestConnectBye()
    {
      server.EnqueueResponse("* BYE not available.\r\n");

      using (var session = new ImapSession(host, port)) {
      }
    }

    [Test, ExpectedException(typeof(ImapConnectionException))]
    public void TestConnectInvalidResponse()
    {
      server.EnqueueResponse("+OK pseudo POP server ready.\r\n");

      using (var session = new ImapSession(host, port)) {
      }
    }

    [Test, ExpectedException(typeof(Smdn.Net.Imap4.Protocol.ImapConnectionException))]
    public void TestConnectTransactionTimeout()
    {
      server.Stop();

      using (var session = new ImapSession(host, port, 500)) {
      }
    }

    [Test]
    public void TestDisconnect()
    {
      var session = Connect();

      session.Disconnect(false);

      Assert.IsTrue(session.IsDisposed);
    }

    [Test]
    public void TestImap4Rev1Incapable()
    {
      using (var session = Connect()) {
        server.EnqueueResponse("* CAPABILITY NONE\r\n0000 OK ImapCapability completed\r\n");

        ImapCapabilityList caps;

        Assert.IsTrue((bool)session.Capability(out caps));
        Assert.IsNotNull(caps);
        Assert.IsFalse(caps.Has(ImapCapability.Imap4Rev1));
        AssertEqualCaseInsensitive("0000 capability\r\n",
                                   server.DequeueRequest());

        try {
          session.HandlesIncapableAsException = true;
          session.NoOp();
          Assert.Fail("exception not occured");
        }
        catch (ImapIncapableException) {
        }
      }
    }

    [Test]
    public void TestGreetingWithCapability()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=PLAIN CHILDREN THREAD=REFERENCES X-EXTENSION1 X-EXTENSION2] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(7, session.ServerCapabilities.Count);

        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Children));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.CRAMMD5));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

        Assert.IsTrue(session.ServerCapabilities.Has("THREAD=REFERENCES"));
        Assert.IsTrue(session.ServerCapabilities.Has("X-EXTENSION1"));
        Assert.IsTrue(session.ServerCapabilities.Has("X-EXTENSION2"));

        try {
          session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestGreetingPreAuthWithCapability()
    {
      server.EnqueueResponse("* PREAUTH [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.AreEqual(2, session.ServerCapabilities.Count);
        Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", host, port)), session.Authority);

        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

        try {
          session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestGreetingOkWithNoImapCapability()
    {
      server.EnqueueResponse("* OK [CAPABILITY AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsTrue(!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));

        session.HandlesIncapableAsException = true;

        session.Login(credential);
      }
    }

    [Test]
    public void TestGreetingHomeServerReferralAsException()
    {
      server.EnqueueResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n");

      try {
        new ImapSession(host, port, true);

        Assert.Fail("connected without exception");
      }
      catch (ImapLoginReferralException ex) {
        Assert.IsTrue(ex.Message.Contains("Server not accepting connections.  Try SERVER2"));
        Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER2/"), ex.ReferToUri);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapConnectionException))]
    public void TestGreetingHomeServerReferralAsError()
    {
      server.EnqueueResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n");

      using (var session = new ImapSession(host, port)) {
      }
    }

    [Test]
    public void TestTransactionTimeoutWhileTransactionProceeding()
    {
      using (var session = Connect()) {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        // not respond NOOP response termination(CRLF)
        server.EnqueueResponse("0000 OK");

        try {
          session.NoOp();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }
      }
    }

    [Test]
    public void TestSocketTimeoutWhileTransactionProceeding()
    {
      using (var session = Connect()) {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond NOOP
        //server.EnqueueResponse("0000 OK\r\n");

        try {
          session.NoOp();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }
      }
    }

    [Test]
    public void TestInternalErrorWhileTransactionProceeding()
    {
      using (var session = Connect()) {
        server.EnqueueResponse("*\r\n" + // invalid response
                               "0000 CAPABILITY Ok\r\n");

        try {
          session.Capability();
          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }
      }
    }

    [Test]
    public void TestNoOpInNonAuthenticatedState()
    {
      using (var session = Connect()) {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        // NOOP transaction
        server.EnqueueResponse("0000 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        AssertEqualCaseInsensitive("0000 NOOP\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestCapability()
    {
      using (var session = Connect()) {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueResponse("* CAPABILITY IMAP4rev1 CHILDREN\r\n" +
                               "0000 OK done\r\n");

        ImapCapabilityList capabilities;

        Assert.IsTrue((bool)session.Capability(out capabilities));

        AssertEqualCaseInsensitive("0000 CAPABILITY\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(capabilities);
        Assert.AreEqual(2, capabilities.Count);
        Assert.IsTrue(capabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(capabilities.Has(ImapCapability.Children));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(2, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Children));

        try {
          session.ServerCapabilities.Add(ImapCapability.Imap4);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestCapabilityNotHandleIncapableAsException()
    {
      using (var session = Connect()) {
        session.HandlesIncapableAsException = true;

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 1
        server.EnqueueResponse("* CAPABILITY IMAP4\r\n" +
                               "0000 OK done\r\n");

        Assert.IsTrue((bool)session.Capability());

        AssertEqualCaseInsensitive("0000 CAPABILITY\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 2
        server.EnqueueResponse("* CAPABILITY IMAP4\r\n" +
                               "0001 OK done\r\n");

        Assert.IsTrue((bool)session.Capability());

        AssertEqualCaseInsensitive("0001 CAPABILITY\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
      }
    }

    [Test]
    public void TestLogin()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.HandlesIncapableAsException = true;

        server.EnqueueResponse("0000 NO incorrect\r\n");

        Assert.IsFalse((bool)session.Login(credential));

        AssertEqualCaseInsensitive(string.Format("0000 login {0} {1}\r\n", username, password),
                                   server.DequeueRequest());

        server.EnqueueResponse("0001 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Login(credential));

        AssertEqualCaseInsensitive(string.Format("0001 login {0} {1}\r\n", username, password),
                                   server.DequeueRequest());

        Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestLoginCredentialNotFound()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse((bool)session.Login(new NullCredential()));
        Assert.IsFalse((bool)session.Login(new NullCredential(), "user"));
      }
    }

    [Test]
    public void TestLoginCredentialUsernameNull()
    {
      LoginCredentialPropertyNull(new NetworkCredential(null, "password"),
                                  "\"\" password");
    }

    [Test]
    public void TestLoginCredentialPasswordNull()
    {
      LoginCredentialPropertyNull(new NetworkCredential("username", null),
                                  "username \"\"");
    }

    private void LoginCredentialPropertyNull(ICredentialsByHost credential, string expectedArgs)
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");
      server.EnqueueResponse("0000 NO failed\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse((bool)session.Login(credential));

        AssertEqualCaseInsensitive(string.Format("0000 LOGIN {0}\r\n", expectedArgs),
                                   server.DequeueRequest());
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestLoginDisabled()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 LOGINDISABLED] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Login(credential);
      }
    }

    [Test]
    public void TestLoginHomerServerReferralAsException()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        server.EnqueueResponse("0000 NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        try {
          session.Login(credential);

          Assert.Fail("logged in without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));
          Assert.AreEqual(new Uri("IMAP://MIKE@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestLoginHomerServerReferralAsError()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, false)) {
        server.EnqueueResponse("0000 NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        Assert.IsFalse((bool)session.Login(credential));
      }
    }

    [Test]
    public void TestLoginMailboxReferralAsException()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        server.EnqueueResponse("0000 OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

        try {
          session.Login(credential);

          Assert.Fail("logged in without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user's personal mailboxes located on Server2, but public mailboxes are available."));
          Assert.AreEqual(new Uri("IMAP://MATTHEW@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestLoginMailboxReferralAsError()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, false)) {
        server.EnqueueResponse("0000 OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

        Assert.IsTrue((bool)session.Login(credential));
      }
    }

    [Test]
    public void TestAuthenticateMailboxReferralAsException()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN LOGIN-REFERRALS] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        server.EnqueueResponse("0000 NO [REFERRAL IMAP://user;AUTH=GSSAPI@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        try {
          session.Authenticate(credential, ImapAuthenticationMechanism.Login);

          Assert.Fail("authentcated without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));
          Assert.AreEqual(new Uri("IMAP://user;AUTH=GSSAPI@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestAuthenticateLogin()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("+ " + Convert.ToBase64String(latin1Encoding.GetBytes("Username:")) + "\r\n");
        server.EnqueueResponse("+ " + Convert.ToBase64String(latin1Encoding.GetBytes("Password:")) + "\r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(credential, ImapAuthenticationMechanism.Login));

        AssertEqualCaseInsensitive("0000 authenticate login\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();

        AssertEqualCaseInsensitive(username,
                                   latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

        requested = server.DequeueRequest();

        AssertEqualCaseInsensitive(password,
                                   latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=LOGIN@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticatePlain()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(credential, ImapAuthenticationMechanism.Plain));

        AssertEqualCaseInsensitive("0000 authenticate plain\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();
        var userpass = latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(new[] {"\0"}, StringSplitOptions.RemoveEmptyEntries);

        Assert.AreEqual(credential.Domain, userpass[0]);
        Assert.AreEqual(credential.UserName, userpass[1]);
        Assert.AreEqual(credential.Password, userpass[2]);

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=PLAIN@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticatePlainSaslIRCapable()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(new NetworkCredential("test", "test", "test"),
                                                 ImapAuthenticationMechanism.Plain));

        AssertEqualCaseInsensitive("0000 authenticate plain dGVzdAB0ZXN0AHRlc3Q=\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(new Uri(string.Format("imap://test;AUTH=PLAIN@{0}:{1}/", host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticateCramMd5()
    {
      AuthenticateCramMd5(false);
    }

    [Test]
    public void TestAuthenticateCramMd5SaslIRCapable()
    {
      AuthenticateCramMd5(true);
    }

    private void AuthenticateCramMd5(bool saslIRCapable)
    {
      if (saslIRCapable)
        server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5] ImapSimulatedServer ready\r\n");
      else
        server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 SASL-IR] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        var timestamp = ((long)DateTime.Now.Subtract(new DateTime(1970, 1, 1)).TotalMilliseconds).ToString();
        var challenge = latin1Encoding.GetBytes(string.Format("<{0}@{1}>", timestamp, host));

        server.EnqueueResponse("+ " + Convert.ToBase64String(challenge) + "\r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(credential, ImapAuthenticationMechanism.CRAMMD5));

        AssertEqualCaseInsensitive("0000 authenticate cram-md5\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();
        var userkeyed = latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(' ');

        Assert.AreEqual(username, userkeyed[0]);

        var keyed = new List<byte>();

        for (var index = 0; index < userkeyed[1].Length; index += 2) {
          keyed.Add(Convert.ToByte(userkeyed[1].Substring(index, 2), 16));
        }

        Assert.AreEqual(Convert.ToBase64String((new HMACMD5(latin1Encoding.GetBytes(password))).ComputeHash(challenge)),
                        Convert.ToBase64String(keyed.ToArray()));

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=CRAM-MD5@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticateInvalidResponse()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("+ xxxxx-invalid-response-xxxxx\r\n");

        try {
          session.Authenticate(new NetworkCredential("test", "test", "test"),
                               ImapAuthenticationMechanism.DigestMD5);
          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException) {
        }

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
      }
    }

    [Test]
    public void TestAuthenticateCancelExchanging()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("+ eD0xLHk9Mix6PTM=\r\n"); // x=1,y=2,z=3
        server.EnqueueResponse("0000 NO AUTHENTICATE failed.\r\n");

        session.Authenticate(new NetworkCredential("test", "test", "test"),
                             ImapAuthenticationMechanism.DigestMD5);

        AssertEqualCaseInsensitive("0000 authenticate digest-md5\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("*\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateCredentialNotFound()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse((bool)session.Authenticate(new NullCredential(), ImapAuthenticationMechanism.DigestMD5));
        Assert.IsFalse((bool)session.Authenticate(new NullCredential(), "user", ImapAuthenticationMechanism.DigestMD5));
      }
    }

    [Test]
    public void TestAuthenticateCredentialUsernameNull()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthenticateCredentialPasswordNull()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential("username", null));
    }

    private void AuthenticateCredentialPropertyNull(ICredentialsByHost credential)
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");
      server.EnqueueResponse("+ \r\n");
      server.EnqueueResponse("0000 NO failed\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse((bool)session.Authenticate(credential,
                                                  ImapAuthenticationMechanism.Plain));

        AssertEqualCaseInsensitive("0000 AUTHENTICATE PLAIN\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("*\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateCredentialUsernameNullSaslIRCapable()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthenticateCredentialPasswordNullSaslIRCapable()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential("username", null));
    }

    private void AuthenticateCredentialPropertyNullSaslIRCapable(ICredentialsByHost credential)
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse((bool)session.Authenticate(credential,
                                                  ImapAuthenticationMechanism.Plain));
      }
    }

    [Test]
    public void TestAuthenticateOkWithCapabilityResponse()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.AreEqual(3, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.SaslIR));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

        //server.EnqueueResponse();
        server.EnqueueResponse("0000 OK [CAPABILITY IMAP4rev1] authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(new NetworkCredential("test", "test", "test"),
                                                 ImapAuthenticationMechanism.Plain));

        AssertEqualCaseInsensitive("0000 AUTHENTICATE PLAIN dGVzdAB0ZXN0AHRlc3Q=\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(new Uri(string.Format("imap://test;AUTH=PLAIN@{0}:{1}/", host, port)), session.Authority);

        Assert.AreEqual(1, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));

        try {
          session.ServerCapabilities.Add(ImapCapability.Imap4);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestLoginInAuthenticatedState()
    {
      using (var session = Authenticate()) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var result = session.Login(credential);

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, Transaction.ImapTransactionResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      }
    }

    [Test]
    public void TestAuthenticateInAuthenticatedState()
    {
      using (var session = Authenticate()) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var result = session.Authenticate(credential, ImapAuthenticationMechanism.Login);

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, Transaction.ImapTransactionResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      }
    }

    [Test]
    public void TestLogoutInNonAuthenticatedState()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");
      server.EnqueueResponse("* BYE Logging out\r\n" + 
                             "0000 OK logged out.\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsTrue(session.Logout().Code == Transaction.ImapTransactionResultCode.Bye);

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);
      }
    }

    [Test]
    public void TestLogoutInAuthenticatedState()
    {
      using (var session = Authenticate()) {
        server.EnqueueResponse("* BYE Logging out\r\n" + 
                               "0001 OK logged out.\r\n");

        Assert.IsTrue(session.Logout().Code == Transaction.ImapTransactionResultCode.Bye);

        AssertEqualCaseInsensitive("0002 logout\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestCloseInNonAuthenticatedState()
    {
      using (var session = Connect()) {
        session.Close();
      }
    }

    [Test]
    public void TestCloseInAuthenticatedState()
    {
      using (var session = Authenticate()) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);

        var result = session.Close();

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, Transaction.ImapTransactionResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    [Test]
    public void TestStartTls()
    {
      using (var session = Connect()) {
        session.HandlesIncapableAsException = true;
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                               "0000 OK CAPABILITY completed\r\n");

        Assert.IsTrue((bool)session.Capability());

        server.DequeueRequest();

        server.EnqueueResponse("0001 OK Begin TLS negotiation now\r\n");

        var prevAuthority = new Uri(session.Authority.ToString());
        var streamUpgraded = false;

        Assert.IsFalse(session.IsSecureConnection);
        Assert.IsTrue((bool)session.StartTls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        }));

        Assert.IsTrue(streamUpgraded, "stream upgraded");

        AssertEqualCaseInsensitive("0001 starttls\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(prevAuthority, session.Authority);

        Assert.Ignore("TLS negotiation is not implemented");
        Assert.IsTrue(session.IsSecureConnection);
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestStartTlsIncapable()
    {
      using (var session = Connect()) {
        server.EnqueueResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                               "0000 OK CAPABILITY completed\r\n");

        Assert.IsTrue((bool)session.Capability());

        server.DequeueRequest();

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        session.HandlesIncapableAsException = true;
        session.StartTls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        });

        Assert.IsFalse(streamUpgraded, "stream upgraded");
      }
    }

    [Test]
    public void TestStartTlsBadResponse()
    {
      using (var session = Connect()) {
        server.EnqueueResponse("0000 BAD command unknown\r\n");

        var prevAuthority = new Uri(session.Authority.ToString());

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        Assert.IsFalse((bool)session.StartTls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        }));

        Assert.IsFalse(streamUpgraded, "stream upgraded");

        AssertEqualCaseInsensitive("0000 starttls\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(prevAuthority, session.Authority);

        Assert.IsFalse(session.IsSecureConnection);
      }
    }

    [Test]
    public void TestListStatusSelect()
    {
      // greeting and login transaction
      using (var session = Connect()) {
        server.EnqueueResponse("0000 OK authenticated\r\n");
  
        session.Login(credential);

        Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);

        server.DequeueRequest();
  
        // CAPABILITY transaction
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0001 OK CAPABILITY completed\r\n");
  
        session.Capability();
  
        server.DequeueRequest();
  
        // LIST transaction
        server.EnqueueResponse("* LIST (\\HasChildren) \".\" \"INBOX\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        Assert.IsTrue((bool)session.List("*", out mailboxes));
  
        AssertEqualCaseInsensitive("0002 list \"\" \"*\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.HasChildren));
        Assert.AreEqual(new Uri(uri, "./INBOX"), mailboxes[0].Url);
  
        // LSUB transaction
        server.EnqueueResponse("* LSUB (\\HasChildren) \"/\" \"Trash\"\r\n" +
                               "* LSUB (\\HasNoChildren) \"/\" \"Trash/mail\"\r\n" +
                               "* LSUB (\\HasNoChildren) \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "* LSUB (\\HasNoChildren) \".\" \"INBOX.&U,BTFw-\"\r\n" +
                               "0003 OK LSUB completed\r\n");
  
        Assert.IsTrue((bool)session.Lsub("*", out mailboxes));
  
        AssertEqualCaseInsensitive("0003 lsub \"\" \"*\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(4, mailboxes.Length);
        Assert.AreEqual("Trash", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.HasChildren));
  
        Assert.AreEqual("Trash/mail", mailboxes[1].Name);
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.HasNoChildren));
        Assert.AreEqual(new Uri(uri, "./Trash/mail"), mailboxes[1].Url);
  
        Assert.AreEqual("INBOX.日本語", mailboxes[2].Name);
        Assert.AreEqual("INBOX.台北", mailboxes[3].Name);
  
        // STATUS transaction
        server.EnqueueResponse("* STATUS \"Trash\" (MESSAGES 3 RECENT 1 UIDVALIDITY 123456 UNSEEN 2 UIDNEXT 4)\r\n" +
                               "0004 OK STATUS completed\r\n");
  
        Assert.IsTrue((bool)session.Status(mailboxes[0], ImapStatusDataItem.StandardAll));
  
        AssertEqualCaseInsensitive("0004 status \"Trash\" (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, mailboxes[0].ExistsMessage);
        Assert.AreEqual(1, mailboxes[0].RecentMessage);
        Assert.AreEqual(123456, mailboxes[0].UidValidity);
        Assert.AreEqual(2, mailboxes[0].UnseenMessage);
        Assert.AreEqual(4, mailboxes[0].UidNext);
  
        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);
  
        server.EnqueueResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                               "* 15 EXISTS\r\n" +
                               "* 2 RECENT\r\n" +
                               "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                               "* OK [UIDNEXT 16]\r\n" +
                               "* OK [UNSEEN 13]\r\n" +
                               "0005 OK [READ-WRITE] SELECT completed\r\n");
  
        var mailbox = mailboxes[2];
  
        Assert.IsTrue((bool)session.Select(mailbox));
  
        AssertEqualCaseInsensitive("0005 select \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.AreSame(mailbox, session.SelectedMailbox);
  
        Assert.AreEqual("INBOX.日本語", mailbox.Name);
        Assert.AreEqual(15, mailbox.ExistsMessage);
        Assert.AreEqual(2, mailbox.RecentMessage);
        Assert.AreEqual(0, mailbox.UnseenMessage);
        Assert.AreEqual(13, mailbox.FirstUnseen);
        Assert.AreEqual(1202674433, mailbox.UidValidity);
        Assert.AreEqual(16, mailbox.UidNext);
        Assert.IsFalse(mailbox.ReadOnly);
        Assert.IsTrue(mailbox.UidPersistent);
        Assert.AreEqual(new Uri(uri, "./INBOX.日本語;UIDVALIDITY=1202674433"), mailbox.Url);

        /*
        Assert.AreEqual(selectedMailbox.Name, mailbox.Name);
        Assert.AreEqual(selectedMailbox.ExistsMessage, mailbox.ExistsMessage);
        Assert.AreEqual(selectedMailbox.RecentMessage, mailbox.RecentMessage);
        Assert.AreEqual(selectedMailbox.UnseenMessage, mailbox.UnseenMessage);
        Assert.AreEqual(selectedMailbox.FirstUnseen, mailbox.FirstUnseen);
        Assert.AreEqual(selectedMailbox.UidValidity, mailbox.UidValidity);
        Assert.AreEqual(selectedMailbox.UidNext, mailbox.UidNext);
        Assert.AreEqual(selectedMailbox.ReadOnly, mailbox.ReadOnly);
        Assert.AreEqual(selectedMailbox.ApplicableFlags.Count, mailbox.ApplicableFlags.Count);
        Assert.AreEqual(selectedMailbox.PermanentFlags.Count, mailbox.PermanentFlags.Count);
        */

        Assert.AreEqual(7, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Has("custom1"));
        Assert.IsTrue(mailbox.ApplicableFlags.Has("custom2"));
  
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.PermanentFlags.Has("custom1"));
  
        // CLOSE transaction
        server.EnqueueResponse("0006 OK CLOSE completed\r\n");
  
        Assert.IsTrue((bool)session.Close());
  
        AssertEqualCaseInsensitive("0006 close\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    [Test]
    public void TestListRoot()
    {
      using (var session = Authenticate()) {
        Assert.IsFalse(session.HierarchyDelimiters.ContainsKey(""));
        Assert.IsFalse(session.HierarchyDelimiters.ContainsKey("#news.comp.mail.misc"));

        // LIST transaction 1
        server.EnqueueResponse("* LIST (\\Noselect) \"/\" \"\"\r\n" +
                               "0002 OK LIST Completed\r\n");

        ImapMailboxList root;

        Assert.IsTrue((bool)session.ListRoot(out root));

        AssertEqualCaseInsensitive("0002 LIST \"\" \"\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(root);
        Assert.IsEmpty(root.Name);
        Assert.AreEqual("/", root.HierarchyDelimiter);
        Assert.AreEqual(1, root.NameAttributes.Count);
        Assert.IsTrue(root.NameAttributes.Has(ImapMailboxFlag.NoSelect));

        Assert.IsTrue(session.HierarchyDelimiters.ContainsKey(""));
        Assert.AreEqual("/", session.HierarchyDelimiters[""]);

        // LIST transaction 2
        server.EnqueueResponse("* LIST (\\Noselect) \".\" #news.\r\n" +
                               "0003 OK LIST Completed\r\n");

        Assert.IsTrue((bool)session.ListRoot("#news.comp.mail.misc", out root));

        AssertEqualCaseInsensitive("0003 LIST \"#news.comp.mail.misc\" \"\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(root);
        Assert.AreEqual("#news.", root.Name);
        Assert.AreEqual(".", root.HierarchyDelimiter);
        Assert.AreEqual(1, root.NameAttributes.Count);
        Assert.IsTrue(root.NameAttributes.Has(ImapMailboxFlag.NoSelect));

        Assert.IsTrue(session.HierarchyDelimiters.ContainsKey("#news.comp.mail.misc"));
        Assert.AreEqual(".", session.HierarchyDelimiters["#news.comp.mail.misc"]);

        try {
          session.HierarchyDelimiters.Add("#refname", "/");
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestList()
    {
      ListRList(false);
    }

    [Test]
    public void TestRList()
    {
      ListRList(true);
    }

    private void ListRList(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RList(out mailboxes));

          AssertEqualCaseInsensitive("0002 RLIST \"\" *\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.List(out mailboxes));

          AssertEqualCaseInsensitive("0002 LIST \"\" *\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Length);

        Assert.AreEqual("blurdybloop", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.NoSelect), "mailbox list #2 flags");

        Assert.AreEqual("foo/bar", mailboxes[2].Name, "mailbox list #3 name");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");
      }
    }

    [Test]
    public void TestListNonEmptyReferenceNameAndEmptyMailboxName()
    {
      ListRListNonEmptyReferenceNameAndEmptyMailboxName(false);
    }

    [Test]
    public void TestRListNonEmptyReferenceNameAndEmptyMailboxName()
    {
      ListRListNonEmptyReferenceNameAndEmptyMailboxName(true);
    }

    private void ListRListNonEmptyReferenceNameAndEmptyMailboxName(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LIST transaction
        server.EnqueueResponse("* LIST (\\Noselect) \".\" #news.\r\n" +
                               "0002 OK LIST Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RList("#news.comp.mail.misc", string.Empty, out mailboxes));

          AssertEqualCaseInsensitive("0002 RLIST \"#news.comp.mail.misc\" \"\"\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.List("#news.comp.mail.misc", string.Empty, out mailboxes));

          AssertEqualCaseInsensitive("0002 LIST \"#news.comp.mail.misc\" \"\"\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(1, mailboxes.Length);

        Assert.AreEqual("#news.", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox list #1 flag count");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoSelect), "mailbox list #1 flags");
      }
    }

    [Test]
    public void TestListNonEmptyReferenceNameAndWildcardMailboxName()
    {
      ListRListNonEmptyReferenceNameAndWildcardMailboxName(false);
    }

    [Test]
    public void TestRListNonEmptyReferenceNameAndWildcardMailboxName()
    {
      ListRListNonEmptyReferenceNameAndWildcardMailboxName(true);
    }

    private void ListRListNonEmptyReferenceNameAndWildcardMailboxName(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LIST transaction
        server.EnqueueResponse("* LIST (\\Noselect) \"/\" ~/Mail/foo\r\n" +
                               "* LIST () \"/\" ~/Mail/meetings\r\n" +
                               "0002 OK LIST Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RList("~/Mail/", "%", out mailboxes));

          AssertEqualCaseInsensitive("0002 RLIST \"~/Mail/\" \"%\"\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.List("~/Mail/", "%", out mailboxes));

          AssertEqualCaseInsensitive("0002 LIST \"~/Mail/\" \"%\"\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Length);

        Assert.AreEqual("~/Mail/foo", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox list #1 flag count");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoSelect), "mailbox list #1 flags");

        Assert.AreEqual("~/Mail/meetings", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
      }
    }

    [Test]
    public void TestLsub()
    {
      LsubRLsub(false);
    }

    [Test]
    public void TestRLsub()
    {
      LsubRLsub(true);
    }

    public void LsubRLsub(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LSUB transaction
        server.EnqueueResponse("* LSUB () \"/\" INBOX\r\n" +
                               "* LSUB () \"/\" INBOX/Sent\r\n" +
                               "* LSUB () \"/\" INBOX/Drafts\r\n" +
                               "* LSUB () \"/\" INBOX/Junk\r\n" +
                                "0002 OK LSUB Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RLsub(out mailboxes));

          AssertEqualCaseInsensitive("0002 RLSUB \"\" *\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.Lsub(out mailboxes));

          AssertEqualCaseInsensitive("0002 LSUB \"\" *\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Length);

        Assert.AreEqual("INBOX", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("INBOX/Sent", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");

        Assert.AreEqual("INBOX/Drafts", mailboxes[2].Name, "mailbox list #3 name");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");

        Assert.AreEqual("INBOX/Junk", mailboxes[3].Name, "mailbox list #4 name");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox list #4 delimiter");
        Assert.AreEqual(0, mailboxes[3].Flags.Count, "mailbox list #4 flag count");
      }
    }

    [Test]
    public void TestLsubEmptyReferenceNameAndNonEmptyMailboxName()
    {
      LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(false);
    }

    [Test]
    public void TestRLsubEmptyReferenceNameAndNonEmptyMailboxName()
    {
      LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(true);
    }

    private void LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LSUB transaction
        server.EnqueueResponse("* LSUB () \"/\" INBOX/Sent\r\n" +
                               "* LSUB () \"/\" INBOX/Drafts\r\n" +
                               "* LSUB () \"/\" INBOX/Junk\r\n" +
                                "0002 OK LSUB Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RLsub("INBOX/*", out mailboxes));

          AssertEqualCaseInsensitive("0002 RLSUB \"\" \"INBOX/*\"\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.Lsub("INBOX/*", out mailboxes));

          AssertEqualCaseInsensitive("0002 LSUB \"\" \"INBOX/*\"\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Length);

        Assert.AreEqual("INBOX/Sent", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("INBOX/Drafts", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");

        Assert.AreEqual("INBOX/Junk", mailboxes[2].Name, "mailbox list #3 name");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");
      }
    }

    [Test]
    public void TestLsubNonEmptyReferenceNameAndWildcardMailboxName()
    {
      LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(false);
    }

    [Test]
    public void TestRLsubNonEmptyReferenceNameAndWildcardMailboxName()
    {
      LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(true);
    }

    private void LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(bool mailboxReferral)
    {
      using (var session = Authenticate(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0])) {
        // LSUB transaction
        server.EnqueueResponse("* LSUB () \".\" #news.comp.mail.mime\r\n" +
                               "* LSUB () \".\" #news.comp.mail.misc\r\n" +
                                "0002 OK LSUB Completed\r\n");

        ImapMailbox[] mailboxes;

        if (mailboxReferral) {
          Assert.IsTrue((bool)session.RLsub("#news.", "comp.mail.*", out mailboxes));

          AssertEqualCaseInsensitive("0002 RLSUB \"#news.\" \"comp.mail.*\"\r\n",
                                     server.DequeueRequest());
        }
        else {
          Assert.IsTrue((bool)session.Lsub("#news.", "comp.mail.*", out mailboxes));

          AssertEqualCaseInsensitive("0002 LSUB \"#news.\" \"comp.mail.*\"\r\n",
                                     server.DequeueRequest());
        }

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Length);

        Assert.AreEqual("#news.comp.mail.mime", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("#news.comp.mail.misc", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestXListIncapable()
    {
      using (var session = Authenticate()) {
        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.GimapXlist));

        session.HandlesIncapableAsException = true;

        ImapMailbox[] mailboxes;

        session.XList(out mailboxes);
      }
    }

    [Test]
    public void TestXList()
    {
      using (var session = Authenticate("XLIST")) {
        // LIST transaction
        server.EnqueueResponse("* XLIST (\\HasNoChildren \\Inbox) \"/\" \"&U9dP4TDIMOwwpA-\"\r\n" +
                               "* XLIST (\\Noselect \\HasChildren) \"/\" \"[Gmail]\"\r\n" +
                               "* XLIST (\\HasNoChildren \\AllMail) \"/\" \"[Gmail]/&MFkweTBmMG4w4TD8MOs-\"\r\n" +
                               "* XLIST (\\HasNoChildren \\Trash) \"/\" \"[Gmail]/&MLQw33ux-\"\r\n" +
                               "* XLIST (\\HasNoChildren \\Starred) \"/\" \"[Gmail]/&MLkwvzD8TtgwTQ-\"\r\n" +
                               "* XLIST (\\HasNoChildren \\Drafts) \"/\" \"[Gmail]/&Tgtm+DBN-\"\r\n" +
                               "* XLIST (\\HasNoChildren \\Spam) \"/\" \"[Gmail]/&j,dg0TDhMPww6w-\"\r\n" +
                               "* XLIST (\\HasNoChildren \\Sent) \"/\" \"[Gmail]/&kAFP4W4IMH8w4TD8MOs-\"\r\n" +
                               "0002 OK LIST Completed\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.XList(out mailboxes));

        AssertEqualCaseInsensitive("0002 XLIST \"\" *\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(8, mailboxes.Length);

        Assert.AreEqual("受信トレイ", mailboxes[0].Name, "mailbox list #1 name");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Inbox), "mailbox list #1 flags");

        Assert.AreEqual("[Gmail]", mailboxes[1].Name, "mailbox list #2 name");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.NoSelect), "mailbox list #2 flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox list #2 flags");

        Assert.AreEqual("[Gmail]/すべてのメール", mailboxes[2].Name, "mailbox list #3 name");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.AllMail), "mailbox list #3 flags");

        Assert.AreEqual("[Gmail]/ゴミ箱", mailboxes[3].Name, "mailbox list #4 name");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.Trash), "mailbox list #4 flags");

        Assert.AreEqual("[Gmail]/スター付き", mailboxes[4].Name, "mailbox list #5 name");
        Assert.IsTrue(mailboxes[4].Flags.Has(ImapMailboxFlag.Starred), "mailbox list #5 flags");

        Assert.AreEqual("[Gmail]/下書き", mailboxes[5].Name, "mailbox list #6 name");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.Drafts), "mailbox list #6 flags");

        Assert.AreEqual("[Gmail]/迷惑メール", mailboxes[6].Name, "mailbox list #7 name");
        Assert.IsTrue(mailboxes[6].Flags.Has(ImapMailboxFlag.Spam), "mailbox list #7 flags");

        Assert.AreEqual("[Gmail]/送信済みメール", mailboxes[7].Name, "mailbox list #8 name");
        Assert.IsTrue(mailboxes[7].Flags.Has(ImapMailboxFlag.Sent), "mailbox list #8 flags");
      }
    }

    [Test]
    public void TestListNonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("LIST");
    }

    [Test]
    public void TestLsubNonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("LSUB");
    }

    [Test]
    public void TestRListNonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("RLIST");
    }

    [Test]
    public void TestRLsubNonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("RLSUB");
    }

    [Test]
    public void TestXListNonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("XLIST");
    }

    private void ListNonAsciiMailboxName(string command)
    {
      using (var session = Authenticate("XLIST", "MAILBOX-REFERRAL")) {
        // list transaction
        server.EnqueueResponse("0002 OK completed\r\n");

        ImapMailbox[] mailboxes;
        Transaction.ImapTransactionResult result;

        switch (command) {
          case "LIST":  result = session.List("INBOX.メールボックス.*", out mailboxes); break;
          case "LSUB":  result = session.Lsub("INBOX.メールボックス.*", out mailboxes); break;
          case "RLIST": result = session.RList("INBOX.メールボックス.*", out mailboxes); break;
          case "RLSUB": result = session.RLsub("INBOX.メールボックス.*", out mailboxes); break;
          case "XLIST": result = session.XList("INBOX.メールボックス.*", out mailboxes); break;
          default:
            return;
        }

        Assert.IsTrue((bool)result);

        AssertEqualCaseInsensitive(string.Format("0002 {0} \"\" \"INBOX.&MOEw,DDrMNwwwzCvMLk-.*\"\r\n", command),
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(0, mailboxes.Length);
      }
    }

    [Test]
    public void TestListExtendedSelectSubscribed()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST (\\Marked \\NoInferiors \\Subscribed) \"/\" \"inbox\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Fruit/Banana\"\r\n" +
                               "* LIST (\\Subscribed \\NonExistent) \"/\" \"Fruit/Peach\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Vegetable\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Vegetable/Broccoli\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("*",
                                                 ImapListSelectionOptions.Subscribed,
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST (SUBSCRIBED) \"\" \"*\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(5, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox #1 delimiter");
        Assert.AreEqual(3, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #1 flags");

        Assert.AreEqual("Fruit/Banana", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox #2 delimiter");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #2 flags");

        Assert.AreEqual("Fruit/Peach", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox #3 delimiter");
        Assert.AreEqual(2, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #3 flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.NonExistent), "mailbox #3 flags");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox #4 delimiter");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("Vegetable/Broccoli", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual("/", mailboxes[4].HierarchyDelimiter, "mailbox #5 delimiter");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #5 flags");
      }
    }

    [Test]
    public void TestListExtendedReturnChildren()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST (\\Marked \\NoInferiors) \"/\" \"inbox\"\r\n" +
                               "* LIST (\\HasChildren) \"/\" \"Fruit\"\r\n" +
                               "* LIST (\\HasNoChildren) \"/\" \"Tofu\"\r\n" +
                               "* LIST (\\HasChildren) \"/\" \"Vegetable\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("%",
                                                 ImapListReturnOptions.Children, out mailboxes));

        //AssertEqualCaseInsensitive("0002 LIST () \"\" \"%\" RETURN (CHILDREN)\r\n",
        AssertEqualCaseInsensitive("0002 LIST \"\" \"%\" RETURN (CHILDREN)\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox #1 delimiter");
        Assert.AreEqual(2, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");

        Assert.AreEqual("Fruit", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox #2 delimiter");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox #2 flags");

        Assert.AreEqual("Tofu", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox #3 delimiter");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.HasNoChildren), "mailbox #3 flags");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox #4 delimiter");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox #4 flags");
      }
    }

    [Test]
    public void TestListExtendedSelectRemoteReturnChildren()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST (\\Marked \\NoInferiors) \"/\" \"inbox\"\r\n" +
                               "* LIST (\\HasChildren) \"/\" \"Fruit\"\r\n" +
                               "* LIST (\\HasNoChildren) \"/\" \"Tofu\"\r\n" +
                               "* LIST (\\HasChildren) \"/\" \"Vegetable\"\r\n" +
                               "* LIST (\\Remote) \"/\" \"Bread\"\r\n" +
                               "* LIST (\\HasChildren \\Remote) \"/\" \"Meat\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("%",
                                                 ImapListSelectionOptions.Remote,
                                                 ImapListReturnOptions.Children,
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST (REMOTE) \"\" \"%\" RETURN (CHILDREN)\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(6, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(2, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");

        Assert.AreEqual("Fruit", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox #2 flags");

        Assert.AreEqual("Tofu", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.HasNoChildren), "mailbox #3 flags");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox #4 flags");

        Assert.AreEqual("Bread", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Has(ImapMailboxFlag.Remote), "mailbox #5 flags");

        Assert.AreEqual("Meat", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(2, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.HasChildren), "mailbox #6 flags");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.Remote), "mailbox #6 flags");
      }
    }

    [Test]
    public void TestListExtendedSelectRemoteSubscribed()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST (\\Marked \\NoInferiors \\Subscribed) \"/\" \"inbox\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Fruit/Banana\"\r\n" +
                               "* LIST (\\Subscribed \\NonExistent) \"/\" \"Fruit/Peach\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Vegetable\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"Vegetable/Broccoli\"\r\n" +
                               "* LIST (\\Remote \\Subscribed) \"/\" \"Bread\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("*",
                                                 ImapListSelectionOptions.Remote + ImapListSelectionOptions.Subscribed,
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST (REMOTE SUBSCRIBED) \"\" \"*\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(6, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(3, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #1 flags");

        Assert.AreEqual("Fruit/Banana", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #2 flags");

        Assert.AreEqual("Fruit/Peach", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(2, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #3 flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.NonExistent), "mailbox #3 flags");
        
        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("Vegetable/Broccoli", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #5 flags");

        Assert.AreEqual("Bread", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(2, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.Remote), "mailbox #6 flags");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
      }
    }

    [Test]
    public void TestListExtendedSelectRecursiveMatchSubscribed()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" \"foo2\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"foo2/bar2\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"baz2/bar2\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"baz2/bar22\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"baz2/bar222\"\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"eps2\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n" +
                               "* LIST (\\Subscribed) \"/\" \"qux2/bar2\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("*2",
                                                 ImapListSelectionOptions.RecursiveMatch + ImapListSelectionOptions.Subscribed,
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST (RECURSIVEMATCH SUBSCRIBED) \"\" \"*2\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(7, mailboxes.Length);

        Assert.AreEqual("foo2", mailboxes[0].Name, "mailbox #1");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        // TODO: child info

        Assert.AreEqual("foo2/bar2", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #2 flags");

        Assert.AreEqual("baz2/bar2", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #3 flags");

        Assert.AreEqual("baz2/bar22", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("baz2/bar222", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #5 flags");

        Assert.AreEqual("eps2", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(1, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
        // TODO: child info

        Assert.AreEqual("qux2/bar2", mailboxes[6].Name, "mailbox #6");
        Assert.AreEqual(1, mailboxes[6].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[6].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
      }
    }

    [Test]
    public void TestListExtendedMultipleMailboxPattern()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" \"INBOX\"\r\n" +
                               "* LIST (\\NoInferiors) \"/\" \"Drafts\"\r\n" +
                               "* LIST () \"/\" \"Sent/March2004\"\r\n" +
                               "* LIST (\\Marked) \"/\" \"Sent/December2003\"\r\n" +
                               "* LIST () \"/\" \"Sent/August2004\"\r\n" +
                               "0002 OK done\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended(out mailboxes,
                                                 "INBOX",
                                                 "Drafts",
                                                 "Sent/%"));

        AssertEqualCaseInsensitive("0002 LIST \"\" (\"INBOX\" \"Drafts\" \"Sent/%\")\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(5, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");

        Assert.AreEqual("Drafts", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.NoInferiors), "mailbox #2 flags");

        Assert.AreEqual("Sent/March2004", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox #3 count of flags");

        Assert.AreEqual("Sent/December2003", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Has(ImapMailboxFlag.Marked), "mailbox #4 flags");

        Assert.AreEqual("Sent/August2004", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(0, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestListExtendedReturnStatusIncapable()
    {
      using (var session = Authenticate("LIST-EXTENDED")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));
        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.ListStatus));

        session.HandlesIncapableAsException = true;

        ImapMailbox[] mailboxes;

        session.ListExtended("%", ImapListReturnOptions.StatusDataItems(ImapStatusDataItem.Messages), out mailboxes);
      }
    }

    [Test]
    public void TestListExtendedReturnStatus()
    {
      using (var session = Authenticate("LIST-EXTENDED", "LIST-STATUS")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListStatus));

        // LIST transaction
        server.EnqueueResponse("* LIST () \".\"  \"INBOX\"\r\n" +
                               "* STATUS \"INBOX\" (MESSAGES 17 UNSEEN 16)\r\n" +
                               "* LIST () \".\" \"foo\"\r\n" +
                               "* STATUS \"foo\" (MESSAGES 30 UNSEEN 29)\r\n" +
                               "* LIST (\\NoSelect) \".\" \"bar\"\r\n" +
                               "0002 OK List completed.\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("%",
                                                 ImapListReturnOptions.StatusDataItems(ImapStatusDataItem.Messages +
                                                                                       ImapStatusDataItem.Unseen),
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST \"\" \"%\" RETURN (STATUS (MESSAGES UNSEEN))\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.AreEqual(17L, mailboxes[0].ExistsMessage, "mailbox #1 exists messages");
        Assert.AreEqual(16L, mailboxes[0].UnseenMessage, "mailbox #1 unseen messages");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.AreEqual(30L, mailboxes[1].ExistsMessage, "mailbox #2 exists messages");
        Assert.AreEqual(29L, mailboxes[1].UnseenMessage, "mailbox #2 unseen messages");

        Assert.AreEqual("bar", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Has(ImapMailboxFlag.NoSelect), "mailbox #3 flags");
        Assert.AreEqual(0L, mailboxes[2].ExistsMessage, "mailbox #3 exists messages");
        Assert.AreEqual(0L, mailboxes[2].UnseenMessage, "mailbox #3 unseen messages");
      }
    }

    [Test]
    public void TestListExtendedReturnStatusSelectSubscribedRecursiveMatch()
    {
      using (var session = Authenticate("LIST-EXTENDED", "LIST-STATUS")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ListStatus));

        // LIST transaction
        server.EnqueueResponse("* LIST (\\Subscribed) \".\"  \"INBOX\"\r\n" +
                               "* STATUS \"INBOX\" (MESSAGES 17)\r\n" +
                               "* LIST () \".\" \"foo\" (CHILDINFO (\"SUBSCRIBED\"))\r\n" +
                               "0002 OK List completed.\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.ListExtended("%",
                                                 ImapListSelectionOptions.Subscribed + ImapListSelectionOptions.RecursiveMatch,
                                                 ImapListReturnOptions.StatusDataItems(ImapStatusDataItem.Messages),
                                                 out mailboxes));

        AssertEqualCaseInsensitive("0002 LIST (SUBSCRIBED RECURSIVEMATCH) \"\" \"%\" RETURN (STATUS (MESSAGES))\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Length);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.Subscribed), "mailbox #1 flags");
        Assert.AreEqual(17L, mailboxes[0].ExistsMessage, "mailbox #1 exists messages");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        // child info
      }
    }

    private ImapSession SelectMailbox(params string[] additionalCapabilities)
    {
      // greeting and login transaction
      var session = Connect();

      server.EnqueueResponse("0000 OK authenticated\r\n");

      session.Login(credential);

      var expectedAuthority = new Uri(string.Format("imap://{0};AUTH=LOGIN@{1}:{2}/", username, host, port));

      Assert.AreEqual(expectedAuthority, session.Authority);

      server.DequeueRequest();

      // CAPABILITY transaction
      server.EnqueueResponse(string.Format("* CAPABILITY IMAP4rev1 {0}\r\n", string.Join(" ", additionalCapabilities)) +
                             "0001 OK CAPABILITY completed\r\n");

      session.Capability();

      server.DequeueRequest();

      // LIST transaction
      server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                             "0002 OK LIST completed\r\n");

      ImapMailbox[] mailboxes;

      session.List(out mailboxes);

      server.DequeueRequest();

      Assert.AreEqual(new Uri(expectedAuthority, "./INBOX"), mailboxes[0].Url);

      // SELECT transaction
      Assert.IsNull(session.SelectedMailbox);

      server.EnqueueResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                             "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                             "* 15 EXISTS\r\n" +
                             "* 2 RECENT\r\n" +
                             "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                             "* OK [UIDNEXT 16]\r\n" +
                             "* OK [UNSEEN 13]\r\n" +
                             "0003 OK [READ-WRITE] SELECT completed\r\n");

      session.Select(mailboxes[0]);

      Assert.AreEqual(new Uri(expectedAuthority, "./INBOX;UIDVALIDITY=1202674433"), mailboxes[0].Url);
      Assert.AreEqual(mailboxes[0], session.SelectedMailbox);

      server.DequeueRequest();

      session.HandlesIncapableAsException = true;

      return session;
    }

    private void CloseMailbox(ImapSession session)
    {
      CloseMailbox(session, "0005");
    }

    private void CloseMailbox(ImapSession session, string tag)
    {
      // CLOSE transaction
      server.EnqueueResponse(tag + " OK CLOSE completed\r\n");

      session.Close();

      AssertEqualCaseInsensitive(tag + " close\r\n",
                                 server.DequeueRequest());

      session.Disconnect(false);
    }

    [Test]
    public void TestSelectUidNotSticky()
    {
      using (var session = Authenticate()) {
        // SELECT transaction
        server.EnqueueResponse("* 1 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 1] Message 1 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] Validity session-only\r\n" +
                               "* OK [UIDNEXT 2] Predicted next UID\r\n" +
                               "* NO [UIDNOTSTICKY] Non-persistent UIDs\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen)] Limited\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.SelectedMailbox.UidPersistent);
        Assert.AreEqual(2, session.SelectedMailbox.UidNext);
        Assert.AreEqual(3857529045, session.SelectedMailbox.UidValidity);
        Assert.IsFalse(session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(new Uri(uri, "./funny;UIDVALIDITY=3857529045"), session.SelectedMailbox.Url);
      }
    }

    [Test]
    public void TestSelectHighestModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [HIGHESTMODSEQ 715194045007]\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(715194045007UL, session.SelectedMailbox.HighestModSeq);
        Assert.AreEqual(new Uri(uri, "./funny;UIDVALIDITY=3857529045"), session.SelectedMailbox.Url);
      }
    }

    [Test]
    public void TestSelectNoModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [NOMODSEQ] Sorry, this mailbox format doesn't support modsequences\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(0UL, session.SelectedMailbox.HighestModSeq);
        Assert.AreEqual(new Uri(uri, "./funny;UIDVALIDITY=3857529045"), session.SelectedMailbox.Url);
      }
    }

    [Test]
    public void TestSelectCondStore()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [HIGHESTMODSEQ 715194045007]\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed, CONDSTORE is now enabled\r\n");

        Assert.IsTrue((bool)session.SelectCondStore("INBOX"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"INBOX\" (CONDSTORE)\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue((bool)session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(715194045007UL, session.SelectedMailbox.HighestModSeq);
        Assert.AreEqual(new Uri(uri, "./INBOX;UIDVALIDITY=3857529045"), session.SelectedMailbox.Url);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSelectCondStoreIncapable()
    {
      using (var session = Authenticate()) {
        session.HandlesIncapableAsException = true;

        // SELECT transaction
        session.SelectCondStore("INBOX");
      }
    }

    [Test]
    public void TestLoginInSelectedState()
    {
      using (var session = SelectMailbox()) {
        var selectedMailbox = session.SelectedMailbox;

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        var result = session.Login(credential);

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, Transaction.ImapTransactionResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        Assert.AreSame(selectedMailbox, session.SelectedMailbox);
      }
    }

    [Test]
    public void TestAuthenticateInSelectedState()
    {
      using (var session = SelectMailbox()) {
        var selectedMailbox = session.SelectedMailbox;

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        var result = session.Authenticate(credential, ImapAuthenticationMechanism.Login);

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, Transaction.ImapTransactionResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        Assert.AreSame(selectedMailbox, session.SelectedMailbox);
      }
    }

    [Test]
    public void TestLogoutInSelectedState()
    {
      using (var session = SelectMailbox()) {
        Assert.IsNotNull(session.SelectedMailbox);

        server.EnqueueResponse("* BYE Logging out\r\n" + 
                               "0004 OK logged out.\r\n");

        Assert.IsTrue(session.Logout().Code == Transaction.ImapTransactionResultCode.Bye);

        AssertEqualCaseInsensitive("0004 LOGOUT\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    [Test]
    public void TestNoOpInSelectedState()
    {
      using (var session = SelectMailbox()) {
        Assert.AreEqual(ImapSessionState.Selected, session.State);

        // NOOP transaction
        server.EnqueueResponse("0004 OK NOOP completed\r\n");

        Assert.IsTrue((bool)session.NoOp());

        AssertEqualCaseInsensitive("0004 NOOP\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestNoOpStatusUpdate()
    {
      using (var session = SelectMailbox()) {
        // NOOP transaction
        server.EnqueueResponse("* 22 EXPUNGE\r\n" +
                               "* 18 EXPUNGE\r\n" +
                               "* 23 EXISTS\r\n" +
                               "* 3 RECENT\r\n" +
                               "* 14 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                               "0004 OK NOOP completed\r\n");
  
        Assert.IsTrue((bool)session.NoOp());
  
        AssertEqualCaseInsensitive("0004 noop\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapUpdatedStatusKind.Expunged | ImapUpdatedStatusKind.Exists | ImapUpdatedStatusKind.Recent | ImapUpdatedStatusKind.Messages,
                        session.LastUpdatedStatus.UpdatedKind);
        Assert.AreEqual(1, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsTrue(session.LastUpdatedStatus.MessageAttributes[0].Flags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(session.LastUpdatedStatus.MessageAttributes[0].Flags.Has(ImapMessageFlag.Deleted));
  
        Assert.AreEqual(2, session.LastUpdatedStatus.Expunged.Length);
        Assert.AreEqual(22, session.LastUpdatedStatus.Expunged[0]);
        Assert.AreEqual(18, session.LastUpdatedStatus.Expunged[1]);
  
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(23, (long)session.LastUpdatedStatus.Exists);
  
        Assert.IsNotNull(session.LastUpdatedStatus.Recent);
        Assert.AreEqual(3, (long)session.LastUpdatedStatus.Recent);
  
        Assert.AreEqual(23, session.SelectedMailbox.ExistsMessage, "selected mailbox updated");
        Assert.AreEqual(3, session.SelectedMailbox.RecentMessage, "selected mailbox updated");
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestCheckStatusUpdate()
    {
      using (var session = SelectMailbox()) {
        // CHECK transaction
        server.EnqueueResponse("* 23 EXISTS\r\n" +
                               "0004 OK CHECK completed\r\n");

        Assert.IsTrue((bool)session.Check());
  
        AssertEqualCaseInsensitive("0004 check\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(0, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Expunged);
        Assert.AreEqual(0, session.LastUpdatedStatus.Expunged.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(23, (long)session.LastUpdatedStatus.Exists);
        Assert.IsNull(session.LastUpdatedStatus.Recent);
  
        Assert.AreEqual(23, session.SelectedMailbox.ExistsMessage, "selected mailbox updated");
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestStatus()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" INBOX\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(1, mailboxes.Length);

        // STATUS transaction
        server.EnqueueResponse("* STATUS \"INBOX\" (UIDNEXT 4 UNSEEN 2)\r\n" +
                               "0003 OK STATUS completed\r\n");

        Assert.IsTrue((bool)session.Status(mailboxes[0],
                                           ImapStatusDataItem.UidNext + ImapStatusDataItem.Unseen));

        AssertEqualCaseInsensitive("0003 STATUS \"INBOX\" (UIDNEXT UNSEEN)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(4L, mailboxes[0].UidNext);
        Assert.AreEqual(2L, mailboxes[0].UnseenMessage);
      }
    }

    [Test]
    public void TestStatusUnmanagedMailbox1()
    {
      using (var session = Authenticate()) {
        // STATUS transaction
        server.EnqueueResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)\r\n" +
                               "0002 OK STATUS completed\r\n");

        ImapMailbox statusMailbox;

        Assert.IsTrue((bool)session.Status("blurdybloop",
                                           ImapStatusDataItem.Messages + ImapStatusDataItem.UidNext,
                                           out statusMailbox));

        AssertEqualCaseInsensitive("0002 STATUS \"blurdybloop\" (MESSAGES UIDNEXT)\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(statusMailbox);
        Assert.AreEqual(44292L, statusMailbox.UidNext);
        Assert.AreEqual(231L, statusMailbox.ExistsMessage);
      }
    }

    [Test]
    public void TestStatusUnmanagedMailbox2()
    {
      using (var session = Authenticate()) {
        // STATUS transaction
        server.EnqueueResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)\r\n" +
                               "0002 OK STATUS completed\r\n");

        ImapStatusAttributeList statusAttributes;

        Assert.IsTrue((bool)session.Status("blurdybloop",
                                           ImapStatusDataItem.Messages + ImapStatusDataItem.UidNext,
                                           out statusAttributes));

        AssertEqualCaseInsensitive("0002 STATUS \"blurdybloop\" (MESSAGES UIDNEXT)\r\n",
                                   server.DequeueRequest());

        Assert.IsNull(statusAttributes.HighestModSeq);
        Assert.IsNull(statusAttributes.Recent);
        Assert.IsNull(statusAttributes.UidValidity);
        Assert.IsNull(statusAttributes.Unseen);
        Assert.AreEqual(44292L, statusAttributes.UidNext);
        Assert.AreEqual(231L, statusAttributes.Messages);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestStatusSelectedMailbox()
    {
      using (var session = SelectMailbox()) {
        Assert.IsNotNull(session.SelectedMailbox);

        session.Status(session.SelectedMailbox, ImapStatusDataItem.UidNext);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestStatusNoSelectMailbox()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(2, mailboxes.Length);
        Assert.AreEqual("foo", mailboxes[0].Name);
        Assert.AreEqual("foo/bar", mailboxes[1].Name);

        session.Status(mailboxes[0], ImapStatusDataItem.UidNext);
      }
    }

    [Test]
    public void TestStatusHighestModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // STATUS transaction
        server.EnqueueResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292 HIGHESTMODSEQ 7011231777)\r\n" +
                               "0002 OK STATUS completed\r\n");

        ImapStatusAttributeList statusAttributes;

        Assert.IsTrue((bool)session.Status("blurdybloop", ImapStatusDataItem.Messages + ImapStatusDataItem.UidNext + ImapStatusDataItem.HighestModSeq, out statusAttributes));

        Assert.IsNull(statusAttributes.Recent);
        Assert.IsNull(statusAttributes.Unseen);
        Assert.IsNull(statusAttributes.UidValidity);

        Assert.IsNotNull(statusAttributes.Messages);
        Assert.AreEqual(231, statusAttributes.Messages);
        Assert.IsNotNull(statusAttributes.UidNext);
        Assert.AreEqual(44292, statusAttributes.UidNext);
        Assert.IsNotNull(statusAttributes.HighestModSeq);
        Assert.AreEqual(7011231777UL, statusAttributes.HighestModSeq);

        AssertEqualCaseInsensitive("0002 STATUS \"blurdybloop\" (MESSAGES UIDNEXT HIGHESTMODSEQ)\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestFetchEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateSet(new long[] {}), ImapFetchDataItem.Fast, out messages);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestFetchEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateUidSet(new long[] {}), ImapFetchDataItem.Fast, out messages);
      }
    }

    [Test]
    public void TestFetchMessageInfo()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (" +
                               "UID 4 " + 
                               "FLAGS (\\Answered custom1) " +
                               "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                               "RFC822.SIZE 1716" + 
                               ")\r\n" + 
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
        ImapFetchDataItem dataItem =
          ImapFetchDataItem.Uid +
          ImapFetchDataItem.Flags +
          ImapFetchDataItem.InternalDate +
          ImapFetchDataItem.Rfc822Size;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), dataItem, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (uid flags internaldate rfc822.size)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        ImapMessage message = messages[0];
  
        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(4, message.Uid);
        Assert.AreEqual(1, message.Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=4"), message.Url);

        Assert.IsTrue(message.Flags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(message.Flags.Has("custom1"));
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), message.InternalDate);
        Assert.AreEqual(1716, message.Rfc822Size);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchEnvelope()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (ENVELOPE (" + 
                               "\"Tue, 19 Feb 2008 08:15:48 +0900\" " +
                               "\"fetch test\" " + 
                               "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                               "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                               "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                               "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                               "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                               "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                               "\"<in.reply.to@mail.example.com>\" " + 
                               "\"<message.id@mail.example.com>\" " + 
                               "))\r\n" +
                               "0004 OK FETCH completed\r\n");
    
        ImapMessage[] messages;
    
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Envelope, out messages));
    
        AssertEqualCaseInsensitive("0004 fetch 1 (envelope)\r\n",
                                   server.DequeueRequest());
    
        Assert.AreEqual(1, messages.Length);
    
        ImapMessage message = messages[0];
    
        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), message.Url);

        // envelope
        ImapEnvelope envelope = message.Envelope;
        Assert.AreEqual("Tue, 19 Feb 2008 08:15:48 +0900", envelope.Date);
        Assert.AreEqual("fetch test", envelope.Subject);
        Assert.AreEqual(1, envelope.From.Length);
        Assert.AreEqual("from address", envelope.From[0].Name);
        Assert.AreEqual(1, envelope.Sender.Length);
        Assert.AreEqual("sender address", envelope.Sender[0].Name);
        Assert.AreEqual(1, envelope.ReplyTo.Length);
        Assert.AreEqual("reply-to address", envelope.ReplyTo[0].Name);
        Assert.AreEqual(1, envelope.To.Length);
        Assert.AreEqual("to address", envelope.To[0].Name);
        Assert.AreEqual(1, envelope.Cc.Length);
        Assert.AreEqual("cc address", envelope.Cc[0].Name);
        Assert.AreEqual(1, envelope.Bcc.Length);
        Assert.AreEqual("bcc address", envelope.Bcc[0].Name);
        Assert.AreEqual("<in.reply.to@mail.example.com>", envelope.InReplyTo);
        Assert.AreEqual("<message.id@mail.example.com>", envelope.MessageId);
    
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchBodyStructureSinglePart()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY (\"text\" \"plain\" (\"charset\" \"ISO-2022-JP\") NIL NIL \"7bit\" 22 1))\r\n" +
                               "0004 OK FETCH completed\r\n");

        ImapMessage[] messages;

        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Body, out messages));

        AssertEqualCaseInsensitive("0004 FETCH 1 (BODY)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(1, messages.Length);

        ImapMessage message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), message.Url);

        // body structure
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsFalse(message.BodyStructure.IsMultiPart);
        Assert.IsTrue(message.BodyStructure is ImapSinglePartBodyStructure);

        var structure = message.BodyStructure as ImapSinglePartBodyStructure;

        Assert.AreEqual("text", structure.Type);
        Assert.AreEqual("plain", structure.Subtype);
        Assert.AreEqual(1, structure.Parameters.Count);
        Assert.AreEqual("ISO-2022-JP", structure.Parameters["charset"]);
        Assert.IsNull(structure.Id);
        Assert.IsNull(structure.Description);
        Assert.AreEqual("7bit", structure.Encoding);
        Assert.AreEqual(22, structure.Size);
        Assert.AreEqual(1, structure.Lines);
        Assert.IsNull(structure.BodyStructure);
        Assert.IsNull(structure.Envelope);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchBodyStructureMultiPart()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY "+
                               "((\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 " + 
                               "23)(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\" \"NAME\" \"cc.diff\") " + 
                               "\"<960723163407.20117h@cac.washington.edu>\" \"Compiler diff\" " + 
                               "\"BASE64\" 4554 73) \"MIXED\")" + 
                               ")\r\n" +
                               "0004 OK FETCH completed\r\n");

        ImapMessage[] messages;

        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Body, out messages));

        AssertEqualCaseInsensitive("0004 FETCH 1 (BODY)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(1, messages.Length);

        ImapMessage message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), message.Url);

        // body structure
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsTrue(message.BodyStructure.IsMultiPart);
        Assert.IsTrue(message.BodyStructure is ImapMultiPartBodyStructure);

        var structure = message.BodyStructure as ImapMultiPartBodyStructure;

        Assert.AreEqual("MIXED", structure.Subtype);
        Assert.AreEqual(2, structure.Nested.Length);

        var part1 = structure.Nested[0] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part1);
        Assert.AreEqual("TEXT", part1.Type);
        Assert.AreEqual("PLAIN", part1.Subtype);
        Assert.AreEqual(1, part1.Parameters.Count);
        Assert.AreEqual("US-ASCII", part1.Parameters["CHARSET"]);
        Assert.IsNull(part1.Id);
        Assert.IsNull(part1.Description);
        Assert.AreEqual("7BIT", part1.Encoding);
        Assert.AreEqual(1152, part1.Size);
        Assert.AreEqual(23, part1.Lines);
        Assert.IsNull(part1.BodyStructure);
        Assert.IsNull(part1.Envelope);

        var part2 = structure.Nested[1] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part2);
        Assert.AreEqual("TEXT", part2.Type);
        Assert.AreEqual("PLAIN", part2.Subtype);
        Assert.AreEqual(2, part2.Parameters.Count);
        Assert.AreEqual("US-ASCII", part2.Parameters["CHARSET"]);
        Assert.AreEqual("cc.diff", part2.Parameters["NAME"]);
        Assert.AreEqual("<960723163407.20117h@cac.washington.edu>", part2.Id);
        Assert.AreEqual("Compiler diff", part2.Description);
        Assert.AreEqual("BASE64", part2.Encoding);
        Assert.AreEqual(4554, part2.Size);
        Assert.AreEqual(73, part2.Lines);
        Assert.IsNull(part2.BodyStructure);
        Assert.IsNull(part2.Envelope);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchByMacro()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (" +
                               "FLAGS (\\Answered custom1) " +
                               "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                               "RFC822.SIZE 1716" + 
                               ")\r\n" + 
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Fast, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 FAST\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), messages[0].InternalDate);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchBody()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[] {12}\r\ntest message)\r\n" +
                               "* 2 FETCH (BODY[] \"test message\")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateRangeSet(1, 2), ImapFetchDataItem.BodyPeek(), out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1:2 (body.peek[])\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(2, messages.Length);
  
        //Assert.AreEqual(session.SelectedMailbox, messages[0].Mailbox);
        //Assert.AreEqual(session.SelectedMailbox, messages[1].Mailbox);
  
        Assert.AreEqual("test message", Encoding.ASCII.GetString(messages[0].Body));
        Assert.AreEqual("test message", Encoding.ASCII.GetString(messages[1].Body));

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), messages[0].Url);

        Assert.AreEqual(0L, messages[1].Uid);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?2"), messages[1].Url);

        CloseMailbox(session);
      }
    }

    private string CreateLiteral(string body, Encoding encoding, int maxoctet)
    {
      var literalizedBody = encoding.GetBytes(body);

      var octetBody = (maxoctet == 0) ? literalizedBody.Length : maxoctet;

      return string.Format("{{{0}}}\r\n{1}",
                           octetBody,
                           latin1Encoding.GetString(literalizedBody, 0, octetBody));
    }

    [Test]
    public void TestFetchSection()
    {
      using (var session = SelectMailbox()) {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";
  
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                               CreateLiteral(body, Encoding.ASCII, 0) +
                               ")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.BodyPeek("header.fields (from to subject date)"), out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (body.peek[header.fields (from to subject date)])\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), messages[0].Url);

        Assert.AreEqual(body, Encoding.ASCII.GetString(messages[0].Body));
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchPartial()
    {
      using (var session = SelectMailbox()) {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";
        var octet = 10;
  
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                               CreateLiteral(body, Encoding.ASCII, octet) +
                               ")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        var messageDataItems = ImapFetchDataItem.BodyText("header.fields (from to subject date)", 0, octet);
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), messageDataItems, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (body[header.fields (from to subject date)]<0." + octet.ToString() + ">)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), messages[0].Url);

        Assert.AreEqual(body.Substring(0, octet), Encoding.ASCII.GetString(messages[0].Body));
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchReturnedInSeparatedResponse()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 2 FETCH (INTERNALDATE \"24-Jan-2010 23:08:54 +0900\")\r\n" +
                               "* 2 FETCH (UID 399)\r\n" +
                               "* 2 FETCH (FLAGS ())\r\n" +
                               "0004 OK done\r\n");

        ImapMessage[] messages;

        var messageDataItems = ImapFetchDataItem.Uid + ImapFetchDataItem.Flags + ImapFetchDataItem.InternalDate;

        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(2), messageDataItems, out messages));

        AssertEqualCaseInsensitive("0004 FETCH 2 (UID FLAGS INTERNALDATE)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(399L, messages[0].Uid);
        Assert.AreEqual(2L, messages[0].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=399"), messages[0].Url);

        Assert.AreEqual(0, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].InternalDate.HasValue);
        Assert.AreEqual(new DateTimeOffset(2010, 1, 24, 23, 8, 54, new TimeSpan(9, 0, 0)), messages[0].InternalDate.Value);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchModSeq()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (MODSEQ (624140003))\r\n" +
                               "* 2 FETCH (MODSEQ (624140007))\r\n" +
                               "* 3 FETCH (MODSEQ (624140005))\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateRangeSet(1, 3), ImapFetchDataItem.ModSeq, out messages));
  
        AssertEqualCaseInsensitive("0004 FETCH 1:3 (MODSEQ)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?1"), messages[0].Url);
        Assert.AreEqual(624140003UL, messages[0].ModSeq);

        Assert.AreEqual(0L, messages[1].Uid);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?2"), messages[1].Url);
        Assert.AreEqual(624140007UL, messages[1].ModSeq);

        Assert.AreEqual(0L, messages[2].Uid);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?3"), messages[2].Url);
        Assert.AreEqual(624140005UL, messages[2].ModSeq);

        Assert.AreEqual(0L, messages[2].Uid);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "?3"), messages[2].Url);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestFetchModSeqIncapable()
    {
      using (var session = SelectMailbox()) {
        session.HandlesIncapableAsException = true;

        // FETCH
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateRangeSet(1, 3), ImapFetchDataItem.Fast + ImapFetchDataItem.ModSeq, out messages);
      }
    }

    [Test]
    public void TestFetchChangedSince()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (UID 4 MODSEQ (65402) FLAGS (\\Seen))\r\n" +
                               "* 2 FETCH (UID 6 MODSEQ (75403) FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (UID 8 MODSEQ (29738) FLAGS ($NoJunk $AutoJunk $MDNSent))\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.FetchChangedSince(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.Flags, 12345UL, out messages));
  
        AssertEqualCaseInsensitive("0004 UID FETCH 1:* (FLAGS) (CHANGEDSINCE 12345)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(4L, messages[0].Uid);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=4"), messages[0].Url);
        Assert.AreEqual(65402UL, messages[0].ModSeq);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(6L, messages[1].Uid);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6"), messages[1].Url);
        Assert.AreEqual(75403UL, messages[1].ModSeq);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(4L, messages[2].Sequence);
        Assert.AreEqual(8L, messages[2].Uid);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=8"), messages[2].Url);
        Assert.AreEqual(29738UL, messages[2].ModSeq);
        Assert.IsTrue(messages[2].Flags.Has("$NoJunk"));
        Assert.IsTrue(messages[2].Flags.Has("$AutoJunk"));
        Assert.IsTrue(messages[2].Flags.Has("$MDNSent"));

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestFetchChangedSinceIncapable()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        ImapMessage[] messages;
        session.FetchChangedSince(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.Flags, 12345UL, out messages);
      }
    }

    [Test]
    public void TestStore()
    {
      using (var session = SelectMailbox()) {
        // STORE
        server.EnqueueResponse("* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                               "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                               "0004 OK STORE completed\r\n");

        ImapMessage[] messages;

        Assert.IsTrue((bool)session.Store(ImapSequenceSet.CreateRangeSet(2, 4),
                                    ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted),
                                    out messages));

        AssertEqualCaseInsensitive("0004 STORE 2:4 +FLAGS (\\Deleted)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(2, messages[0].Sequence);
        Assert.AreEqual(0UL, messages[0].ModSeq);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messages[1].Sequence);
        Assert.AreEqual(0UL, messages[1].ModSeq);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(4, messages[2].Sequence);
        Assert.AreEqual(0UL, messages[2].ModSeq);
        Assert.AreEqual(3, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Seen));

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestStoreEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        session.Store(ImapSequenceSet.CreateSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestStoreEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        session.Store(ImapSequenceSet.CreateUidSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
      }
    }

    [Test]
    public void TestStoreUnchangedSince()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // STORE
        server.EnqueueResponse("* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                               "* 5 FETCH (MODSEQ (320162350))\r\n" +
                               "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                               "0004 OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;
  
        Assert.IsTrue((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateSet(7, 5, 9),
                                                  ImapStoreDataItem.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                  320162338UL,
                                                  out messages,
                                                  out failedMessageSet));
  
        AssertEqualCaseInsensitive("0004 STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(7, messages[0].Sequence);
        Assert.AreEqual(320162342UL, messages[0].ModSeq);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(5, messages[1].Sequence);
        Assert.AreEqual(320162350UL, messages[1].ModSeq);
        Assert.AreEqual(0, messages[1].Flags.Count);

        Assert.AreEqual(9, messages[2].Sequence);
        Assert.AreEqual(320162349UL, messages[2].ModSeq);
        Assert.AreEqual(1, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Answered));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(2, failed.Length);
        Assert.AreEqual(7, failed[0]);
        Assert.AreEqual(9, failed[1]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestStoreUnchangedSinceWithTaggedNoResponse()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // STORE
        server.EnqueueResponse("* 1 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "* 3 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "0004 NO [MODIFIED 2] Some of the messages no longer exist.\r\n");

        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;

        Assert.IsFalse((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                                   ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                                   320172338UL,
                                                   out messages,
                                                   out failedMessageSet));

        AssertEqualCaseInsensitive("0004 STORE 1:7 (UNCHANGEDSINCE 320172338) +FLAGS (\\SEEN)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(2, messages.Length);

        Assert.AreEqual(1, messages[0].Sequence);
        Assert.AreEqual(320172342UL, messages[0].ModSeq);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messages[1].Sequence);
        Assert.AreEqual(320172342UL, messages[1].ModSeq);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Seen));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(1, failed.Length);
        Assert.AreEqual(2, failed[0]);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestStoreUnchangedSinceIncapable()
    {
      using (var session = SelectMailbox()) {
        // STORE
        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;
        session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                    ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                    320172338UL,
                                    out messages,
                                    out failedMessageSet);
      }
    }

    [Test]
    public void TestSearch()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Unseen, out matched));

        AssertEqualCaseInsensitive("0004 SEARCH UNSEEN\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2,    messages[0]);
        Assert.AreEqual(84,   messages[1]);
        Assert.AreEqual(882,  messages[2]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchWithLiteralAndDefaultCharset()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Subject("subject"), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH CHARSET US-ASCII SUBJECT {7}\r\n" +
                                   "subject\r\n",
                                   server.DequeueAll());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchWithLiteralAndSpecifiedCharset()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Subject("件名"), Encoding.GetEncoding(932), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH CHARSET SHIFT_JIS SUBJECT {4}\r\n" +
                                   "\x008c\x008f\x0096\x00bc\r\n",
                                   server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchNothingMatched()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Text("string not in mailbox"), Encoding.UTF8, out matched));

        AssertEqualCaseInsensitive("0004 SEARCH CHARSET UTF-8 TEXT {21}\r\n" +
                                   "string not in mailbox\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchWithCriteriaFromUri()
    {
      using (var session = SelectMailbox("LITERAL+")) {
        var b = new ImapUriBuilder(session.SelectedMailbox.Url);

        b.SearchCriteria = ImapSearchCriteria.From("差出人") & (ImapSearchCriteria.Seen | ImapSearchCriteria.Subject("件名"));
        b.Charset = Encoding.UTF8;

        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.FromUri(b.Uri),
                                           out matched));

        AssertEqualCaseInsensitive("0004 SEARCH CHARSET utf-8 FROM {9+}\r\n" +
                                   "\x00E5\x00B7\x00AE\x00E5\x0087\x00BA\x00E4\x00BA\x00BA OR (SEEN) (SUBJECT {6+}\r\n" +
                                   "\x00E4\x00BB\x00B6\x00E5\x0090\x008D)\r\n",
                                   server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test, Ignore("not implemented")]
    public void TestSearchBadCharset()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("0004 NO [BADCHARSET ONDUL EUC-JP] Unknown charset\r\n");
        server.EnqueueResponse("+ OK\r\n");
        server.EnqueueResponse("0005 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Body("ほげほげ"), out matched));
  
        AssertEqualCaseInsensitive("0004 search charset utf-8 BODY {12}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 search charset EUC-JP BODY {8}\r\n",
                                   server.DequeueRequest());
        server.DequeueRequest();
  
        CloseMailbox(session, "0006");
      }
    }

    [Test]
    public void TestSearchCondStoreModSeq()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeq(620162338UL), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH MODSEQ 620162338\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchCondStoreModSeqWithEntry()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH MODSEQ \"/flags/\\\\draft\" all 620162338\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(10, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);
        Assert.AreEqual(23, messages[9]);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSearchCondStoreIncapable()
    {
      using (var session = SelectMailbox()) {
        ImapMatchedSequenceSet matched;

        session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched);
      }
    }

    [Test]
    public void TestESearch()
    {
      using (var session = SelectMailbox("ESEARCH")) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") MIN 2 COUNT 3\r\n" +
                               "0004 OK SEARCH completed\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Min + ImapSearchResultOptions.Count, out matched));

        AssertEqualCaseInsensitive("0004 SEARCH RETURN (MIN COUNT) CHARSET US-ASCII FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);
        Assert.IsNotNull(matched.Tag);
        Assert.AreEqual("0004", matched.Tag);
        Assert.IsNotNull(matched.Min);
        Assert.AreEqual(2, matched.Min);
        Assert.IsNotNull(matched.Count);
        Assert.AreEqual(3, matched.Count);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestESearchIncapable()
    {
      using (var session = SelectMailbox()) {
        ImapMatchedSequenceSet matched;

        session.ESearch(ImapSearchCriteria.All, ImapSearchResultOptions.Min, out matched);
      }
    }

    [Test]
    public void TestESearchSearchres()
    {
      using (var session = SelectMailbox("ESEARCH", "SEARCHRES")) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));

        AssertEqualCaseInsensitive("0004 SEARCH RETURN (SAVE) CHARSET US-ASCII FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        // FETCH transaction
        server.EnqueueResponse("* 2 FETCH (UID 14)\r\n" +
                               "* 84 FETCH (UID 100)\r\n" +
                               "* 882 FETCH (UID 1115)\r\n" +
                               "0005 OK completed\r\n");

        ImapMessage[] fetched;

        Assert.IsTrue((bool)session.Fetch(matched, ImapFetchDataItem.Uid, out fetched));

        AssertEqualCaseInsensitive("0005 FETCH $ (UID)\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session, "0006");
      }
    }

    [Test]
    public void TestESearchSearchresSearchWithSavedResult()
    {
      using (var session = SelectMailbox("ESEARCH", "SEARCHRES")) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));

        AssertEqualCaseInsensitive("0004 SEARCH RETURN (SAVE) CHARSET US-ASCII FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 17 900 901\r\n" +
                               "0005 OK completed\r\n");

        ImapMatchedSequenceSet matched2;

        Assert.IsTrue((bool)session.UidSearch(ImapSearchCriteria.Uid(matched) & ImapSearchCriteria.Smaller(4096), out matched2));

        AssertEqualCaseInsensitive("0005 UID SEARCH UID $ SMALLER 4096\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(matched2.IsUidSet);
        Assert.IsFalse(matched2.IsEmpty);

        var arr = matched2.ToArray();

        Assert.AreEqual(3, arr.Length);
        Assert.AreEqual(17, arr[0]);
        Assert.AreEqual(900, arr[1]);
        Assert.AreEqual(901, arr[2]);

        CloseMailbox(session, "0006");
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestESearchSearchresIncapable()
    {
      using (var session = SelectMailbox("ESEARCH")) {
        // SEARCH transaction
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        session.ESearch(criteria, ImapSearchResultOptions.Save, out matched);
      }
    }

    [Test]
    public void TestSort()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All, out matched));

        AssertEqualCaseInsensitive("0004 SORT (REVERSE SIZE DATE) UTF-8 ALL\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(5, messages.Length);
        Assert.AreEqual(5, messages[0]);
        Assert.AreEqual(3, messages[1]);
        Assert.AreEqual(4, messages[2]);
        Assert.AreEqual(1, messages[3]);
        Assert.AreEqual(2, messages[4]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSortWithLiteralAndDefaultCharset()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                         ImapSearchCriteria.Subject("件名"),
                                         out matched));

        AssertEqualCaseInsensitive("0004 SORT (REVERSE SIZE DATE) UTF-8 SUBJECT {6}\r\n" +
                                   "件名\r\n",
                                   server.DequeueAll());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSortWithLiteralAndSpecifiedCharset()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                         ImapSearchCriteria.Subject("件名"),
                                         Encoding.GetEncoding(932),
                                         out matched));

        AssertEqualCaseInsensitive("0004 SORT (REVERSE SIZE DATE) SHIFT_JIS SUBJECT {4}\r\n" +
                                   "\x008c\x008f\x0096\x00bc\r\n",
                                   server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestSortIncapable()
    {
      using (var session = SelectMailbox()) {
        ImapMatchedSequenceSet matched;

        session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All, out matched);
      }
    }

    [Test]
    public void TestESort()
    {
      using (var session = SelectMailbox("SORT", "ESORT")) {
        // SORT transaction
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") UID ALL 23765,23764,23763,23761\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.UidESort(ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All,
                                       out matched));

        AssertEqualCaseInsensitive("0004 UID SORT RETURN (ALL) (REVERSE DATE) UTF-8 UNDELETED UNKEYWORD $Junk\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(matched.IsUidSet);
        Assert.AreEqual("0004", matched.Tag);

        var messages = matched.ToArray();

        Assert.AreEqual(4, messages.Length);
        Assert.AreEqual(23765, messages[0]);
        Assert.AreEqual(23764, messages[1]);
        Assert.AreEqual(23763, messages[2]);
        Assert.AreEqual(23761, messages[3]);

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestESortIncapable()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.UidESort(ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All,
                                       out matched));
      }
    }

    [Test]
    public void TestThread()
    {
      using (var session = SelectMailbox("THREAD=REFERENCES")) {
        // THREAD transaction
        server.EnqueueResponse("* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList));

        AssertEqualCaseInsensitive("0004 THREAD REFERENCES UTF-8 ALL\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(25, threadList.Children.Length);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestThreadWithLiteralAndDefaultCharset()
    {
      using (var session = SelectMailbox("THREAD=REFERENCES")) {
        // THREAD transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* THREAD (206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References,
                                           ImapSearchCriteria.Subject("件名"),
                                           out threadList));

        AssertEqualCaseInsensitive("0004 THREAD REFERENCES UTF-8 SUBJECT {6}\r\n" +
                                   "件名\r\n",
                                   server.DequeueAll());

        Assert.IsNotNull(threadList);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestThreadWithLiteralAndSpecifiedCharset()
    {
      using (var session = SelectMailbox("THREAD=REFERENCES")) {
        // THREAD transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* THREAD (206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References,
                                           ImapSearchCriteria.Subject("件名"),
                                           Encoding.GetEncoding(932),
                                           out threadList));

        AssertEqualCaseInsensitive("0004 THREAD REFERENCES SHIFT_JIS SUBJECT {4}\r\n" +
                                   "\x008c\x008f\x0096\x00bc\r\n",
                                   server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(threadList);

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestThreadIncapable()
    {
      using (var session = SelectMailbox("THREAD=ORDEREDSUBJECT")) {
        ImapThreadList threadList;

        session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList);
      }
    }

    [Test]
    public void TestExpunge()
    {
      using (var session = SelectMailbox()) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 4 EXPUNGE\r\n" +
                               "* 7 EXPUNGE\r\n" +
                               "* 11 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");
  
        long[] expunged;
  
        Assert.IsTrue((bool)session.Expunge(out expunged));
  
        AssertEqualCaseInsensitive("0004 expunge\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(4, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(4,  expunged[1]);
        Assert.AreEqual(7,  expunged[2]);
        Assert.AreEqual(11, expunged[3]);
  
        CloseMailbox(session);
      }
    }


    [Test]
    public void TestUidExpunge()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");
  
        long[] expunged;
  
        Assert.IsTrue((bool)session.UidExpunge(ImapSequenceSet.CreateUidFromSet(1), out expunged));
  
        AssertEqualCaseInsensitive("0004 UID EXPUNGE 1:*\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(3,  expunged[1]);
        Assert.AreEqual(3,  expunged[2]);
  
        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestExpungeSequenceSet()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        long[] expunged;

        session.UidExpunge(ImapSequenceSet.CreateFromSet(1), out expunged);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestUidExpungeEmptyUidSet()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        long[] expunged;

        session.UidExpunge(ImapSequenceSet.CreateUidSet(new long[] {}), out expunged);
      }
    }

    [Test]
    public void TestCopy()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 OK COPY completed\r\n");
  
        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX"));
  
        AssertEqualCaseInsensitive("0004 copy 1:10 \"INBOX\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestCopyEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        session.Copy(ImapSequenceSet.CreateSet(new long[] {}), "INBOX");
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestCopyEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        session.Copy(ImapSequenceSet.CreateUidSet(new long[] {}), "INBOX");
      }
    }

    [Test]
    public void TestCopyTryCreate()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("0006 OK COPY completed\r\n");
  
        ImapMailbox created;
  
        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateUidRangeSet(1, 10), "INBOX.copyto", out created));
  
        Assert.AreEqual("INBOX.copyto", created.Name);
        Assert.AreEqual(new Uri(uri, "./INBOX.copyto"), created.Url);

        AssertEqualCaseInsensitive("0004 uid copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 create \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0006 uid copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session, "0007");
      }
    }

    [Test]
    public void TestCopyDontTryCreate()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
  
        Assert.IsFalse((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX.copyto"));
  
        AssertEqualCaseInsensitive("0004 copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestCopyWithAppendUidResponseCode()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 OK [COPYUID 38505 304,319:320 3956:3958] Done\r\n");
  
        ImapCopiedUidSet copied;

        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(2, 4), "meeting", out copied));
  
        AssertEqualCaseInsensitive("0004 COPY 2:4 \"meeting\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(copied);
        Assert.AreEqual(38505, copied.UidValidity);

        var copiedUids = copied.CopiedUidSet.ToArray();

        Assert.AreEqual(3, copiedUids.Length);
        Assert.AreEqual(304, copiedUids[0]);
        Assert.AreEqual(319, copiedUids[1]);
        Assert.AreEqual(320, copiedUids[2]);

        var assignedUids = copied.AssignedUidSet.ToArray();

        Assert.AreEqual(3, assignedUids.Length);
        Assert.AreEqual(3956, assignedUids[0]);
        Assert.AreEqual(3957, assignedUids[1]);
        Assert.AreEqual(3958, assignedUids[2]);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppend()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message\r\n",
                                   server.DequeueAll());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendNonSynchronizedLiterals()
    {
      using (var session = SelectMailbox("LITERAL+")) {
        // APPEND transaction
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33+}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message\r\n",
                                   server.DequeueAll());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendTryCreate()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0006 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(new[] {"$label2"}, ImapMessageFlag.Draft));
  
        ImapMailbox created;
  
        Assert.IsTrue((bool)session.Append(message, "INBOX.appendto", out created));
  
        Assert.AreEqual("INBOX.appendto", created.Name);
        Assert.AreEqual(new Uri(uri, "./INBOX.appendto"), created.Url);
  
        AssertEqualCaseInsensitive("0004 append \"INBOX.appendto\" (\\Draft $label2) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 create \"INBOX.appendto\"\r\n",
                                   server.DequeueRequest());

        AssertEqualCaseInsensitive("0006 append \"INBOX.appendto\" (\\Draft $label2) {33}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message\r\n",
                                   server.DequeueAll());
  
        CloseMailbox(session, "0007");
      }
    }

    [Test]
    public void TestAppendDontTryCreate()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new DateTimeOffset(2008, 2, 25, 15, 1, 12, new TimeSpan(+9, 0, 0)));
  
        var internalDate = "\"25-Feb-2008 15:01:12 +0900\"";
  
        Assert.IsFalse((bool)session.Append(message, "INBOX.appendto"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX.appendto\" " + internalDate + " {33}\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendRemoveNonApplicableFlags()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen,
                                                              ImapMessageFlag.Recent,
                                                              ImapMessageFlag.AllowedCreateKeywords));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message\r\n",
                                   server.DequeueAll());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendWithAppendUidResponseCode()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK [APPENDUID 38505 3955] APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.Append(message, "INBOX", out appended));

        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message\r\n",
                                   server.DequeueAll());

        Assert.IsNotNull(appended);
        Assert.AreEqual(38505L, appended.UidValidity);
        Assert.AreEqual(3955L, appended.ToNumber());
        Assert.AreEqual("3955", appended.ToString());
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendBeginAppendEndAppend()
    {
      AppendBeginAppendEndAppend(false);
    }

    [Test]
    public void TestAppendBeginAppendEndAppendNonSynchronizedLiterals()
    {
      AppendBeginAppendEndAppend(false);
    }

    private void AppendBeginAppendEndAppend(bool nonSynchronizedLiterals)
    {
      var messageBody = @"Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)
From: Fred Foobar <foobar@Blurdybloop.example.COM>
Subject: afternoon meeting
To: mooch@owatagu.example.net
Message-Id: <B27397-0100000@Blurdybloop.example.COM>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

Hello Joe, do you think we can meet at 3:30 tomorrow?
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var caps = nonSynchronizedLiterals
        ? new[] {"IMAP4REV1", "LITERAL+"}
        : new[] {"IMAP4REV1"};

      using (var session = Authenticate(caps)) {
        var messageBodyStream = new MemoryStream(NetworkTransferEncoding.Transfer8Bit.GetBytes(messageBody));

        var asyncResult = session.BeginAppend(messageBodyStream, null, null, "INBOX");

        Assert.IsNotNull(asyncResult);
        Assert.IsFalse(asyncResult.IsCompleted);
        Assert.IsTrue(session.IsTransactionProceeding);

        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0002 OK [APPENDUID 38505 3955] APPEND completed\r\n");

        if (!asyncResult.AsyncWaitHandle.WaitOne(1000))
          Assert.Fail("append timed out");

        Assert.IsTrue(asyncResult.IsCompleted);

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)session.EndAppend(asyncResult, out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding);

        var expected =
          string.Format("0002 APPEND \"INBOX\" {{{0}}}\r\n", messageBodyStream.Length) +
          messageBody +
          "\r\n";

        AssertEqualCaseInsensitive(expected, server.DequeueAll());

        Assert.IsNotNull(appendedUid);
        Assert.AreEqual(38505L, appendedUid.UidValidity);
        Assert.AreEqual(3955L, appendedUid.ToNumber());
      }
    }

    [Test]
    public void TestAppendBeginAppendNestedCall()
    {
      using (var session = Authenticate()) {
        var messageBodyStream = new MemoryStream(Encoding.ASCII.GetBytes("message"));

        var asyncResult = session.BeginAppend(messageBodyStream, null, null, "INBOX");

        Assert.IsNotNull(asyncResult);
        Assert.IsFalse(asyncResult.IsCompleted);
        Assert.IsTrue(session.IsTransactionProceeding);

        try {
          session.BeginAppend(messageBodyStream, null, null, "INBOX");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
        }

        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0002 OK APPEND completed\r\n");

        if (!asyncResult.AsyncWaitHandle.WaitOne(1000))
          Assert.Fail("append timed out");

        Assert.IsTrue(asyncResult.IsCompleted);

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)session.EndAppend(asyncResult, out appendedUid));

        AssertEqualCaseInsensitive("0002 APPEND \"INBOX\" {7}\r\nmessage\r\n",
                                   server.DequeueAll());

        Assert.IsNull(appendedUid);
      }
    }

    [Test]
    public void TestAppendBeginAppendContinuationNo()
    {
      using (var session = Authenticate()) {
        var messageBodyStream = new MemoryStream(Encoding.ASCII.GetBytes("message"));

        var asyncResult = session.BeginAppend(messageBodyStream, null, null, "INBOX");

        Assert.IsNotNull(asyncResult);
        Assert.IsFalse(asyncResult.IsCompleted);
        Assert.IsTrue(session.IsTransactionProceeding);

        server.EnqueueResponse("0002 NO APPEND failed\r\n");

        if (!asyncResult.AsyncWaitHandle.WaitOne(1000))
          Assert.Fail("append timed out");

        Assert.IsTrue(asyncResult.IsCompleted);

        ImapAppendedUidSet appendedUid;

        Assert.IsFalse((bool)session.EndAppend(asyncResult));

        AssertEqualCaseInsensitive("0002 APPEND \"INBOX\" {7}\r\n",
                                   server.DequeueAll());
      }
    }

    [Test]
    public void TestAppendEndAppendInvalidAsyncResult()
    {
      using (var session = Authenticate()) {
        try {
          session.EndAppend(null);
          Assert.Fail("ArgumentException not thrown");
        }
        catch (ArgumentException) {
        }
      }
    }

    [Test]
    public void TestAppendMultiple()
    {
      AppendMultiple(false);
    }

    [Test]
    public void TestAppendMultipleWithAppendUidResponseCode()
    {
      AppendMultiple(true);
    }

    private void AppendMultiple(bool respondAppendUid)
    {
      using (var session = SelectMailbox("MULTIAPPEND")) {
        // APPEND transaction
        server.EnqueueResponse("+ Ready for literal data\r\n");
        server.EnqueueResponse(string.Empty);
        server.EnqueueResponse("+ Ready for literal data\r\n");
        server.EnqueueResponse(string.Empty);

        if (respondAppendUid)
          server.EnqueueResponse("0004 OK [APPENDUID 38505 3955:3956] APPEND completed\r\n");
        else
          server.EnqueueResponse("0004 OK APPEND completed\r\n");

        var messageBody1 = @"Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)
From: Fred Foobar <foobar@Blurdybloop.example.COM>
Subject: afternoon meeting
To: mooch@owatagu.example.net
Message-Id: <B27397-0100000@Blurdybloop.example.COM>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

Hello Joe, do you think we can meet at 3:30 tomorrow?
".Replace("\r\n", "\n").Replace("\n", "\r\n");
        var messageBody2 = @"Date: Mon, 7 Feb 1994 22:43:04 -0800 (PST)
From: Joe Mooch <mooch@OWaTaGu.example.net>
Subject: Re: afternoon meeting
To: foobar@blurdybloop.example.com
Message-Id: <a0434793874930@OWaTaGu.example.net>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

3:30 is fine with me.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var message1 = new ImapMessage(Encoding.ASCII.GetBytes(messageBody1),
                                       new ImapMessageFlagList(ImapMessageFlag.Seen));
        var message2 = new ImapMessage(Encoding.ASCII.GetBytes(messageBody2),
                                       new DateTimeOffset(new DateTime(1994, 2, 7, 22, 43, 04), new TimeSpan(-8, 0, 0)),
                                       new ImapMessageFlagList(ImapMessageFlag.Seen));

        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages", out appended));

        var expected =
          string.Format("0004 APPEND \"saved-messages\" (\\Seen) {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody1)) +
          messageBody1 +
          string.Format(" (\\Seen) \"07-Feb-1994 22:43:04 -0800\" {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody2)) +
          messageBody2 +
          "\r\n";

        AssertEqualCaseInsensitive(expected, server.DequeueAll());

        if (respondAppendUid) {
          Assert.IsNotNull(appended);
          Assert.AreEqual(38505L, appended.UidValidity);

          var appendedUids = appended.ToArray();

          Assert.AreEqual(2, appendedUids.Length);
          Assert.AreEqual(3955L, appendedUids[0]);
          Assert.AreEqual(3956L, appendedUids[1]);
        }

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendMultipleWithNonSyncedLiteral()
    {
      using (var session = SelectMailbox("MULTIAPPEND", "LITERAL+")) {
        // APPEND transaction
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message1 = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message1"));
        var message2 = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message2"));
        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages"));

        AssertEqualCaseInsensitive("0004 APPEND \"saved-messages\" {34+}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message1 {34+}\r\n" +
                                   "MIME-Version: 1.0\r\n" +
                                   "\r\n" +
                                   "test message2\r\n",
                                   server.DequeueAll());

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestAppendMultipleIncapable()
    {
      using (var session = SelectMailbox()) {
        session.HandlesIncapableAsException = true;
        // APPEND transaction
        session.AppendMultiple(new[] {
          new ImapMessage(Encoding.ASCII.GetBytes("message")),
          new ImapMessage(Encoding.ASCII.GetBytes("message2")),
        }, "saved-messages");
      }
    }

    [Test]
    public void TestUnselect()
    {
      using (var session = SelectMailbox("UNSELECT")) {
        // EXPUNGE transaction
        server.EnqueueResponse("0004 OK UNSELECT completed\r\n");
  
        Assert.IsTrue((bool)session.Unselect());
  
        AssertEqualCaseInsensitive("0004 unselect\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    private ImapSession Authenticate(params string[] additionalCapabilities)
    {
      // greeting and login transaction
      var session = Connect();

      server.EnqueueResponse("0000 OK authenticated\r\n");

      session.Login(credential);

      Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);

      server.DequeueRequest();

      // CAPABILITY transaction
      server.EnqueueResponse(string.Format("* CAPABILITY {0} IMAP4rev1\r\n", string.Join(" ", additionalCapabilities)) +
                             "0001 OK CAPABILITY completed\r\n");

      session.Capability();

      server.DequeueRequest();

      return session;
    }

    [Test]
    public void TestNoOpInAuthenticatedState()
    {
      using (var session = Authenticate()) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        // NOOP transaction
        server.EnqueueResponse("0002 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        AssertEqualCaseInsensitive("0002 NOOP\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestCreate()
    {
      using (var session = Authenticate()) {
        // CREATE transaction
        server.EnqueueResponse("0002 OK CREATE completed\r\n");

        ImapMailbox createdMailbox;

        Assert.IsTrue((bool)session.Create("INBOX.日本語", out createdMailbox));

        AssertEqualCaseInsensitive("0002 create \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("INBOX.日本語", createdMailbox.Name);
        Assert.AreEqual(new Uri(uri, "./INBOX.日本語"), createdMailbox.Url);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestCreateInbox()
    {
      using (var session = Authenticate()) {
        ImapMailbox createdMailbox;

        session.Create("INBOX", out createdMailbox);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestCreateExistentMailbox()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX.日本語", mailboxes[0].Name);

        ImapMailbox createdMailbox;

        session.Create(mailboxes[0].Name, out createdMailbox);
      }
    }

    [Test]
    public void TestDelete()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("blurdybloop", mailboxes[0].Name);

        // DELETE transaction
        server.EnqueueResponse("0003 OK DELETE completed\r\n");

        Assert.IsTrue((bool)session.Delete(mailboxes[0]));

        AssertEqualCaseInsensitive("0003 delete \"blurdybloop\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestDeleteUnmanagedMailbox()
    {
      using (var session = Authenticate()) {
        // DELETE transaction
        server.EnqueueResponse("0002 OK DELETE completed\r\n");

        Assert.IsTrue((bool)session.Delete("INBOX.☺"));

        AssertEqualCaseInsensitive("0002 DELETE \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestDeleteInbox1()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].IsInbox);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);

        session.Delete(mailboxes[0]);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestDeleteInbox2()
    {
      using (var session = Authenticate()) {
        session.Delete("inbox");
      }
    }

    [Test]
    public void TestDeleteSelectedMailbox()
    {
      using (var session = Authenticate()) {
        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueResponse("* 17 EXISTS\r\n" +
                               "* 2 RECENT\r\n" +
                               "* OK [UNSEEN 8] Message 8 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS ()] No permanent flags permitted\r\n" +
                               "0002 OK [READ-WRITE] EXAMINE completed\r\n");

        Assert.IsTrue((bool)session.Select("blurdybloop"));

        AssertEqualCaseInsensitive("0002 select \"blurdybloop\"\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.Selected, session.State);

        // DELETE transaction
        server.EnqueueResponse("0003 OK DELETE completed\r\n");

        Assert.IsNotNull(session.SelectedMailbox);
        Assert.IsTrue((bool)session.Delete(session.SelectedMailbox));

        AssertEqualCaseInsensitive("0003 delete \"blurdybloop\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      }
    }

    [Test]
    public void TestDeleteHierarchy()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(3, mailboxes.Length);
        Assert.AreEqual("foo", mailboxes[1].Name);
        Assert.AreEqual("foo/bar", mailboxes[2].Name);

        // DELETE transaction
        server.EnqueueResponse("0003 OK DELETE completed\r\n");

        Assert.IsTrue((bool)session.Delete(mailboxes[2]));

        AssertEqualCaseInsensitive("0003 delete \"foo/bar\"\r\n",
                                   server.DequeueRequest());

        server.EnqueueResponse("0004 OK DELETE completed\r\n");

        Assert.IsTrue((bool)session.Delete(mailboxes[1]));

        AssertEqualCaseInsensitive("0004 delete \"foo\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestDeleteParentMailbox()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(3, mailboxes.Length);
        Assert.AreEqual("foo", mailboxes[1].Name);

        session.Delete(mailboxes[1]);
      }
    }

    [Test]
    public void TestRename()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(2, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.AreEqual("INBOX.日本語", mailboxes[1].Name);
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter);

        // REMANE transaction
        server.EnqueueResponse("0003 OK RENAME completed\r\n");

        ImapMailbox renamingMailbox = mailboxes[1];
        ImapMailbox renamedMailbox;

        Assert.AreEqual(new Uri(uri, "./INBOX.日本語"), renamingMailbox.Url);

        Assert.IsTrue((bool)session.Rename(renamingMailbox, "INBOX.☺", out renamedMailbox));

        AssertEqualCaseInsensitive("0003 rename \"INBOX.&ZeVnLIqe-\" \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(renamedMailbox);
        Assert.AreEqual("INBOX.☺", renamedMailbox.Name);
        Assert.IsTrue(object.ReferenceEquals(renamingMailbox, renamedMailbox));
        Assert.AreEqual(new Uri(uri, "./INBOX.☺"), renamedMailbox.Url);
      }
    }

    [Test]
    public void TestRenameInbox1()
    {
      using (var session = Authenticate()) {
        // REMANE transaction
        server.EnqueueResponse("0002 OK RENAME completed\r\n");

        ImapMailbox renamedMailbox;

        Assert.IsTrue((bool)session.Rename("INBOX", "INBOX.☺", out renamedMailbox));

        AssertEqualCaseInsensitive("0002 RENAME \"INBOX\" \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual("INBOX.☺", renamedMailbox.Name);
        Assert.IsFalse(renamedMailbox.IsInbox);
      }
    }

    [Test]
    public void TestRenameInbox2()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);

        // REMANE transaction
        server.EnqueueResponse("0003 OK RENAME completed\r\n");

        ImapMailbox inbox = mailboxes[0];
        ImapMailbox renamedMailbox;

        Assert.IsTrue(inbox.IsInbox);
        Assert.IsTrue((bool)session.Rename(inbox, "INBOX.☺", out renamedMailbox));

        AssertEqualCaseInsensitive("0003 RENAME \"INBOX\" \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(renamedMailbox);
        Assert.AreEqual("INBOX.☺", renamedMailbox.Name);
        Assert.IsTrue(object.ReferenceEquals(inbox, renamedMailbox));
        Assert.IsFalse(renamedMailbox.IsInbox);
      }
    }

    [Test]
    public void TestRenameUnmanagedMailbox()
    {
      using (var session = Authenticate()) {
        // REMANE transaction
        server.EnqueueResponse("0002 OK RENAME completed\r\n");

        ImapMailbox renamedMailbox;

        Assert.IsTrue((bool)session.Rename("INBOX", "INBOX.☺", out renamedMailbox));

        AssertEqualCaseInsensitive("0002 RENAME \"INBOX\" \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(renamedMailbox);
        Assert.AreEqual("INBOX.☺", renamedMailbox.Name);
      }
    }

    [Test]
    public void TestRenameHierarchy()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(3, mailboxes.Length);
        Assert.AreEqual("blurdybloop", mailboxes[0].Name);
        Assert.AreEqual("foo", mailboxes[1].Name);
        Assert.AreEqual("foo/bar", mailboxes[2].Name);

        // REMANE transaction
        server.EnqueueResponse("0003 OK RENAME completed\r\n");

        Assert.IsTrue((bool)session.Rename("foo", "zowie"));

        AssertEqualCaseInsensitive("0003 RENAME \"foo\" \"zowie\"\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual("blurdybloop", mailboxes[0].Name);
        Assert.AreEqual("zowie", mailboxes[1].Name);
        Assert.AreEqual("zowie/bar", mailboxes[2].Name);
      }
    }

    [Test]
    public void TestRenameSelectedMailbox()
    {
      using (var session = Authenticate()) {
        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueResponse("* 17 EXISTS\r\n" +
                               "* 2 RECENT\r\n" +
                               "* OK [UNSEEN 8] Message 8 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS ()] No permanent flags permitted\r\n" +
                               "0002 OK [READ-WRITE] EXAMINE completed\r\n");

        Assert.IsTrue((bool)session.Select("blurdybloop"));

        AssertEqualCaseInsensitive("0002 select \"blurdybloop\"\r\n",
                                   server.DequeueRequest());

        // REMANE transaction
        server.EnqueueResponse("0003 OK RENAME completed\r\n");

        Assert.IsNotNull(session.SelectedMailbox);

        var selectedMailbox = session.SelectedMailbox;

        Assert.IsTrue((bool)session.Rename(selectedMailbox, "foo"));

        AssertEqualCaseInsensitive("0003 RENAME \"blurdybloop\" \"foo\"\r\n",
                                   server.DequeueRequest());

        Assert.AreSame(selectedMailbox, session.SelectedMailbox);
        Assert.AreEqual("foo", selectedMailbox.Name);
      }
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestRenameToExistMailbox()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \"/\" blurdybloop\r\n" +
                               "* LIST (\\Noselect) \"/\" foo\r\n" +
                               "* LIST () \"/\" foo/bar\r\n" +
                               "0002 OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        session.List(out mailboxes);

        server.DequeueRequest();

        Assert.AreEqual(3, mailboxes.Length);
        Assert.AreEqual("blurdybloop", mailboxes[0].Name);
        Assert.AreEqual("foo", mailboxes[1].Name);

        session.Rename("blurdybloop", "foo");
      }
    }

    [Test]
    public void TestRenameMailboxReferralAsException()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List(out mailboxes);
  
        server.DequeueRequest();
  
        // REMANE transaction
        server.EnqueueResponse("0003 NO [REFERRAL IMAP://user;AUTH=*@SERVER1/FOO IMAP://user;AUTH=*@SERVER2/BAR] Unable to rename mailbox across servers\r\n");
  
        try {
          session.HandlesReferralAsException = true;
  
          ImapMailbox renamedMailbox;
          session.Rename(mailboxes[1], "INBOX.☺", out renamedMailbox);
  
          Assert.Fail("renamed without exception");
        }
        catch (ImapMailboxReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Unable to rename mailbox across servers"));
          Assert.AreEqual(2, ex.Referrals.Length);
          Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER1/FOO"), ex.Referrals[0]);
          Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER2/BAR"), ex.Referrals[1]);
        }
      }
    }

    [Test]
    public void TestSubscribe()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List(out mailboxes);
  
        server.DequeueRequest();
  
        // SUBSCRIBE transaction
        server.EnqueueResponse("0003 OK SUBSCRIBE completed\r\n");
  
        Assert.IsTrue((bool)session.Subscribe(mailboxes[1]));
  
        AssertEqualCaseInsensitive("0003 subscribe \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestUnsubscribe()
    {
      using (var session = Authenticate()) {
        // LSUB transaction
        server.EnqueueResponse("* LSUB () \".\" \"INBOX\"\r\n" +
                               "* LSUB () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LSUB completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.Lsub("INBOX*", out mailboxes);
  
        server.DequeueRequest();
  
        // UNSUBSCRIBE transaction
        server.EnqueueResponse("0003 OK UNSUBSCRIBE completed\r\n");
  
        Assert.IsTrue((bool)session.Unsubscribe(mailboxes[1]));
  
        AssertEqualCaseInsensitive("0003 unsubscribe \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestNamespace()
    {
      using (var session = Authenticate("NAMESPACE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Namespace));
  
        // NAMESPACE transaction
        server.EnqueueResponse("* NAMESPACE ((\"\" \"/\")) ((\"~\" \"/\")) ((\"#shared/\" \"/\")(\"#public/\" \"/\")(\"#ftp/\" \"/\")(\"#news.\" \".\"))\r\n" +
                               "0002 OK NAMESPACE command completed\r\n");
  
        ImapNamespace namespaces = null;
  
        Assert.IsTrue((bool)session.Namespace(out namespaces));
  
        AssertEqualCaseInsensitive("0002 NAMESPACE\r\n",
                                   server.DequeueRequest());
  
        Assert.IsNotNull(namespaces);
        Assert.AreEqual(1, namespaces.PersonalNamespaces.Length);
        Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);
  
        Assert.AreEqual(1, namespaces.OtherUsersNamespaces.Length);
        Assert.AreEqual("~", namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.OtherUsersNamespaces[0].Extensions.Count);
  
        Assert.AreEqual(4, namespaces.SharedNamespaces.Length);
        Assert.AreEqual("#shared/", namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("#public/", namespaces.SharedNamespaces[1].Prefix);
        Assert.AreEqual("#ftp/", namespaces.SharedNamespaces[2].Prefix);
        Assert.AreEqual("#news.", namespaces.SharedNamespaces[3].Prefix);
  
        Assert.IsTrue(namespaces.PersonalNamespaces.Length == session.Namespaces.PersonalNamespaces.Length);
        Assert.IsTrue(namespaces.OtherUsersNamespaces.Length == session.Namespaces.OtherUsersNamespaces.Length);
        Assert.IsTrue(namespaces.SharedNamespaces.Length == session.Namespaces.SharedNamespaces.Length);
        Assert.IsTrue(namespaces.PersonalNamespaces[0].Prefix == session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.OtherUsersNamespaces[0].Prefix == session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[0].Prefix == session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[1].Prefix == session.Namespaces.SharedNamespaces[1].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[2].Prefix == session.Namespaces.SharedNamespaces[2].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[3].Prefix == session.Namespaces.SharedNamespaces[3].Prefix);
      }
    }

    [Test]
    public void TestIdleBeginIdleEndIdle()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        session.SendTimeout         = 100;
        session.ReceiveTimeout      = 100;
        session.TransactionTimeout  = 100;

        // IDLE
        server.EnqueueResponse("+ idling\r\n");

        var asyncResult = session.BeginIdle();

        Assert.IsNotNull(asyncResult);
        Assert.IsTrue(session.IsIdling);
        Assert.IsTrue(session.IsTransactionProceeding);
        Assert.AreEqual(100, session.SendTimeout);
        Assert.AreNotEqual(100, session.ReceiveTimeout);
        Assert.AreEqual(100, session.TransactionTimeout);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());

        Thread.Sleep(500);

        // DONE
        server.EnqueueResponse("0004 OK done.\r\n");

        Assert.IsTrue((bool)session.EndIdle(asyncResult));

        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);
        Assert.AreEqual(100, session.SendTimeout);
        Assert.AreEqual(100, session.ReceiveTimeout);
        Assert.AreEqual(100, session.TransactionTimeout);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestIdleBeginIdleNo()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        server.EnqueueResponse("0004 NO not allowed at this time.\r\n");

        Assert.IsFalse((bool)session.EndIdle(session.BeginIdle()));

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.IsIdling);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestIdleBeginIdleNestedCall()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        server.EnqueueResponse("+ idling\r\n");
        server.EnqueueResponse("0004 OK done.\r\n");

        var asyncResult = session.BeginIdle();

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());

        try {
          session.BeginIdle();
          Assert.Fail("ImapProtocolViolationException not thrown");
        }
        catch (ImapProtocolViolationException) {
        }

        session.EndIdle(asyncResult);

        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestIdleBeginIdleIncapable()
    {
      using (var session = SelectMailbox()) {
        Assert.IsFalse(session.IsIdling);

        session.HandlesIncapableAsException = true;

        session.BeginIdle();
      }
    }

    [Test]
    public void TestIdleEndIdleInvalidAsyncResult()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        try {
          session.EndIdle(null);
          Assert.Fail("ArgumentException not thrown");
        }
        catch (ArgumentException) {
        }

        CloseMailbox(session, "0004");
      }
    }

    [Test]
    public void TestIdleStatusUpdate()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        session.SendTimeout         = 50;
        session.ReceiveTimeout      = 50;
        session.TransactionTimeout  = 50;

        // IDLE
        server.EnqueueResponse("* 2 EXPUNGE\r\n" +
                               "+ idling\r\n" +
                               "* 3 EXISTS\r\n");

        var asyncResult = session.BeginIdle();

        Assert.IsNotNull(asyncResult);
        Assert.IsTrue(session.IsIdling);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());

        Thread.Sleep(500);

        Assert.IsFalse(3 == session.SelectedMailbox.ExistsMessage);

        server.EnqueueResponse("* 4 EXISTS\r\n" +
                               "0004 OK IDLE terminated\r\n");

        Assert.IsTrue((bool)session.EndIdle(asyncResult));

        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.IsIdling);
        Assert.AreEqual(4, session.SelectedMailbox.ExistsMessage);

        Assert.AreEqual(ImapUpdatedStatusKind.Expunged | ImapUpdatedStatusKind.Exists,
                        session.LastUpdatedStatus.UpdatedKind);

        Assert.AreEqual(0, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Expunged);
        Assert.AreEqual(1, session.LastUpdatedStatus.Expunged.Length);
        Assert.AreEqual(2, session.LastUpdatedStatus.Expunged[0]);
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(4, (long)session.LastUpdatedStatus.Exists);
        Assert.IsNull(session.LastUpdatedStatus.Recent);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestIdle()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        server.EnqueueResponse("+ idling\r\n");
        server.EnqueueResponse("0004 OK done.\r\n");

        var sw = new Stopwatch();

        sw.Reset();
        sw.Start();

        Assert.IsTrue((bool)session.Idle(500));

        sw.Stop();

        Assert.Greater(sw.ElapsedMilliseconds, 500);
        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestIdleNo()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        server.EnqueueResponse("0004 NO not allowed at this time.\r\n");

        var sw = new Stopwatch();

        sw.Reset();
        sw.Start();

        Assert.IsFalse((bool)session.Idle(500));

        sw.Stop();

        Assert.Less(sw.ElapsedMilliseconds, 500);
        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session);
      }
    }

    private class IdleState {
      public ImapSession Session;
      public long CurrentMessageCount;
    }

    [Test]
    public void TestIdleKeepIdleCallback()
    {
      using (var session = Authenticate("IDLE")) {
        // SELECT
        server.EnqueueResponse("* FLAGS (\\Deleted \\Seen)\r\n" +
                               "* 4 EXISTS\r\n" +
                               "* 0 RECENT\r\n" +
                               "* OK [UIDVALIDITY 1]\r\n" +
                               "0002 OK SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("INBOX"));

        AssertEqualCaseInsensitive("0002 SELECT \"INBOX\"\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(4, session.SelectedMailbox.ExistsMessage);

        // IDLE
        Assert.IsFalse(session.IsIdling);

        server.WaitForRequest = false;

        session.SendTimeout         = 50;
        session.ReceiveTimeout      = 50;
        session.TransactionTimeout  = 50;

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          var s = state as ImapPseudoServer;

          s.EnqueueResponse("* 2 EXPUNGE\r\n" +
                            "* 3 EXISTS\r\n" +
                            "+ idling\r\n");

          Thread.Sleep(250);

          // ...time passes; another client expunges message 3...
          s.EnqueueResponse("* 3 EXPUNGE\r\n" +
                            "* 2 EXISTS\r\n");

          Thread.Sleep(250);

          // ...time passes; new mail arrives...
          s.EnqueueResponse("* 3 EXISTS\r\n");

          s.EnqueueResponse("0003 OK IDLE terminated\r\n");
        }, server);

        var idleState = new IdleState();

        idleState.Session = session;
        idleState.CurrentMessageCount = session.SelectedMailbox.ExistsMessage;

        Assert.IsTrue((bool)session.Idle(1000, idleState, delegate(object state, ImapUpdatedStatus updatedStatus) {
          var s = state as IdleState;

          if (updatedStatus.Expunged != null)
            s.CurrentMessageCount -= updatedStatus.Expunged.Length;

          if (updatedStatus.Exists.HasValue) {
            if (s.CurrentMessageCount < updatedStatus.Exists.Value)
              return false;
            else
              s.CurrentMessageCount = updatedStatus.Exists.Value;
          }

          return true;
        }));

        AssertEqualCaseInsensitive("0003 IDLE\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3L, session.SelectedMailbox.ExistsMessage);

        CloseMailbox(session, "0004");
      }
    }

    [Test]
    public void TestIdleKeepIdleCallbackTimeout()
    {
      using (var session = SelectMailbox("IDLE")) {
        // IDLE
        Assert.IsFalse(session.IsIdling);

        session.SendTimeout         = 50;
        session.ReceiveTimeout      = 50;
        session.TransactionTimeout  = 50;

        server.WaitForRequest = false;

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          var s = state as ImapPseudoServer;

          s.EnqueueResponse("+ idling\r\n");

          Thread.Sleep(250);

          s.EnqueueResponse("* 3 EXPUNGE\r\n" +
                            "* 2 EXISTS\r\n");

          Thread.Sleep(250);

          s.EnqueueResponse("0004 OK done.\r\n");
        }, server);

        var sw = new Stopwatch();

        sw.Reset();
        sw.Start();

        Assert.IsTrue((bool)session.Idle(500, null, delegate(object state, ImapUpdatedStatus updatedStatus) {
          return true;
        }));

        sw.Stop();

        Assert.GreaterOrEqual(sw.ElapsedMilliseconds, 500);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("DONE\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestIdleKeepIdleCallbackException()
    {
      using (var session = SelectMailbox("IDLE")) {
        // IDLE
        Assert.IsFalse(session.IsIdling);

        session.SendTimeout         = 50;
        session.ReceiveTimeout      = 50;
        session.TransactionTimeout  = 50;

        server.WaitForRequest = false;

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          var s = state as ImapPseudoServer;

          s.EnqueueResponse("+ idling\r\n");

          Thread.Sleep(250);

          s.EnqueueResponse("* 3 EXPUNGE\r\n" +
                            "* 2 EXISTS\r\n");
        }, server);

        try {
          session.Idle(1000, null, delegate(object state, ImapUpdatedStatus updatedStatus) {
            throw new NullReferenceException();
          });

          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException ex) {
          Assert.IsInstanceOfType(typeof(NullReferenceException), ex.InnerException);
        }

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        //Assert.IsFalse(session.IsIdling); throws NullReferenceException
        Assert.IsFalse(session.IsTransactionProceeding);

        AssertEqualCaseInsensitive("0004 IDLE\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestEnable()
    {
      using (var session = Authenticate("ENABLE", "X-GOOD-IDEA")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Enable));
  
        // ENABLE transaction
        server.EnqueueResponse("* ENABLED X-GOOD-IDEA\r\n" +
                               "0002 OK ENABLE completed\r\n");
  
        ImapCapabilityList enabledCapas;
  
        Assert.IsTrue((bool)session.Enable(out enabledCapas, "CONDSTORE", "X-GOOD-IDEA"));
  
        AssertEqualCaseInsensitive("0002 ENABLE CONDSTORE X-GOOD-IDEA\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, enabledCapas.Count);
        Assert.IsTrue(enabledCapas.Has("X-GOOD-IDEA"));
      }
    }

    [Test]
    public void TestGenericCommandNoWithArguments()
    {
      using (var session = Authenticate()) {
        // generic command transaction
        server.EnqueueResponse("* X-EXT-COMMAND data\r\n" +
                               "0002 OK X-EXT-COMMAND completed\r\n");
  
        ImapDataResponse[] dataResps;
  
        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps));
  
        AssertEqualCaseInsensitive("0002 X-EXT-COMMAND\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, dataResps.Length);
        Assert.AreEqual(2, dataResps[0].Data.Length);
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[0].Format);
        Assert.AreEqual("X-EXT-COMMAND", dataResps[0].Data[0].Text.ToString());
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[1].Format);
        Assert.AreEqual("data", dataResps[0].Data[1].Text.ToString());
      }
    }

    [Test]
    public void TestGenericCommandWithArguments()
    {
      using (var session = Authenticate()) {
        // generic command transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0002 OK X-EXT-COMMAND completed\r\n");
  
        ImapDataResponse[] dataResps;
  
        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps,
                                           new ImapParenthesizedString("arg1",
                                                                       "arg2",
                                                                       new ImapParenthesizedString("arg3-1", "arg3-2"),
                                                                       new ImapLiteralString("ascii text", NetworkTransferEncoding.Transfer8Bit)),
                                           new ImapParenthesizedString(new ImapQuotedString("quoted"),
                                                                       "nonquoted")
                                           ));
  
        AssertEqualCaseInsensitive("0002 X-EXT-COMMAND (arg1 arg2 (arg3-1 arg3-2) {10}\r\n" +
                                   "ascii text) (\"quoted\" nonquoted)\r\n",
                                   server.DequeueAll());
  
        Assert.AreEqual(0, dataResps.Length);
      }
    }

    [Test]
    public void TestID()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));

        // ID transaction
        server.EnqueueResponse("* ID NIL\r\n" +
                               "0002 OK ID completed\r\n");

        Dictionary<string, string> serverParameterList;

        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", "sodr"},
          {"version", "19.34"},
          {"vendor", "Pink Floyd Music Limited"},
        }, out serverParameterList));

        AssertEqualCaseInsensitive("0002 ID (\"name\" \"sodr\" \"version\" \"19.34\" \"vendor\" \"Pink Floyd Music Limited\")\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(serverParameterList);
        Assert.AreEqual(0, serverParameterList.Count);
        Assert.AreEqual(0, session.ServerID.Count);

        try {
          session.ServerID.Add("key", "value");
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestIDParameterNil()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));

        // ID transaction
        server.EnqueueResponse("* ID (\"name\" \"Cyrus\" \"version\" \"1.5\")\r\n" +
                               "0002 OK ID completed\r\n");

        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", null},
          {"version", null},
          {"vendor", null},
        }));

        AssertEqualCaseInsensitive("0002 ID (\"name\" NIL \"version\" NIL \"vendor\" NIL)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(2, session.ServerID.Count);
        Assert.AreEqual("Cyrus", session.ServerID["name"]);
        Assert.AreEqual("1.5", session.ServerID["version"]);

        try {
          session.ServerID.Add("key", "value");
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIDTooMatchFieldValuePair()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        for (var i = 0; i < 31; i++) {
          paramList.Add(string.Format("field{0}", i), string.Format("value{0}", i));
        }
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIDTooLongField()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {new string('x', 31), "value"},
        }));
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIDTooLongValue()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"field", new string('x', 1025)},
        }));
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIDContaisNoAsciiChar()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"オワタ", "＼(＾o＾)／"},
        }));
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestLanguageIncapable()
    {
      using (var session = Authenticate()) {
        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.Language));

        session.HandlesIncapableAsException = true;

        session.Language();
      }
    }

    [Test]
    public void TestLanguageListSupported()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (EN DE IT i-default)\r\n" +
                               "0002 OK Supported languages have been enumerated\r\n");

        string[] supportedLanguageTags;

        Assert.IsTrue((bool)session.Language(out supportedLanguageTags));

        AssertEqualCaseInsensitive("0002 LANGUAGE\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(4, supportedLanguageTags.Length);
        Assert.AreEqual("EN", supportedLanguageTags[0]);
        Assert.AreEqual("DE", supportedLanguageTags[1]);
        Assert.AreEqual("IT", supportedLanguageTags[2]);
        Assert.AreEqual("i-default", supportedLanguageTags[3]);
      }
    }

    [Test]
    public void TestLanguageSelectDefault()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (DE)\r\n" +
                               "0002 OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n");

        string selectedLanguageTag;

        var result = session.Language(out selectedLanguageTag);

        Assert.IsTrue((bool)result);
        Assert.AreEqual("DE", selectedLanguageTag);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        AssertEqualCaseInsensitive("0002 LANGUAGE default\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual("DE", session.SelectedLanguage);
      }
    }

    [Test]
    public void TestLanguageSelectSupported()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (EN)\r\n" +
                               "0002 OK Now speaking English\r\n");

        string selectedLanguageTag;

        var result = session.Language(out selectedLanguageTag, "FR-CA", "EN-CA");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("EN", selectedLanguageTag);
        Assert.AreEqual("Now speaking English", result.TaggedStatusResponse.ResponseText.Text);

        AssertEqualCaseInsensitive("0002 LANGUAGE FR-CA EN-CA\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual("EN", session.SelectedLanguage);
      }
    }

    [Test]
    public void TestLanguageSelectWithCultureInfo()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("0002 NO Unsupported language\r\n");

        var result = session.Language(System.Globalization.CultureInfo.GetCultureInfo("ja-JP"));

        Assert.IsFalse((bool)result);
        Assert.AreEqual("Unsupported language", result.TaggedStatusResponse.ResponseText.Text);

        AssertEqualCaseInsensitive("0002 LANGUAGE ja-JP\r\n",
                                   server.DequeueRequest());

        Assert.IsNull(session.SelectedLanguage);
      }
    }

    [Test]
    public void TestLanguageNamespaceResponse()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (DE)\r\n" +
                               "* NAMESPACE ((\"\" \"/\")) ((\"Other Users/\" \"/\" \"TRANSLATION\"" +
                                  " (\"Andere Ben&APw-tzer/\"))) ((\"Public Folders/\" \"/\" " +
                                  " \"TRANSLATION\" (\"Gemeinsame Postf&AM8-cher/\")))\r\n" +
                               "0002 OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n");

        var result = session.Language("DE");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        AssertEqualCaseInsensitive("0002 LANGUAGE DE\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual("DE", session.SelectedLanguage);

        Assert.AreEqual(1, session.Namespaces.PersonalNamespaces.Length);
        Assert.AreEqual(string.Empty, session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.PersonalNamespaces[0].HierarchyDelimiter);

        Assert.AreEqual(1, session.Namespaces.OtherUsersNamespaces.Length);
        Assert.AreEqual("Other Users/", session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual("Andere Benützer/", session.Namespaces.OtherUsersNamespaces[0].TranslatedPrefix);

        Assert.AreEqual(1, session.Namespaces.SharedNamespaces.Length);
        Assert.AreEqual("Public Folders/", session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.SharedNamespaces[0].HierarchyDelimiter);
        // RFC error? (Gemeinsame Postfächer)
        Assert.AreEqual("Gemeinsame PostfÏcher/", session.Namespaces.SharedNamespaces[0].TranslatedPrefix);
      }
    }

    [Test]
    public void TestLanguageReceiveResponseAsUTF8()
    {
      using (var session = Authenticate("LANGUAGE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (JA)\r\n" +
                               "0002 OK 完了\r\n");

        var result = session.Language("JA");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);

        AssertEqualCaseInsensitive("0002 LANGUAGE JA\r\n",
                                   server.DequeueRequest());

        // following command/response
        server.EnqueueResponse("0003 OK 完了\r\n");

        session.NoOp();

        Assert.IsTrue((bool)result);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestComparatorIncapable()
    {
      using (var session = Authenticate()) {
        Assert.IsFalse(session.ServerCapabilities.Has(ImapCapability.I18NLevel2));

        session.HandlesIncapableAsException = true;

        session.Comparator();
      }
    }

    [Test]
    public void TestComparator()
    {
      using (var session = Authenticate("I18NLEVEL=2")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.I18NLevel2));
        Assert.AreEqual(ImapCollationAlgorithm.Default, session.ActiveComparator);

        // COMPARATOR transaction
        server.EnqueueResponse("* COMPARATOR i;unicode-casemap\r\n" +
                               "0002 OK Will use i;unicode-casemap for collation\r\n");

        ImapCollationAlgorithm activeComparator;

        Assert.IsTrue((bool)session.Comparator(out activeComparator));

        AssertEqualCaseInsensitive("0002 COMPARATOR\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, session.ActiveComparator);
      }
    }

    [Test]
    public void TestComparatorChangeComparator()
    {
      using (var session = Authenticate("I18NLEVEL=2")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.I18NLevel2));
        Assert.AreEqual(ImapCollationAlgorithm.Default, session.ActiveComparator);

        // COMPARATOR transaction
        server.EnqueueResponse("* COMPARATOR i;basic\r\n" +
                               "0002 OK Will use i;basic for collation\r\n");

        ImapCollationAlgorithm activeComparator;
        ImapCollationAlgorithm[] matchingComparators;

        Assert.IsTrue((bool)session.Comparator(out activeComparator,
                                               out matchingComparators,
                                               new ImapCollationAlgorithm("cz;*"),
                                               new ImapCollationAlgorithm("i;basic")));

        AssertEqualCaseInsensitive("0002 COMPARATOR \"cz;*\" \"i;basic\"\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), activeComparator);
        Assert.AreEqual(0, matchingComparators.Length);

        Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), session.ActiveComparator);
      }
    }

    private int port;
    private string host;
    private Uri uri;
    private const string username = "imapuser";
    private const string password = "mikumiku";
    private NetworkCredential credential;

    private Encoding latin1Encoding = Encoding.GetEncoding("latin1");

    private ImapPseudoServer server;
  }
}
